#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "xmalloc.h"
#include "hxd.h"

static struct htlc_chat __chat_list, *chat_list = &__chat_list;

static u_int32_t chat_ref = 1;

int
chat_isset (struct htlc_conn *htlc, struct htlc_chat *chat)
{
	return FD_ISSET(htlc->uid, &chat->fds);
}

void
chat_set (struct htlc_conn *htlc, struct htlc_chat *chat)
{
	FD_SET(htlc->uid, &chat->fds);
	chat->nusers++;
}

void
chat_clr (struct htlc_conn *htlc, struct htlc_chat *chat)
{
	FD_CLR(htlc->uid, &chat->fds);
	chat->nusers--;
}

struct htlc_chat *
chat_lookup_ref (u_int32_t ref)
{
	struct htlc_chat *chatp;

	for (chatp = chat_list; chatp; chatp = chatp->next)
		if (chatp->ref == ref)
			return chatp;

	return 0;
}

struct htlc_chat *
chat_new (void)
{
	struct htlc_chat *chat;

	chat = xmalloc(sizeof(*chat));
	chat->ref = chat_ref++;
	FD_ZERO(&chat->fds);
	chat->nusers = 0;
	chat->subject[0] = 0;
	chat->next = chat_list->next;
	chat->prev = chat_list;
	if (chat_list->next)
		chat_list->next->prev = chat;
	chat_list->next = chat;

	return chat;
}

void
chat_delete (struct htlc_chat *chat)
{
	if (chat->prev)
		chat->prev->next = chat->next;
	if (chat->next)
		chat->next->prev = chat->prev;
	xfree(chat);
}

void
chat_remove_from_all (struct htlc_conn *htlc)
{
	struct htlc_chat *chatp, *next;

	for (chatp = chat_list; chatp; chatp = next) {
		next = chatp->next;
		if (chat_isset(htlc, chatp)) {
			chat_clr(htlc, chatp);
			if (!chatp->nusers) {
				chat_delete(chatp);
			} else {
				struct htlc_conn *htlcp;
				u_int32_t ref;
				u_int16_t uid;

				ref = htonl(chatp->ref);
				uid = htons(htlc->uid);
				for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
					if (chat_isset(htlcp, chatp)) {
						hlwrite(htlcp, HTLS_HDR_CHAT_USER_PART, 0, 2,
							HTLS_DATA_CHAT_ID, sizeof(ref), &ref,
							HTLS_DATA_UID, sizeof(uid), &uid);
					}
				}
			}
		}
	}
}

void
rcv_chat_create (struct htlc_conn *htlc)
{
	struct htlc_conn *htlcp;
	u_int32_t uid = 0, ref;
	u_int16_t uid16, icon, color;

	dh_start(htlc)
		if (ntohs(dh->type) != HTLC_DATA_UID)
			continue;
		dh_getint(uid);
		if ((htlcp = isclient(uid))) {
			struct htlc_chat *chat = chat_new();

			chat_set(htlc, chat);
			ref = htonl(chat->ref);
			uid16 = htons(htlc->uid);
			icon = htons(htlc->icon);
			color = htons(htlc->color);
			hlwrite(htlc, HTLS_HDR_TASK, 0, 5,
				HTLS_DATA_CHAT_ID, sizeof(ref), &ref,
				HTLS_DATA_UID, sizeof(uid16), &uid16,
				HTLS_DATA_ICON, sizeof(icon), &icon,
				HTLS_DATA_COLOUR, sizeof(color), &color,
				HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
			if (htlc->uid != (u_int16_t)uid)
				hlwrite(htlcp, HTLS_HDR_CHAT_INVITE, 0, 3,
					HTLS_DATA_CHAT_ID, sizeof(ref), &ref,
					HTLS_DATA_UID, sizeof(uid16), &uid16,
					HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
		} else
			hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "who?!?");
	dh_end()
}

void
rcv_chat_invite (struct htlc_conn *htlc)
{
	struct htlc_conn *htlcp;
	u_int32_t uid = 0, ref = 0;
	u_int16_t uid16;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_UID:
				dh_getint(uid);
				break;
			case HTLC_DATA_CHAT_ID:
				dh_getint(ref);
				break;
		}
	dh_end()
	if (ref && uid) {
		if ((htlcp = isclient(uid))) {
			struct htlc_chat *chat = chat_lookup_ref(ref);
			if (!chat || !chat_isset(htlc, chat)) {
				hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "who?!?");
				return;
			}
			if (chat_isset(htlcp, chat)) {
				char buf[64];
				u_int16_t len;

				len = snprintf(buf, sizeof(buf),
					       "%s is already in chat 0x%x",
						htlcp->name, chat->ref);
				if (len == (u_int16_t)-1)
					len = sizeof(buf);
				hlwrite(htlc, HTLS_HDR_TASK, 1, 1,
					HTLS_DATA_TASKERROR, len, buf);
				return;
			}
			hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
			ref = htonl(ref);
			uid16 = htons(htlc->uid);
			hlwrite(htlcp, HTLS_HDR_CHAT_INVITE, 0, 3,
				HTLS_DATA_CHAT_ID, sizeof(ref), &ref,
				HTLS_DATA_UID, sizeof(uid16), &uid16,
				HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
		} else
			hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "who?!?");
	} else
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
}

void
rcv_chat_join (struct htlc_conn *htlc)
{
	u_int32_t ref = 0;
	u_int16_t uid, icon16, color;
	struct htlc_chat *chat;

	dh_start(htlc)
		if (ntohs(dh->type) != HTLC_DATA_CHAT_ID)
			continue;
		dh_getint(ref);
		if ((chat = chat_lookup_ref(ref))) {
			chat_set(htlc, chat);
			{
			struct qbuf *q = &htlc->out;
			u_int32_t this_off = q->pos + q->len, pos = this_off + SIZEOF_HL_HDR;
			struct hl_hdr h;
			struct hl_userlist_hdr uh;
			u_int16_t nlen;
			struct htlc_conn *htlcp;

			q->len += SIZEOF_HL_HDR;
			q->buf = xrealloc(q->buf, q->pos + q->len);
			ref = htonl(ref);
			uid = htons(htlc->uid);
			icon16 = htons(htlc->icon);
			color = htons(htlc->color);
			for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
				if (!chat_isset(htlcp, chat))
					continue;
				if (htlcp != htlc)
					hlwrite(htlcp, HTLS_HDR_CHAT_USER_CHANGE, 0, 5,
						HTLS_DATA_CHAT_ID, sizeof(ref), &ref,
						HTLS_DATA_UID, sizeof(uid), &uid,
						HTLS_DATA_ICON, sizeof(icon16), &icon16,
						HTLS_DATA_COLOUR, sizeof(color), &color,
						HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
				nlen = strlen(htlcp->name);
				uh.type = htons(HTLS_DATA_USER_LIST);
				uh.len = htons(8 + nlen);
				uh.uid = htons(htlcp->uid);
				uh.icon = htons(htlcp->icon);
				uh.color = htons(htlcp->color);
				uh.nlen = htons(nlen);
				q->len += SIZEOF_HL_USERLIST_HDR + nlen;
				q->buf = xrealloc(q->buf, q->pos + q->len);
				memcpy(&q->buf[pos], &uh, SIZEOF_HL_USERLIST_HDR);
				pos += SIZEOF_HL_USERLIST_HDR;
				memcpy(&q->buf[pos], htlcp->name, nlen);
				pos += nlen;
			}

			h.type = htonl(HTLS_HDR_TASK);
			h.trans = htonl(htlc->trans);
			htlc->trans++;
			h.uid = 0;
			h.flag = 0;
			h.len = h.len2 = htonl((pos - this_off) - (SIZEOF_HL_HDR - sizeof(h.hc)));
			h.hc = htons(chat->nusers);
			memory_copy(q->buf + this_off, &h, SIZEOF_HL_HDR);

			FD_SET(htlc->fd, &hxd_wfds);
			}
		} else
			hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "who?!?");
	dh_end()
}

void
rcv_chat_part (struct htlc_conn *htlc)
{
	u_int32_t ref = 0;
	struct htlc_chat *chat;

	dh_start(htlc)
		if (ntohs(dh->type) != HTLC_DATA_CHAT_ID)
			continue;
		dh_getint(ref);
		if ((chat = chat_lookup_ref(ref)) && chat_isset(htlc, chat)) {
			chat_clr(htlc, chat);
			if (!chat->nusers)
				chat_delete(chat);
			else {
				struct htlc_conn *htlcp;
				u_int16_t uid;

				ref = htonl(ref);
				uid = htons(htlc->uid);
				for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next)
					if (chat_isset(htlcp, chat))
						hlwrite(htlcp, HTLS_HDR_CHAT_USER_PART, 0, 2,
							HTLS_DATA_CHAT_ID, sizeof(ref), &ref,
							HTLS_DATA_UID, sizeof(uid), &uid);
			}
		}
	dh_end()
}

void
rcv_chat_decline (struct htlc_conn *htlc __attribute__((__unused__)))
{
#if 0
	dh_start(htlc)
	dh_end()
#endif
}

void
rcv_chat_subject (struct htlc_conn *htlc)
{
	u_int32_t ref = 0;
	struct htlc_chat *chat;
	struct htlc_conn *htlcp;
	u_int8_t subject[256];
	u_int16_t slen = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_CHAT_ID:
				dh_getint(ref);
				break;
			case HTLC_DATA_CHAT_SUBJECT:
				slen = (u_int16_t)ntohs(dh->len) >= sizeof(subject) ?
					sizeof(subject) - 1 :
					ntohs(dh->len);
				memcpy(subject, dh->data, slen);
				subject[slen] = 0;
				break;
		}
	dh_end()
	if ((chat = chat_lookup_ref(ref)) && chat_isset(htlc, chat) && slen) {
		memcpy(chat->subject, subject, slen);
		chat->subject[slen] = 0;
		ref = htonl(ref);
		for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
			if (!chat_isset(htlcp, chat))
				continue;
			hlwrite(htlcp, HTLS_HDR_CHAT_SUBJECT, 0, 2,
				HTLS_DATA_CHAT_ID, sizeof(ref), &ref,
				HTLS_DATA_CHAT_SUBJECT, slen, subject);
		}
	} else {
		u_int16_t style = 0;

		hlwrite(htlc, HTLS_HDR_MSG, 0, 2,
			HTLS_DATA_STYLE, sizeof(style), &style,
			HTLS_DATA_MSG, 6, "huh?!?");
	}
}
