#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include "hxd.h"
#include "xmalloc.h"
#ifdef CONFIG_HOPE
#include "md5.h"
#include "sha.h"
#include "haval.h"
#endif
#ifdef SOCKS
#include "socks.h"
#endif
#ifdef CONFIG_HFS
#include "hfs.h"
#endif

#if !defined(NAME_MAX) && defined(FILENAME_MAX)
#define NAME_MAX FILENAME_MAX
#endif

#if defined(__hpux__) && NAME_MAX == 14
#define NAME_MAX 255
#endif

#ifdef CONFIG_HTXF_PREVIEW
#include "magick/magick.h"
#endif

#if defined(CONFIG_HTXF_PTHREAD)
#include <pthread.h>
#elif defined(CONFIG_HTXF_CLONE)
#include <sched.h>
#define CLONE_STACKSIZE	0x10000
#endif

#define HTXF_BUFSIZE		0xf000
#define COPY_BUFSIZE		0xf000
#define DIRCHAR			'/'
#define ROOTDIR			(htlc->rootdir)
#define UNKNOWN_TYPECREA	"TEXTR*ch"

#define log_list	0
#define log_comment	0
#define log_rename	1
#define log_move	1
#define log_mkdir	1
#define log_symlink	1
#define log_delete	1
#define log_getinfo	0
#define log_hash	0
#define log_download	1
#define log_upload	1

u_int16_t nr_gets = 0;
u_int16_t nr_puts = 0;

static int
hldir_to_path (struct hl_data_hdr *dh, const char *root, char *rlpath)
{
	u_int16_t rootlen, pos, count;
	u_int8_t *p, nlen;
	char path[MAXPATHLEN];

	rootlen = pos = strlen(root);
	memcpy(path, root, pos);
	path[pos++] = DIRCHAR;
	count = ntohs(*((u_int16_t *)dh->data));
	if (count > 32)
		count = 32;
	/* compatibility with broken hx */
	if (ntohs(dh->len) < 5) {
		strcpy(rlpath, root);
		return 0;
	}
	p = &dh->data[4];
	for (;;) {
		nlen = *p++;
		if (pos + nlen >= MAXPATHLEN) {
			return ENAMETOOLONG;
		}
		if (!(nlen == 2 && p[0] == '.' && p[1] == '.') && !memchr(p, DIRCHAR, nlen)) {
			memcpy(&path[pos], p, nlen);
			pos += nlen;
		}
		if (--count) {
			path[pos++] = DIRCHAR;
			p += 2 + nlen;
		} else {
			path[pos] = 0;
			break;
		}
	}
	if (!realpath(path, rlpath))
		return errno;

	return 0;
}

static void
read_filename (char *filename, const char *in, size_t len)
{
	if ((len == 2 && in[0] == '.' && in[1] == '.') || memchr(in, DIRCHAR, len)) {
		len = 0;
	} else {
		memcpy(filename, in, len);
	}
	filename[len] = 0;
}

#ifdef CONFIG_HTXF_PREVIEW
static int
preview_path (char *previewpath, const char *path, struct stat *statbuf)
{
	int i, len;
	char pathbuf[MAXPATHLEN];
	struct stat sb;

	len = strlen(path);
	if (len + 16 >= MAXPATHLEN)
		return ENAMETOOLONG;
	strcpy(pathbuf, path);
	for (i = len - 1; i > 0; i--)
		if (pathbuf[i] == DIRCHAR) {
			pathbuf[i++] = 0;
			break;
		}

	snprintf(previewpath, MAXPATHLEN, "%s/.preview", pathbuf);
	if (stat(previewpath, &sb)) {
		if (statbuf)
			return ENOENT;
		if (mkdir(previewpath, hxd_cfg.permissions.directories))
			return errno;
	}
	pathbuf[--i] = DIRCHAR;
	strcat(previewpath, &pathbuf[i]);
	if (statbuf && stat(previewpath, statbuf))
		return errno;

	return 0;
}
#endif

#if defined(CONFIG_HFS)
static int
finderinfo_path (char *infopath, const char *path, struct stat *statbuf)
{
	int i, len;
	char pathbuf[MAXPATHLEN];
	struct stat sb;

	len = strlen(path);
	if (len + 16 >= MAXPATHLEN)
		return ENAMETOOLONG;
	strcpy(pathbuf, path);
	for (i = len - 1; i > 0; i--)
		if (pathbuf[i] == DIRCHAR) {
			pathbuf[i++] = 0;
			break;
		}

	switch (hxd_cfg.files.fork) {
		default:
		case HFS_FORK_CAP:
			snprintf(infopath, MAXPATHLEN, "%s/.finderinfo", pathbuf);
			break;
		case HFS_FORK_NETATALK:
			if (!stat(path, &sb) && S_ISDIR(sb.st_mode)) {
				snprintf(infopath, MAXPATHLEN, "%s/.AppleDouble", path);
				strcpy(&pathbuf[i], ".Parent");
			} else
				snprintf(infopath, MAXPATHLEN, "%s/.AppleDouble", pathbuf);
			break;
		case HFS_FORK_DOUBLE:
			snprintf(infopath, MAXPATHLEN, "%s/%%%s", pathbuf, &pathbuf[i]);
			goto no_dir;
	}
	if (stat(infopath, &sb)) {
		if (statbuf)
			return ENOENT;
		if (mkdir(infopath, hxd_cfg.permissions.directories))
			return errno;
	}
	pathbuf[--i] = DIRCHAR;
	strcat(infopath, &pathbuf[i]);

no_dir:
	if (statbuf && stat(infopath, statbuf))
		return errno;

	return 0;
}

static int
resource_path (char *rsrcpath, const char *path, struct stat *statbuf)
{
	int i, len;
	char pathbuf[MAXPATHLEN];
	struct stat sb;

	len = strlen(path);
	if (len + 16 >= MAXPATHLEN)
		return ENAMETOOLONG;
	strcpy(pathbuf, path);
	for (i = len - 1; i > 0; i--)
		if (pathbuf[i] == DIRCHAR) {
			pathbuf[i++] = 0;
			break;
		}

	snprintf(rsrcpath, MAXPATHLEN, "%s/.resource", pathbuf);
	if (stat(rsrcpath, &sb)) {
		if (statbuf)
			return ENOENT;
		if (mkdir(rsrcpath, hxd_cfg.permissions.directories))
			return errno;
	}
	pathbuf[--i] = DIRCHAR;
	strcat(rsrcpath, &pathbuf[i]);
	if (statbuf && stat(rsrcpath, statbuf))
		return errno;

	return 0;
}

static int
resource_open (const char *path, int mode, int perm)
{
	if (hxd_cfg.files.fork == HFS_FORK_CAP) {
		char rsrcpath[MAXPATHLEN];

		if (resource_path(rsrcpath, path, 0))
			return -1;
		return open(rsrcpath, mode, perm);
	} else { /* AppleDouble */
		char infopath[MAXPATHLEN];
		struct hfs_dbl_hdr dbl;
		struct stat sb;
		ssize_t r;
		u_int16_t i;
		int f;

		if (!finderinfo_path(infopath, path, &sb)) {
			f = open(infopath, O_RDONLY);
			if (f < 0)
				return 0;
			r = read(f, &dbl, SIZEOF_HFS_DBL_HDR);
			if (r != SIZEOF_HFS_DBL_HDR)
				goto funkdat;
			if (ntohs(dbl.entries) > HFS_HDR_MAX)
				dbl.entries = htons(HFS_HDR_MAX);
			r = read(f, &dbl.descrs, SIZEOF_HFS_HDR_DESCR * ntohs(dbl.entries));
			if (r != SIZEOF_HFS_HDR_DESCR * ntohs(dbl.entries))
				goto funkdat;
			for (i = 0; i < ntohs(dbl.entries); i++) {
				struct hfs_hdr_descr *descr = (struct hfs_hdr_descr *)(&dbl.descrs[SIZEOF_HFS_HDR_DESCR * i]);
				if (ntohl(descr->id) == HFS_HDR_RSRC) {
					if (lseek(f, ntohl(descr->offset), SEEK_SET) != (off_t)ntohl(descr->offset))
						goto funkdat;
					return f;
				}
			}
funkdat:
			close(f);
			return -1;
		}
	}

	return -1;
}

static size_t
resource_len (const char *path)
{
	size_t len = 0;

	if (hxd_cfg.files.fork == HFS_FORK_CAP) {
		char rsrcpath[MAXPATHLEN];
		struct stat sb;

		if (!resource_path(rsrcpath, path, &sb))
			len = sb.st_size;
	} else { /* AppleDouble */
		char infopath[MAXPATHLEN];
		struct hfs_dbl_hdr dbl;
		struct stat sb;
		ssize_t r;
		u_int16_t i;
		int f;

		if (!finderinfo_path(infopath, path, &sb)) {
			f = open(infopath, O_RDONLY);
			if (f < 0)
				return 0;
			r = read(f, &dbl, SIZEOF_HFS_DBL_HDR);
			if (r != SIZEOF_HFS_DBL_HDR)
				goto funkdat;
			if (ntohs(dbl.entries) > HFS_HDR_MAX)
				dbl.entries = htons(HFS_HDR_MAX);
			r = read(f, &dbl.descrs, SIZEOF_HFS_HDR_DESCR * ntohs(dbl.entries));
			if (r != SIZEOF_HFS_HDR_DESCR * ntohs(dbl.entries))
				goto funkdat;
			for (i = 0; i < ntohs(dbl.entries); i++) {
				struct hfs_hdr_descr *descr = (struct hfs_hdr_descr *)(&dbl.descrs[SIZEOF_HFS_HDR_DESCR * i]);
				if (ntohl(descr->id) == HFS_HDR_RSRC) {
					len = ntohl(descr->length);
					break;
				}
			}
funkdat:
			close(f);
		}
	}

	return len;
}

static inline char *
suffix (const char *path)
{
	char *p = (char *)path + strlen(path);

	while (p-- > path)
		if (*p == '.')
			return p + 1;
	return 0;
}

static void
suffix_type_creator (char *buf, const char *path)
{
	u_int8_t *tc = UNKNOWN_TYPECREA;
	char *suff;

	if (!(suff = suffix(path)))
		goto cpy;

	if (!strcmp(suff, "jpg") || !strcmp(suff, "jpeg"))
		tc = "JPEGGKON";
	else if (!strcmp(suff, "png"))
		tc = "PNGfGKON";
	else if (!strcmp(suff, "gif"))
		tc = "GIFfGKON";
	else if (!strcmp(suff, "mp3") || !strcmp(suff, "mp2"))
		tc = "MP3 MAmp";
	else if (!strcmp(suff, "mpg") || !strcmp(suff, "mpeg"))
		tc = "MPEGTVOD";
	else if (!strcmp(suff, "mov"))
		tc = "MooVTVOD";
	else if (!strcmp(suff, "sit"))
		tc = "SITDSIT!";
	else if (!strcmp(suff, "zip"))
		tc = "ZIP ZIP ";
	else if (!strcmp(suff, "app") || !strcmp(suff, "sea"))
		tc = "APPLpeff";
	else if (!strcmp(suff, "img"))
		tc = "rohdWrap";
	else if (!strcmp(suff, "pict"))
		tc = "PICTGKON";

cpy:
	memcpy(buf, tc, 8);
}

static void
type_creator (u_int8_t *buf, const char *path)
{
	char infopath[MAXPATHLEN];
	struct stat sb;

	if (!finderinfo_path(infopath, path, &sb)) {
		int r, f;

		f = open(infopath, O_RDONLY);
		if (f < 0)
			goto use_suffix;
		switch (hxd_cfg.files.fork) {
			default:
			case HFS_FORK_CAP:
				r = read(f, buf, 8);
				if (r != 8 || !(*(u_int32_t *)(&buf[0])) || !(*(u_int32_t *)(&buf[4])))
					r = 0;
				break;
			case HFS_FORK_NETATALK:
			case HFS_FORK_DOUBLE:
				{
					struct hfs_dbl_hdr dbl;
					int i;

					r = read(f, &dbl, SIZEOF_HFS_DBL_HDR);
					if (r != SIZEOF_HFS_DBL_HDR) {
						r = -1;
						break;
					}
					if (ntohs(dbl.entries) > HFS_HDR_MAX)
						dbl.entries = htons(HFS_HDR_MAX);
					r = read(f, &dbl.descrs, SIZEOF_HFS_HDR_DESCR * ntohs(dbl.entries));
					if (r != SIZEOF_HFS_HDR_DESCR * ntohs(dbl.entries)) {
						r = -1;
						break;
					}
					for (i = 0; i < ntohs(dbl.entries); i++) {
						struct hfs_hdr_descr *descr = (struct hfs_hdr_descr *)(&dbl.descrs[SIZEOF_HFS_HDR_DESCR * i]);
						if (ntohl(descr->id) == HFS_HDR_FINFO) {
							if (lseek(f, ntohl(descr->offset), SEEK_SET) != (off_t)ntohl(descr->offset))
								continue;
							r = 8;
							r = read(f, buf, 8);
							break;
						}
					}
				}
				break;
		}
		close(f);
		if (r == 8)
			return;
	}

use_suffix:
	suffix_type_creator(buf, path);
}

struct hfsinfo {
	u_int8_t type[4];
	u_int8_t creator[4];
	u_int32_t create_time;
	u_int32_t modify_time;
	u_int32_t comlen;
	u_int8_t comment[200];
};

static void
hfsinfo_read (const char *path, struct hfsinfo *fi)
{
	char infopath[MAXPATHLEN];
	int f, i, r;
	struct stat sb;
	union {
		struct hfs_cap_info cap;
		struct hfs_dbl_hdr dbl;
	} hdr;

	memset(fi, 0, sizeof(struct hfsinfo));
	if (!finderinfo_path(infopath, path, &sb)) {
		f = open(infopath, O_RDONLY);
		if (f >= 0) {
			switch (hxd_cfg.files.fork) {
				case HFS_FORK_CAP:
					r = read(f, &hdr.cap, SIZEOF_HFS_CAP_INFO);
					if (r != SIZEOF_HFS_CAP_INFO)
						break;
					memcpy(fi->type, hdr.cap.fi_fndr, 8);
					if (hdr.cap.fi_datevalid & HFS_CAP_CDATE)
						memcpy(&fi->create_time, hdr.cap.fi_ctime, 4);
					else
						fi->create_time = 0;
					if (hdr.cap.fi_datevalid & HFS_CAP_MDATE)
						memcpy(&fi->modify_time, hdr.cap.fi_mtime, 4);
					else
						fi->modify_time = 0;
					fi->comlen = hdr.cap.fi_comln > 200 ? 200 : hdr.cap.fi_comln;
					memcpy(fi->comment, hdr.cap.fi_comnt, fi->comlen);
					break;
				case HFS_FORK_NETATALK:
				case HFS_FORK_DOUBLE:
					r = read(f, &hdr.dbl, SIZEOF_HFS_DBL_HDR);
					if (r != SIZEOF_HFS_DBL_HDR)
						break;
					if (ntohs(hdr.dbl.entries) > HFS_HDR_MAX)
						hdr.dbl.entries = htons(HFS_HDR_MAX);
					r = read(f, &hdr.dbl.descrs, SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries));
					if (r != SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries))
						break;
					for (i = 0; i < ntohs(hdr.dbl.entries); i++) {
						struct hfs_hdr_descr *descr = (struct hfs_hdr_descr *)(&hdr.dbl.descrs[SIZEOF_HFS_HDR_DESCR * i]);
						if (lseek(f, ntohl(descr->offset), SEEK_SET) != (off_t)ntohl(descr->offset))
							continue;
						switch (ntohl(descr->id)) {
							case HFS_HDR_COMNT:
								fi->comlen = ntohl(descr->length) > 200 ? 200 : ntohl(descr->length);
								r = read(f, fi->comment, fi->comlen);
								if (r != (int)fi->comlen)
									fi->comlen = 0;
								break;
							case HFS_HDR_OLDI:
							case HFS_HDR_DATES:
								r = read(f, &fi->create_time, 8);
								if (r != 8)
									fi->create_time = fi->modify_time = 0;
								break;
							case HFS_HDR_FINFO:
								r = read(f, fi->type, 8);
								if (r != 8)
									memset(fi->type, 0, 8);
								break;
						}
					}
					break;
			}
			close(f);
		}
	}

	if (!(*(u_int32_t *)(&fi->type[0])) || !(*(u_int32_t *)(&fi->type[4])))
		suffix_type_creator(fi->type, path);
	if (!fi->create_time || !fi->modify_time) {
		if (!stat(path, &sb)) {
			u_int32_t htime = hfs_u_to_htime(sb.st_mtime);
			if (!fi->create_time)
				fi->create_time = htime;
			if (!fi->modify_time)
				fi->modify_time = htime;
		}
	}
	if (!fi->comlen) {
		fi->comlen = strlen(hxd_cfg.files.comment);
		if (fi->comlen > 200)
			fi->comlen = 200;
		memcpy(fi->comment, hxd_cfg.files.comment, fi->comlen);
	}
}

static void
hfsinfo_write (const char *path, struct hfsinfo *fi)
{
	char infopath[MAXPATHLEN];
	int f, i, r;
	union {
		struct hfs_cap_info cap;
		struct hfs_dbl_hdr dbl;
	} hdr;

	if (!finderinfo_path(infopath, path, 0)) {
		f = open(infopath, O_RDWR|O_CREAT, hxd_cfg.permissions.files);
		if (f >= 0) {
			switch (hxd_cfg.files.fork) {
				case HFS_FORK_CAP:
					memset(&hdr.cap, 0, sizeof(struct hfs_cap_info));
					hdr.cap.fi_magic1 = HFS_CAP_MAGIC1;
					hdr.cap.fi_version = HFS_CAP_VERSION;
					hdr.cap.fi_magic = HFS_CAP_MAGIC;
					hdr.cap.fi_datemagic = HFS_CAP_DMAGIC;
					hdr.cap.fi_datevalid = HFS_CAP_MDATE|HFS_CAP_CDATE;
					memcpy(hdr.cap.fi_fndr, fi->type, 8);
					memcpy(hdr.cap.fi_ctime, &fi->create_time, 8);
					hdr.cap.fi_comln = fi->comlen > 200 ? 200 : fi->comlen;
					memcpy(hdr.cap.fi_comnt, fi->comment, hdr.cap.fi_comln);
					write(f, &hdr.cap, SIZEOF_HFS_CAP_INFO);
					break;
				case HFS_FORK_NETATALK:
				case HFS_FORK_DOUBLE:
					r = read(f, &hdr.dbl, SIZEOF_HFS_DBL_HDR);
					if (r != SIZEOF_HFS_DBL_HDR) {
						struct hfs_hdr_descr *descr;
						hdr.dbl.entries = htons(3);
						descr = (struct hfs_hdr_descr *)(&hdr.dbl.descrs[SIZEOF_HFS_HDR_DESCR * 0]);
						descr->id = htonl(HFS_HDR_COMNT);
						descr->offset = htonl(SIZEOF_HFS_DBL_HDR + SIZEOF_HFS_HDR_DESCR * 3);
						descr->length = htonl(fi->comlen);
						descr = (struct hfs_hdr_descr *)(&hdr.dbl.descrs[SIZEOF_HFS_HDR_DESCR * 1]);
						descr->id = htonl(HFS_HDR_DATES);
						descr->offset = htonl(SIZEOF_HFS_DBL_HDR + SIZEOF_HFS_HDR_DESCR * 3 + fi->comlen);
						descr->length = htonl(8);
						descr = (struct hfs_hdr_descr *)(&hdr.dbl.descrs[SIZEOF_HFS_HDR_DESCR * 2]);
						descr->id = htonl(HFS_HDR_FINFO);
						descr->offset = htonl(SIZEOF_HFS_DBL_HDR + SIZEOF_HFS_HDR_DESCR * 3 + fi->comlen + 8);
						descr->length = htonl(8);
					} else {
						if (ntohs(hdr.dbl.entries) > HFS_HDR_MAX)
							hdr.dbl.entries = htons(HFS_HDR_MAX);
						r = read(f, &hdr.dbl.descrs, SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries));
						if (r != SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries))
							break;
					}
					for (i = 0; i < ntohs(hdr.dbl.entries); i++) {
						struct hfs_hdr_descr *descr = (struct hfs_hdr_descr *)(&hdr.dbl.descrs[SIZEOF_HFS_HDR_DESCR * i]);
						if (lseek(f, ntohl(descr->offset), SEEK_SET) != (off_t)ntohl(descr->offset))
							continue;
						switch (ntohl(descr->id)) {
							case HFS_HDR_COMNT:
								if (r == SIZEOF_HFS_DBL_HDR)
									break;
								descr->length = htonl(fi->comlen);
								write(f, fi->comment, fi->comlen);
								break;
							case HFS_HDR_OLDI:
							case HFS_HDR_DATES:
								if (descr->length < 8)
									descr->length = 8;
								write(f, &fi->create_time, 8);
								break;
							case HFS_HDR_FINFO:
								if (descr->length < 8)
									descr->length = 8;
								write(f, fi->type, 8);
								break;
						}
					}
					hdr.dbl.magic = htonl(HFS_DBL_MAGIC);
					if (hxd_cfg.files.fork == HFS_FORK_NETATALK)
						hdr.dbl.version = htonl(HFS_HDR_VERSION_1);
					else
						hdr.dbl.version = htonl(HFS_HDR_VERSION_2);
					lseek(f, 0, SEEK_SET);
					write(f, &hdr.dbl, SIZEOF_HFS_DBL_HDR + SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries));
					break;
			}
			fsync(f);
			close(f);
		}
	}
}

static inline int
comment_len (const char *path)
{
	char infopath[MAXPATHLEN];
	int f, i, r, len = 0;
	struct stat sb;
	union {
		struct hfs_cap_info cap;
		struct hfs_dbl_hdr dbl;
	} hdr;

	if (!finderinfo_path(infopath, path, &sb)) {
		f = open(infopath, O_RDONLY);
		if (f >= 0) {
			switch (hxd_cfg.files.fork) {
				case HFS_FORK_CAP:
					if (read(f, &hdr.cap, SIZEOF_HFS_CAP_INFO) == SIZEOF_HFS_CAP_INFO)
						len = hdr.cap.fi_comln > 200 ? 200 : hdr.cap.fi_comln;
					break;
				case HFS_FORK_NETATALK:
				case HFS_FORK_DOUBLE:
					r = read(f, &hdr.dbl, SIZEOF_HFS_DBL_HDR);
					if (r != SIZEOF_HFS_DBL_HDR)
						break;
					if (ntohs(hdr.dbl.entries) > HFS_HDR_MAX)
						hdr.dbl.entries = htons(HFS_HDR_MAX);
					r = read(f, &hdr.dbl.descrs, SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries));
					if (r != SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries))
						break;
					for (i = 0; i < ntohs(hdr.dbl.entries); i++) {
						struct hfs_hdr_descr *descr = (struct hfs_hdr_descr *)(&hdr.dbl.descrs[SIZEOF_HFS_HDR_DESCR * i]);
						if (ntohl(descr->id) == HFS_HDR_COMNT)
							len = ntohl(descr->length) > 200 ? 200 : ntohl(descr->length);
					}
					break;
			}
			close(f);
		}
	}
	if (!len) {
		len = strlen(hxd_cfg.files.comment);
		if (len > 200)
			len = 200;
	}

	return len;
}

static inline void
comment_write (const char *path, char *comment, int comlen)
{
	char infopath[MAXPATHLEN];
	int f;
#if 0
	int i, r;
#endif
	union {
		struct hfs_cap_info cap;
		struct hfs_dbl_hdr dbl;
	} hdr;
	if (!finderinfo_path(infopath, path, 0)) {
		f = open(infopath, O_RDWR|O_CREAT, hxd_cfg.permissions.files);
		if (f >= 0) {
			if (comlen > 200)
				comlen = 200;
			switch (hxd_cfg.files.fork) {
				case HFS_FORK_CAP:
					if (read(f, &hdr.cap, SIZEOF_HFS_CAP_INFO) != SIZEOF_HFS_CAP_INFO) {
						memset(&hdr.cap, 0, sizeof(struct hfs_cap_info));
						hdr.cap.fi_magic1 = HFS_CAP_MAGIC1;
						hdr.cap.fi_version = HFS_CAP_VERSION;
						hdr.cap.fi_magic = HFS_CAP_MAGIC;
						hdr.cap.fi_datemagic = HFS_CAP_DMAGIC;
						suffix_type_creator(hdr.cap.fi_fndr, path);
					}
					hdr.cap.fi_comln = comlen;
					memcpy(hdr.cap.fi_comnt, comment, comlen);
					lseek(f, 0, SEEK_SET);
					write(f, &hdr.cap, SIZEOF_HFS_CAP_INFO);
					break;
#if 0
				case HFS_FORK_NETATALK:
				case HFS_FORK_DOUBLE:
					r = read(f, &hdr.dbl, SIZEOF_HFS_DBL_HDR);
					if (r != SIZEOF_HFS_DBL_HDR)
						break;
					if (ntohs(hdr.dbl.entries) > HFS_HDR_MAX)
						hdr.dbl.entries = htons(HFS_HDR_MAX);
					r = read(f, &hdr.dbl.descrs, SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries));
					if (r != SIZEOF_HFS_HDR_DESCR * ntohs(hdr.dbl.entries))
						break;
					for (i = 0; i < ntohs(hdr.dbl.entries); i++) {
						struct hfs_hdr_descr *descr = (struct hfs_hdr_descr *)(&hdr.dbl.descrs[SIZEOF_HFS_HDR_DESCR * i]);
						if (ntohl(descr->id) == HFS_HDR_COMNT) {
							descr->length = htonl(comlen);
							if (lseek(f, ntohl(descr->offset), SEEK_SET) != (off_t)ntohl(descr->offset))
								continue;
							write(f, comment, comlen);
							if (lseek(f, SIZEOF_HFS_DBL_HDR + SIZEOF_HFS_HDR_DESCR * i, SEEK_SET)
							    != (off_t)(SIZEOF_HFS_DBL_HDR + SIZEOF_HFS_HDR_DESCR * i))
								continue;
							write(f, descr, SIZEOF_HFS_HDR_DESCR);
						}
					}
					break;
#endif
			}
			fsync(f);
			close(f);
		}
	}
}
#endif /* CONFIG_HFS */

static void
snd_file_list (struct htlc_conn *htlc, struct hl_filelist_hdr **fhdrs, u_int16_t fhdrcount)
{
	struct hl_hdr h;
	u_int16_t i;
	u_int32_t pos, len = 2;

	h.type = htonl(HTLS_HDR_TASK);
	h.trans = htonl(htlc->trans);
	htlc->trans++;
	h.uid = 0;
	h.flag = 0;
	h.hc = htons(fhdrcount);

	for (i = 0; i < fhdrcount; i++)
		len += SIZEOF_HL_FILELIST_HDR + ntohl(fhdrs[i]->fnlen);
	h.len = h.len2 = htonl(len);
	pos = htlc->out.pos + htlc->out.len;
	qbuf_set(&htlc->out, htlc->out.pos, htlc->out.len + len + 20);
	memcpy(&htlc->out.buf[pos], &h, SIZEOF_HL_HDR);
	pos += SIZEOF_HL_HDR;
	FD_SET(htlc->fd, &hxd_wfds);
	for (i = 0; i < fhdrcount; i++) {
		memcpy(&htlc->out.buf[pos], fhdrs[i], SIZEOF_HL_FILELIST_HDR + ntohl(fhdrs[i]->fnlen));
		pos += SIZEOF_HL_FILELIST_HDR + ntohl(fhdrs[i]->fnlen);
	}
}

static int
fh_compare (const void *p1, const void *p2)
{
	struct hl_filelist_hdr *fh1 = *(struct hl_filelist_hdr **)p1, *fh2 = *(struct hl_filelist_hdr **)p2;
	int i, len1, len2, minlen;
	char *n1, *n2;

	n1 = fh1->fname;
	n2 = fh2->fname;
	len1 = ntohl(fh1->fnlen);
	len2 = ntohl(fh2->fnlen);
	minlen = len1 > len2 ? len2 : len1;
	for (i = 0; i < minlen; i++) {
		if (n1[i] > n2[i])
			return 1;
		else if (n1[i] != n2[i])
			return -1;
	}
	if (len1 > len2)
		return 1;
	else if (len1 != len2)
		return -1;

	return 0;
}

static inline void
fh_sort (struct hl_filelist_hdr **fhdrs, u_int16_t count)
{
#if WANT_TO_TIME_SORT
	static int use_qsort = 1;
	struct timeval start, now;
	double diff;
	char secsbuf[32];

	if (use_qsort) {
		gettimeofday(&start, 0);
		qsort(fhdrs, count, sizeof(struct hl_filelist_hdr *), fh_compare);
		gettimeofday(&now, 0);
	} else {
		u_int16_t i, j;
		struct hl_filelist_hdr *fhp;

		gettimeofday(&start, 0);
		for (i = 0; i < count; i++)
			for (j = 1; j < count - i; j++)
				if (fhdrs[j - 1]->fname[0] > fhdrs[j]->fname[0]) {
					fhp = fhdrs[j - 1];
					fhdrs[j - 1] = fhdrs[j];
					fhdrs[j] = fhp;
				}
		gettimeofday(&now, 0);
	}
	diff = (double)(now.tv_sec + now.tv_usec * 0.000001)
	     - (double)(start.tv_sec + start.tv_usec * 0.000001);
	sprintf(secsbuf, "%g", diff);
	hxd_log("%ssort %u took %s seconds", use_qsort ? "q" : "bubble ", count, secsbuf);
	use_qsort = !use_qsort;
#else
	qsort(fhdrs, count, sizeof(struct hl_filelist_hdr *), fh_compare);
#endif
}

static void
hxd_scandir (struct htlc_conn *htlc, const char *path)
{
	DIR *dir;
	struct dirent *de;
	struct hl_filelist_hdr **fhdrs = 0;
	u_int16_t count = 0, maxcount = 0;
	struct stat sb;
	char pathbuf[MAXPATHLEN];
	u_int8_t nlen;

	if (!(dir = opendir(path))) {
		snd_strerror(htlc, errno);
		return;
	}
	while ((de = readdir(dir))) {
		if (count >= maxcount) {
			maxcount += 16;
			fhdrs = xrealloc(fhdrs, sizeof(struct hl_filelist_hdr *) * maxcount);
		}
		if (de->d_name[0] == '.')
			continue;
		nlen = (u_int8_t)strlen(de->d_name);
		snprintf(pathbuf, sizeof(pathbuf), "%s/%s", path, de->d_name);
		fhdrs[count] = xmalloc(SIZEOF_HL_FILELIST_HDR + nlen);
		fhdrs[count]->type = htons(HTLS_DATA_FILE_LIST);
		fhdrs[count]->len = htons((SIZEOF_HL_FILELIST_HDR - 4) + nlen);
		fhdrs[count]->unknown = 0;
		fhdrs[count]->fnlen = htonl((u_int32_t)nlen);
		memcpy(fhdrs[count]->fname, de->d_name, nlen);
		if (stat(pathbuf, &sb)) {
			fhdrs[count]->fsize = 0;
			fhdrs[count]->ftype = htonl(0x616c6973);	/* 'alis' */
			fhdrs[count]->fcreator = 0;
			count++;
			continue;
		}
		if (S_ISDIR(sb.st_mode)) {
			u_int32_t ndirs = 0;
			DIR *subdir;
			struct dirent *subde;
			if ((subdir = opendir(pathbuf))) {
				while ((subde = readdir(subdir)))
					if (subde->d_name[0] != '.')
						ndirs++;
				closedir(subdir);
			}
			fhdrs[count]->fsize = htonl(ndirs);
			fhdrs[count]->ftype = htonl(0x666c6472);	/* 'fldr' */
			fhdrs[count]->fcreator = 0;
		} else {
			u_int32_t size = sb.st_size;
#if defined(CONFIG_HFS)
			size += resource_len(pathbuf);
#endif
			fhdrs[count]->fsize = htonl(size);
#if defined(CONFIG_HFS)
			type_creator((u_int8_t *)&fhdrs[count]->ftype, pathbuf);
#else
			fhdrs[count]->ftype = 0;
			fhdrs[count]->fcreator = 0;
#endif
		}
		count++;
	}
	closedir(dir);
	fh_sort(fhdrs, count);
	snd_file_list(htlc, fhdrs, count);
	while (count)
		xfree(fhdrs[--count]);
	xfree(fhdrs);
}

void
rcv_file_list (struct htlc_conn *htlc)
{
	char path[MAXPATHLEN];
	int err;

	if (htlc->in.pos == SIZEOF_HL_HDR) {
		hxd_scandir(htlc, ROOTDIR);
		return;
	}
	dh_start(htlc)
		if (ntohs(dh->type) != HTLC_DATA_DIR)
			return;
		if ((err = hldir_to_path(dh, ROOTDIR, path))) {
			snd_strerror(htlc, err);
			return;
		}
		if (!htlc->access.view_drop_boxes && strcasestr(path, "DROP BOX") && strcmp(path, htlc->dropbox)) {
			/* snd_strerror(htlc, EPERM); */
			hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
			return;
		}
		if (log_list)
			hxd_log("%s:%s:%u - list %s", htlc->name, htlc->login, htlc->uid, path);
		hxd_scandir(htlc, path);
	dh_end()
}

void
rcv_file_setinfo (struct htlc_conn *htlc)
{
	u_int16_t fnlen = 0, newfnlen = 0, comlen = 0;
	char dir[MAXPATHLEN], oldbuf[MAXPATHLEN], newbuf[MAXPATHLEN],
	     filename[NAME_MAX], newfilename[NAME_MAX];
	char rsrcpath_old[MAXPATHLEN], rsrcpath_new[MAXPATHLEN];
	u_int8_t comment[200];
	struct stat sb;
	int err;

	dir[0] = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_FILE_RENAME:
				newfnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(newfilename, dh->data, newfnlen);
				break;
			case HTLC_DATA_DIR:
				if ((err = hldir_to_path(dh, ROOTDIR, dir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
			case HTLC_DATA_FILE_COMMENT:
				comlen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
				memcpy(comment, dh->data, comlen);
				break;
		}
	dh_end()

	if (!fnlen || (!newfnlen && !comlen)) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}

	snprintf(oldbuf, sizeof(oldbuf), "%s/%s", dir[0] ? dir : ROOTDIR, filename);
	if (stat(oldbuf, &sb)) {
		snd_strerror(htlc, errno);
		return;
	}
	if (!htlc->access.view_drop_boxes && strcasestr(oldbuf, "DROP BOX") && strcmp(oldbuf, htlc->dropbox)) {
		snd_strerror(htlc, EPERM);
		return;
	}

#if defined(CONFIG_HFS) 
	if (comlen) {
		if (S_ISDIR(sb.st_mode)) {
			if (!htlc->access.comment_folders) {
				snd_strerror(htlc, EPERM);
				return;
			}
		} else {
			if (!htlc->access.comment_files) {
				snd_strerror(htlc, EPERM);
				return;
			}
		}
		if (log_comment)
			hxd_log("%s:%s:%u - comment %s to %.*s", htlc->name, htlc->login, htlc->uid, oldbuf, comlen, comment);
		comment_write(oldbuf, comment, comlen);
	}
#endif

	if (!newfnlen)
		goto ret;

	if (dir[0])
		snprintf(newbuf, sizeof(newbuf), "%s/%s", dir, newfilename);
	else
		snprintf(newbuf, sizeof(newbuf), "%s/%s", ROOTDIR, newfilename);

	if (S_ISDIR(sb.st_mode)) {
		if (!htlc->access.rename_folders) {
			snd_strerror(htlc, EPERM);
			return;
		}
	} else {
		if (!htlc->access.rename_files) {
			snd_strerror(htlc, EPERM);
			return;
		}
	}

	if (log_rename)
		hxd_log("%s:%s:%u - rename %s to %s", htlc->name, htlc->login, htlc->uid, oldbuf, newbuf);

	if (rename(oldbuf, newbuf)) {
		snd_strerror(htlc, errno);
		return;
	}

#if defined(CONFIG_HFS)
	if (hxd_cfg.files.fork == HFS_FORK_CAP) {
		if (!resource_path(rsrcpath_old, oldbuf, &sb)) {
			if ((err = resource_path(rsrcpath_new, newbuf, 0))) {
				/* (void)rename(newbuf, oldbuf); */
				snd_strerror(htlc, err);
				return;
			} else {
				if (rename(rsrcpath_old, rsrcpath_new)) {
					/* (void)rename(newbuf, oldbuf); */
					snd_strerror(htlc, errno);
					return;
				}
			}
		}
	}
	if (!finderinfo_path(rsrcpath_old, oldbuf, &sb)) {
		if ((err = finderinfo_path(rsrcpath_new, newbuf, 0))) {
			/* (void)rename(newbuf, oldbuf); */
			snd_strerror(htlc, err);
			return;
		} else {
			if (rename(rsrcpath_old, rsrcpath_new)) {
				/* (void)rename(newbuf, oldbuf); */
				snd_strerror(htlc, errno);
				return;
			}
		}
	}
#endif /* CONFIG_HFS */

ret:
	hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
}

int
copy_and_unlink (const char *oldpath, const char *newpath)
{
	int rfd, wfd, err;
	ssize_t r;
	char *buf;

	rfd = open(oldpath, O_RDONLY);
	if (rfd < 0)
		return rfd;
	wfd = open(newpath, O_WRONLY|O_CREAT|O_TRUNC, hxd_cfg.permissions.files);
	if (wfd < 0) {
		close(rfd);
		return wfd;
	}
	buf = xmalloc(COPY_BUFSIZE);
	for (;;) {
		r = read(rfd, buf, COPY_BUFSIZE);
		if (r == 0) {
			err = 0;
			break;
		}
		if (r < 0) {
			err = errno;
			break;
		}
		if (write(wfd, buf, r) != r) {
			err = errno;
			break;
		}
	}
	xfree(buf);

	close(rfd);
	close(wfd);

	if (err) {
		errno = err;
		return -1;
	} else
		return unlink(oldpath);
}

void
rcv_file_move (struct htlc_conn *htlc)
{
	u_int16_t fnlen = 0;
	char dir[MAXPATHLEN], newdir[MAXPATHLEN],
	     filename[NAME_MAX], oldbuf[MAXPATHLEN], newbuf[MAXPATHLEN];
	char rsrcpath_old[MAXPATHLEN], rsrcpath_new[MAXPATHLEN];
	struct stat sb, rsb;
	int err;
	dev_t diff_device;

	dir[0] = newdir[0] = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_DIR:
				if ((err = hldir_to_path(dh, ROOTDIR, dir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
			case HTLC_DATA_DIR_RENAME:
				if ((err = hldir_to_path(dh, ROOTDIR, newdir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
		}
	dh_end()

	if ((!dir[0] && !fnlen) || !newdir[0]) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}
	if (!dir[0])
		strcpy(dir, ROOTDIR);
	if (fnlen) {
		snprintf(oldbuf, sizeof(oldbuf), "%s/%s", dir, filename);
		snprintf(newbuf, sizeof(newbuf), "%s/%s", newdir, filename);
	} else {
		strcpy(oldbuf, dir);
		strcpy(newbuf, newdir);
	}

	if (stat(oldbuf, &sb)) {
		snd_strerror(htlc, errno);
		return;
	}
	if (S_ISDIR(sb.st_mode)) {
		if (!htlc->access.move_folders) {
			snd_strerror(htlc, EPERM);
			return;
		}
	} else {
		if (!htlc->access.move_files) {
			snd_strerror(htlc, EPERM);
			return;
		}
	}

	if (log_move)
		hxd_log("%s:%s:%u - move %s to %s", htlc->name, htlc->login, htlc->uid, oldbuf, newbuf);

	diff_device = sb.st_dev;
	if (stat(newdir, &sb)) {
		snd_strerror(htlc, errno);
		return;
	}

#if 0 /* gcc on hpux does not like this */
	diff_device &= ~sb.st_dev;
#else
	diff_device = diff_device != sb.st_dev;
#endif

	if (diff_device ? copy_and_unlink(oldbuf, newbuf) : rename(oldbuf, newbuf)) {
		snd_strerror(htlc, errno);
		return;
	}

#if defined(CONFIG_HFS)
	if (hxd_cfg.files.fork == HFS_FORK_CAP) {
		if (!resource_path(rsrcpath_old, oldbuf, &rsb)) {
			if ((err = resource_path(rsrcpath_new, newbuf, 0))) {
				/* (void)rename(newbuf, oldbuf); */
				snd_strerror(htlc, err);
				return;
			} else {
				if (diff_device ? copy_and_unlink(rsrcpath_old, rsrcpath_new) : rename(rsrcpath_old, rsrcpath_new)) {
					/* (void)rename(newbuf, oldbuf); */
					snd_strerror(htlc, errno);
					return;
				}
			}
		}
	}
	if (!finderinfo_path(rsrcpath_old, oldbuf, &rsb)) {
		if ((err = finderinfo_path(rsrcpath_new, newbuf, 0))) {
			/* (void)rename(newbuf, oldbuf); */
			snd_strerror(htlc, err);
			return;
		} else {
			if (diff_device ? copy_and_unlink(rsrcpath_old, rsrcpath_new) : rename(rsrcpath_old, rsrcpath_new)) {
				/* (void)rename(newbuf, oldbuf); */
				snd_strerror(htlc, errno);
				return;
			}
		}
	}
#endif /* CONFIG_HFS */

	hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
}

void
rcv_file_mkdir (struct htlc_conn *htlc)
{
	u_int16_t fnlen = 0;
	char dir[MAXPATHLEN], filename[NAME_MAX], newbuf[MAXPATHLEN];
	int err;

	dir[0] = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_DIR:
				err = hldir_to_path(dh, ROOTDIR, dir);
				if (err && err != ENOENT) {
					snd_strerror(htlc, err);
					return;
				}
				break;
		}
	dh_end()

	if (!fnlen && !dir[0]) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}
	if (dir[0]) {
		if (fnlen)
			snprintf(newbuf, sizeof(newbuf), "%s/%s", dir, filename);
		else
			strcpy(newbuf, dir);
	} else {
		snprintf(newbuf, sizeof(newbuf), "%s/%s", ROOTDIR, filename);
	}

	if (log_mkdir)
		hxd_log("%s:%s:%u - mkdir %s", htlc->name, htlc->login, htlc->uid, newbuf);

	if (mkdir(newbuf, hxd_cfg.permissions.directories))
		snd_strerror(htlc, errno);
	else
		hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
}

void
rcv_file_symlink (struct htlc_conn *htlc)
{
	u_int16_t fnlen = 0, newfnlen = 0;
	char dir[MAXPATHLEN], newdir[MAXPATHLEN],
	     filename[NAME_MAX], newfilename[NAME_MAX], oldbuf[MAXPATHLEN], newbuf[MAXPATHLEN];
	char rsrcpath_old[MAXPATHLEN], rsrcpath_new[MAXPATHLEN];
	struct stat rsb;
	int err;

	dir[0] = newdir[0] = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_FILE_RENAME:
				newfnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(newfilename, dh->data, newfnlen);
				break;
			case HTLC_DATA_DIR:
				if ((err = hldir_to_path(dh, ROOTDIR, dir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
			case HTLC_DATA_DIR_RENAME:
				if ((err = hldir_to_path(dh, ROOTDIR, newdir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
		}
	dh_end()

	if ((!dir[0] && !fnlen) || !newdir[0]) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}
	if (!dir[0])
		strcpy(dir, ROOTDIR);
	if (fnlen) {
		snprintf(oldbuf, sizeof(oldbuf), "%s/%s", dir, filename);
		snprintf(newbuf, sizeof(newbuf), "%s/%s", newdir, newfnlen ? newfilename : filename);
	} else {
		strcpy(oldbuf, dir);
		strcpy(newbuf, newdir);
	}
	if (!htlc->access.view_drop_boxes && strcasestr(oldbuf, "DROP BOX") && strcmp(oldbuf, htlc->dropbox)) {
		snd_strerror(htlc, EPERM);
		return;
	}

	if (log_symlink)
		hxd_log("%s:%s:%u - symlink %s to %s", htlc->name, htlc->login, htlc->uid, newbuf, oldbuf);

	if (symlink(oldbuf, newbuf))
		snd_strerror(htlc, errno);
	else
		hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
#if defined(CONFIG_HFS)
	if (hxd_cfg.files.fork == HFS_FORK_CAP) {
		if (!resource_path(rsrcpath_old, oldbuf, &rsb)) {
			if ((err = resource_path(rsrcpath_new, newbuf, 0))) {
				/* (void)unlink(newbuf); */
				snd_strerror(htlc, err);
				return;
			} else {
				if (symlink(rsrcpath_old, rsrcpath_new)) {
					/* (void)unlink(newbuf); */
					snd_strerror(htlc, errno);
					return;
				}
			}
		}
	}
	if (!finderinfo_path(rsrcpath_old, oldbuf, &rsb)) {
		if ((err = finderinfo_path(rsrcpath_new, newbuf, 0))) {
			/* (void)unlink(newbuf); */
			snd_strerror(htlc, err);
			return;
		} else {
			if (symlink(rsrcpath_old, rsrcpath_new)) {
				/* (void)unlink(newbuf); */
				snd_strerror(htlc, errno);
				return;
			}
		}
	}
#endif /* CONFIG_HFS */
}

static int
recursive_rmdir (char *dirpath)
{
	DIR *dir;
	struct dirent *de;
	char pathbuf[MAXPATHLEN];
	struct stat sb;
	int err;

	dir = opendir(dirpath);
	if (!dir)
		return 1;
	while ((de = readdir(dir))) {
		if (de->d_name[0] == '.' && ((de->d_name[1] == '.' && !de->d_name[2]) || !de->d_name[1]))
			continue;
		snprintf(pathbuf, MAXPATHLEN, "%s/%s", dirpath, de->d_name);
		if ((err = lstat(pathbuf, &sb)))
			goto ret;
		if (S_ISDIR(sb.st_mode) && !S_ISLNK(sb.st_mode))
			err = recursive_rmdir(pathbuf);
		else
			err = unlink(pathbuf);
		if (err)
			goto ret;
	}
	closedir(dir);
	return rmdir(dirpath);
ret:
	closedir(dir);

	return err;
}

void
rcv_file_delete (struct htlc_conn *htlc)
{
	u_int16_t fnlen = 0;
	char dir[MAXPATHLEN], filename[NAME_MAX], oldbuf[MAXPATHLEN];
	char rsrcpath_old[MAXPATHLEN];
	struct stat sb, rsb;
	int err;

	dir[0] = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_DIR:
				if ((err = hldir_to_path(dh, ROOTDIR, dir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
		}
	dh_end()

	if (!fnlen && !dir[0]) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}

	if (dir[0]) {
		if (fnlen)
			snprintf(oldbuf, sizeof(oldbuf), "%s/%s", dir, filename);
		else
			strcpy(oldbuf, dir);
	} else {
		snprintf(oldbuf, sizeof(oldbuf), "%s/%s", ROOTDIR, filename);
	}
	if (!htlc->access.view_drop_boxes && strcasestr(oldbuf, "DROP BOX") && strcmp(oldbuf, htlc->dropbox)) {
		snd_strerror(htlc, EPERM);
		return;
	}

	if (log_delete)
		hxd_log("%s:%s:%u - delete %s", htlc->name, htlc->login, htlc->uid, oldbuf);

	if (lstat(oldbuf, &sb)) {
		snd_strerror(htlc, errno);
		return;
	}

#if defined(CONFIG_HFS)
	if (hxd_cfg.files.fork == HFS_FORK_CAP) {
		if (!S_ISDIR(sb.st_mode) && !resource_path(rsrcpath_old, oldbuf, &rsb)
		    && !S_ISDIR(rsb.st_mode)) {
			if (unlink(rsrcpath_old)) {
				snd_strerror(htlc, errno);
				return;
			}
		}
	}
	if (!finderinfo_path(rsrcpath_old, oldbuf, &rsb)) {
		if (unlink(rsrcpath_old)) {
			snd_strerror(htlc, errno);
			return;
		}
	}
#endif /* CONFIG_HFS */

	if (S_ISDIR(sb.st_mode) && !S_ISLNK(sb.st_mode))
		err = recursive_rmdir(oldbuf);
	else
		err = unlink(oldbuf);

	if (err)
		snd_strerror(htlc, errno);
	else
		hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
}

void
rcv_file_getinfo (struct htlc_conn *htlc)
{
	u_int32_t size;
	u_int8_t date_create[8], date_modify[8];
	u_int16_t fnlen = 0;
	char dir[MAXPATHLEN], filename[NAME_MAX], oldbuf[MAXPATHLEN];
	struct stat sb;
	int err;
	u_int32_t mactime;
#if defined(CONFIG_HFS)
	struct hfsinfo fi;
#endif

	dir[0] = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) > NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_DIR:
				if ((err = hldir_to_path(dh, ROOTDIR, dir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
		}
	dh_end()

	if (!fnlen && !dir[0]) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}

	if (dir[0]) {
		if (fnlen)
			snprintf(oldbuf, sizeof(oldbuf), "%s/%s", dir, filename);
		else {
			int i, len = strlen(dir);

			for (i = len - 1; i > 0; i--)
				if (dir[i] == DIRCHAR) {
					fnlen = len - i > NAME_MAX ? NAME_MAX : len - i;
					strcpy(filename, &dir[len - fnlen + 1]);
					break;
				}
			strcpy(oldbuf, dir);
		}
	} else {
		snprintf(oldbuf, sizeof(oldbuf), "%s/%s", ROOTDIR, filename);
	}
	if (!htlc->access.view_drop_boxes && strcasestr(oldbuf, "DROP BOX") && strcmp(oldbuf, htlc->dropbox)) {
		snd_strerror(htlc, EPERM);
		return;
	}

#if 0
	{
		char *p;
		for (p = oldbuf+strlen(ROOTDIR); *p; p++)
			if ((*p != '.' && *p != '/') || (p-2 >= oldbuf && *p == '.' && *(p-1) == '.' && *(p-2) == '.'))
				goto ok;
		snd_strerror(htlc, EPERM);
		return;
	}

ok:
#endif
	if (log_getinfo)
		hxd_log("%s:%s:%u - getinfo %s", htlc->name, htlc->login, htlc->uid, oldbuf);

	if (stat(oldbuf, &sb)) {
		snd_strerror(htlc, errno);
		return;
	}
	size = sb.st_size;
#if defined(CONFIG_HFS)
	size += resource_len(oldbuf);
#endif
	size = htonl(size);
	memset(date_create, 0, 8);
	memset(date_modify, 0, 8);

#if defined(CONFIG_HFS)
	hfsinfo_read(oldbuf, &fi);
	mactime = hfs_h_to_mtime(fi.create_time);
	*((u_int16_t *)date_create) = htons(1904);
	*((u_int32_t *)(date_create+4)) = mactime;
	mactime = hfs_h_to_mtime(fi.modify_time);
	*((u_int16_t *)date_modify) = htons(1904);
	*((u_int32_t *)(date_modify+4)) = mactime;
	if (S_ISDIR(sb.st_mode))
		memcpy(fi.type, "fldr\0\0\0\0", 8);

	hlwrite(htlc, HTLS_HDR_TASK, 0, 8,
		HTLS_DATA_FILE_ICON, 4, fi.type,
		HTLS_DATA_FILE_TYPE, 4, fi.type,
		HTLS_DATA_FILE_CREATOR, 4, fi.creator,
		HTLS_DATA_FILE_SIZE, sizeof(size), &size,
		HTLS_DATA_FILE_NAME, fnlen, filename,
		HTLS_DATA_FILE_DATE_CREATE, 8, date_create,
		HTLS_DATA_FILE_DATE_MODIFY, 8, date_modify,
		HTLS_DATA_FILE_COMMENT, fi.comlen > 200 ? 200 : fi.comlen, fi.comment);
#else
	mactime = htonl(sb.st_mtime + 2082844800);
	*((u_int16_t *)date_modify) = 1904;
	*((u_int32_t *)(date_modify+4)) = mactime;
	hlwrite(htlc, HTLS_HDR_TASK, 0, 3,
		HTLS_DATA_FILE_SIZE, sizeof(size), &size,
		HTLS_DATA_FILE_NAME, fnlen, filename,
		HTLS_DATA_FILE_DATE_MODIFY, 8, date_modify);
#endif
}

#ifdef CONFIG_HOPE
void
rcv_file_hash (struct htlc_conn *htlc)
{
	u_int16_t fnlen = 0;
	char dir[MAXPATHLEN], filename[NAME_MAX], pathbuf[MAXPATHLEN];
	int err;
	int fd;
	u_int32_t data_len = 0, rsrc_len = 0;
	u_int16_t haval_len = 16, haval_passes = 3, hash_types = 0;
	u_int8_t md5[32], haval[64], sha1[40];
#if defined(CONFIG_HFS)
	int rfd;
	off_t off;
#endif

	dir[0] = 0;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) > NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_DIR:
				if ((err = hldir_to_path(dh, ROOTDIR, dir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
			case HTLC_DATA_RFLT:
				if (ntohs(dh->len) >= 66) {
					L32NTOH(data_len, &dh->data[46]);
					L32NTOH(rsrc_len, &dh->data[62]);
				}
				break;
			case HTLC_DATA_HASH_MD5:
				hash_types |= 0x01;
				break;
			case HTLC_DATA_HASH_HAVAL:
				hash_types |= 0x02;
				if (ntohs(dh->len) == 2) {
					haval_len = dh->data[0];
					haval_passes = dh->data[1];
				}
				if (haval_len > 32)
					haval_len = 32;
				if (haval_passes < 3)
					haval_passes = 3;
				if (haval_passes > 5)
					haval_passes = 5;
				break;
			case HTLC_DATA_HASH_SHA1:
				hash_types |= 0x04;
				break;
		}
	dh_end()

	if (!fnlen && !dir[0]) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}
	if (dir[0]) {
		if (fnlen)
			snprintf(pathbuf, sizeof(pathbuf), "%s/%s", dir, filename);
		else
			strcpy(pathbuf, dir);
	} else {
		snprintf(pathbuf, sizeof(pathbuf), "%s/%s", ROOTDIR, filename);
	}
	if (!htlc->access.view_drop_boxes && strcasestr(pathbuf, "DROP BOX") && strcmp(pathbuf, htlc->dropbox)) {
		snd_strerror(htlc, EPERM);
		return;
	}

	if (log_hash)
		hxd_log("%s:%s:%u - hash %s", htlc->name, htlc->login, htlc->uid, pathbuf);

	fd = open(pathbuf, O_RDONLY);
	if (fd < 0) {
		snd_strerror(htlc, errno);
		return;
	}
#if defined(CONFIG_HFS)
	rfd = resource_open(pathbuf, O_RDONLY, 0);
	off = lseek(rfd, 0, SEEK_CUR);
	if (off == (off_t)-1) {
		if (rfd >= 0)
			close(rfd);
		rfd = -1;
	}
#endif
	if (hash_types & 0x01) {
		memset(md5, 0, 32);
		md5_fd(fd, data_len, &md5[0]);
#if defined(CONFIG_HFS)
		if (rfd >= 0)
			md5_fd(rfd, rsrc_len, &md5[16]);
#endif
	}
	if (hash_types & 0x02) {
		memset(haval, 0, haval_len * 2);
		lseek(fd, 0, SEEK_SET);
		haval_fd(fd, data_len, &haval[0], haval_len * 8, haval_passes);
#if defined(CONFIG_HFS)
		if (rfd >= 0) {
			lseek(rfd, off, SEEK_SET);
			haval_fd(rfd, rsrc_len, &haval[haval_len], haval_len * 8, haval_passes);
		}
#endif
	}
	if (hash_types & 0x04) {
		memset(sha1, 0, 40);
		lseek(fd, 0, SEEK_SET);
		sha_fd(fd, data_len, &sha1[0]);
#if defined(CONFIG_HFS)
		if (rfd >= 0) {
			lseek(rfd, off, SEEK_SET);
			sha_fd(rfd, rsrc_len, &sha1[20]);
		}
#endif
	}
#if defined(CONFIG_HFS)
	if (rfd >= 0)
		close(rfd);
#endif
	close(fd);

#if defined(CONFIG_HFS)
	hlwrite(htlc, HTLS_HDR_TASK, 0, 3,
		HTLS_DATA_HASH_MD5, 32, md5,
		HTLS_DATA_HASH_HAVAL, haval_len * 2, haval,
		HTLS_DATA_HASH_SHA1, 40, sha1);
#else
	hlwrite(htlc, HTLS_HDR_TASK, 0, 3,
		HTLS_DATA_HASH_MD5, 16, md5,
		HTLS_DATA_HASH_HAVAL, haval_len, haval,
		HTLS_DATA_HASH_SHA1, 20, sha1);
#endif
}
#endif

#if defined(CONFIG_HTXF_PTHREAD) || defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
static int
make_htxf_sock (struct sockaddr_in *saddr)
{
	int s;
	int x;

	s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (s < 0)
		return s;
	x = 1;
	setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x));
#if defined(SO_LINGER)
	{
		struct linger lingivitis;

		lingivitis.l_onoff = 1;
		lingivitis.l_linger = 1000;
		setsockopt(s, SOL_SOCKET, SO_LINGER, (char *)&lingivitis, sizeof(lingivitis));
	}
#endif
	if (bind(s, (struct sockaddr *)saddr, sizeof(struct sockaddr_in))) {
		close(s);
		return -1;
	}
	if (listen(s, 1)) {
		close(s);
		return -1;
	}

	return s;
}

typedef ssize_t watch_time_t;

struct watch {
	struct timeval begin;
	/* 1/10^6 */
	watch_time_t time;
};

static struct watch only_watch;

static inline struct watch *
watch_new (void)
{
	return &only_watch;
}

static inline void
watch_reset (struct watch *wp)
{
	memset(wp, 0, sizeof(struct watch));
}

static inline void
watch_start (struct watch *wp)
{
	gettimeofday(&wp->begin, 0);
}

static inline void
watch_stop (struct watch *wp)
{
	struct timeval now;
	watch_time_t sec, usec;

	gettimeofday(&now, 0);
	if ((usec = now.tv_usec - wp->begin.tv_usec) < 0) {
		usec += 1000000;
		sec = now.tv_sec - wp->begin.tv_sec - 1;
	} else {
		sec = now.tv_sec - wp->begin.tv_sec;
	}
	wp->time = usec + (sec * 1000000);
}

static void
shape_out (ssize_t bytes, watch_time_t usecs, ssize_t max_Bps)
{
	float Bpus, max_Bpus = (float)(max_Bps) / 1000000;

	Bpus = (float)bytes / (float)usecs;
	if (Bpus > max_Bpus) {
		size_t st = (size_t)((Bpus / max_Bpus) * (float)usecs) - usecs;
		usleep(st);
	}
}

static RETSIGTYPE
apple_sucks_exit (int sig)
{
#if defined(CONFIG_HTXF_PTHREAD)
	pthread_exit((void *)sig);
#else
	_exit(sig);
#endif
}

#if defined(CONFIG_HTXF_PTHREAD)
typedef void * thread_return_type;
#else
typedef int thread_return_type;
#endif

static thread_return_type
get_thread (void *__arg)
{
	struct htxf_conn *htxf = (struct htxf_conn *)__arg;
	struct htlc_conn *htlc = htxf->htlc;
	char *path = htxf->path;
	u_int32_t data_size = htxf->data_size, data_pos = htxf->data_pos;
	u_int32_t rsrc_size = htxf->rsrc_size, rsrc_pos = htxf->rsrc_pos;
	u_int32_t ref = htxf->ref;
	u_int16_t preview = htxf->preview;
	u_int32_t addr = htxf->sockaddr.sin_addr.s_addr;

	int ls, s, f, r, x, retval = 0;
	struct sockaddr_in sa;
	u_int8_t *buf;
	char abuf[16];
	struct htxf_hdr hdr;
	struct sigaction act;
	struct watch *wp;

	ls = make_htxf_sock(&htxf->listen_sockaddr);

	act.sa_flags = 0;
	act.sa_handler = apple_sucks_exit;
	sigfillset(&act.sa_mask);
	sigaction(SIGALRM, &act, 0);
	alarm(10);
next:
	x = sizeof(sa);
	s = accept(ls, (struct sockaddr *)&sa, &x);
	if (s < 0) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		retval = errno;
		goto ret_nobuf;
	}
	inet_ntoa_r(sa.sin_addr, abuf, sizeof(abuf));
	hxd_log("htxf get %s from %s:%u (%s:%s); data_pos %u; rsrc_pos %u", path, abuf, ntohs(sa.sin_port), htlc->name, htlc->login, data_pos, rsrc_pos);
	if (addr != sa.sin_addr.s_addr) {
		hxd_log("WRONG ADDRESS");
		goto next;
	}
	if (read(s, &hdr, 16) != 16) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		goto next;
	}
	if (hdr.ref != ref) {
		hxd_log("wrong ref");
		goto next;
	}
	alarm(0);
	close(ls);

	buf = malloc(HTXF_BUFSIZE + 512);
	if (!buf) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		retval = errno;
		goto ret_nobuf;
	}
	if (!preview) {
#if defined(CONFIG_HFS)
		struct hfsinfo fi;

		hfsinfo_read(path, &fi);
#endif
		memcpy(buf, "\
FILP\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\2INFO\0\0\0\0\0\0\0\0\0\0\0^AMAC\
TYPECREA\
\0\0\0\0\0\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\
\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\7\160\0\0\0\0\0\0\7\160\0\0\0\0\0\0\0\0\0\3hxd", 115);
#if defined(CONFIG_HFS)
		memcpy(buf+44, fi.type, 8);
		buf[116] = fi.comlen;
		memcpy(buf+117, fi.comment, fi.comlen);
		if ((rsrc_size - rsrc_pos))
			buf[23] = 3;
		if (65 + fi.comlen + 12 > 0xff)
			buf[38] = 1;
		buf[39] = 65 + fi.comlen + 12;

		/* *((u_int16_t *)(&buf[92])) = htons(1904); */
		*((u_int32_t *)(&buf[96])) = hfs_h_to_mtime(fi.create_time);
		/* *((u_int16_t *)(&buf[100])) = htons(1904); */
		*((u_int32_t *)(&buf[104])) = hfs_h_to_mtime(fi.modify_time);
		memcpy(&buf[117] + fi.comlen, "DATA\0\0\0\0\0\0\0\0", 12);
		S32HTON((data_size - data_pos), &buf[129 + fi.comlen]);
		if (write(s, buf, 133 + fi.comlen) != (ssize_t)(133 + fi.comlen)) {
			hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
			goto ret_err;
		}
#else
		buf[39] = 65 + 12;
		memcpy(&buf[117], "DATA\0\0\0\0\0\0\0\0", 12);
		S32HTON((data_size - data_pos), &buf[129]);
		if (write(s, buf, 133) != (ssize_t)(133)) {
			hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
			goto ret_err;
		}
#endif /* CONFIG_HFS */
	}

	if (htlc->limit_out_Bps)
		wp = watch_new();
	else
		wp = 0;

	if ((data_size - data_pos)) {
		f = open(path, O_RDONLY);
		if (f < 0) {
			hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
			goto ret_err;
		}
		if (data_pos)
			if (lseek(f, data_pos, SEEK_SET) != (off_t)data_pos) {
				hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
				goto ret_err;
			}
		for (;;) {
			r = read(f, buf, HTXF_BUFSIZE > (data_size - data_pos) ? (data_size - data_pos) : HTXF_BUFSIZE);
			if (r <= 0) {
				hxd_log("%s:%d: r == %d; %s (%d)", __FILE__, __LINE__, r, strerror(errno), errno);
				goto ret_err;
			}
			if (wp) {
				watch_reset(wp);
				watch_start(wp);
			}
			if (write(s, buf, r) != r) {
				hxd_log("%s:%d: r == %d; %s (%d)", __FILE__, __LINE__, r, strerror(errno), errno);
				goto ret_err;
			}
			if (wp)
				watch_stop(wp);
			data_pos += r;
			htxf->data_pos = data_pos;
			if (data_pos >= data_size)
				break;
			if (wp)
				shape_out(r, wp->time, htlc->limit_out_Bps);
		}
		close(f);
	}
#if defined(CONFIG_HFS)
	if (preview || !(rsrc_size - rsrc_pos))
		goto done;
	memcpy(buf, "MACR\0\0\0\0\0\0\0\0", 12);
	S32HTON((rsrc_size - rsrc_pos), &buf[12]);
	if (write(s, buf, 16) != 16) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		goto ret_err;
	}
	if ((rsrc_size - rsrc_pos)) {
		f = resource_open(path, O_RDONLY, 0);
		if (f < 0) {
			hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
			goto ret_err;
		}
		if (rsrc_pos)
			if (lseek(f, rsrc_pos, SEEK_CUR) == (off_t)-1) {
				hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
				goto ret_err;
			}
		for (;;) {
			r = read(f, buf, HTXF_BUFSIZE > (rsrc_size - rsrc_pos) ? (rsrc_size - rsrc_pos) : HTXF_BUFSIZE);
			if (r <= 0) {
				hxd_log("%s:%d: r == %d; %s (%d)", __FILE__, __LINE__, r, strerror(errno), errno);
				goto ret_err;
			}
			if (wp) {
				watch_reset(wp);
				watch_start(wp);
			}
			if (write(s, buf, r) != r) {
				hxd_log("%s:%d: r == %d; %s (%d)", __FILE__, __LINE__, r, strerror(errno), errno);
				goto ret_err;
			}
			if (wp)
				watch_stop(wp);
			rsrc_pos += r;
			htxf->rsrc_pos = rsrc_pos;
			if (rsrc_pos >= rsrc_size)
				break;
			if (wp)
				shape_out(r, wp->time, htlc->limit_out_Bps);
		}
		close(f);
	}
done:
#endif /* CONFIG_HFS */
#ifdef CONFIG_WINDOS_CLIENT_FIX
	fd_blocking(s, 0);
	r = 0;
	while (read(s, buf, HTXF_BUFSIZE) && errno == EWOULDBLOCK) {
		if (r++ > 3)
			break;
		sleep(1);
	}
#endif
	retval = 0;
	goto ret;
ret_err:
	retval = errno;
ret:
	close(s);
	free(buf);
ret_nobuf:
	hxd_log("closed htxf get %s from %s:%u (%s:%s); data_pos %u; rsrc_pos %u", path, abuf, ntohs(sa.sin_port), htlc->name, htlc->login, data_pos, rsrc_pos);

#if defined(CONFIG_HTXF_PTHREAD)
	{
		u_int16_t i;

		LOCK_HTXF(htlc);
		for (i = 0; i < HTXF_GET_MAX; i++)
			if (htlc->htxf_out[i] == htxf)
				goto ok;
		goto notok;
ok:
		xfree(htlc->htxf_out[i]);
		htlc->htxf_out[i] = 0;
		htlc->nr_gets--;
		nr_gets--;
notok:
		UNLOCK_HTXF(htlc);
	}
#endif

	return (thread_return_type)retval;
}

static thread_return_type
put_thread (void *__arg)
{
	struct htxf_conn *htxf = (struct htxf_conn *)__arg;
	struct htlc_conn *htlc = htxf->htlc;
	char *path = htxf->path;
	u_int32_t data_size = htxf->data_size, data_pos = htxf->data_pos;
	u_int32_t rsrc_size = htxf->rsrc_size, rsrc_pos = htxf->rsrc_pos;
	u_int32_t ref = htxf->ref;
	u_int32_t addr = htxf->sockaddr.sin_addr.s_addr;

	int ls, s, f, r, x, retval;
	struct sockaddr_in sa;
	u_int8_t *buf;
	char abuf[16], typecrea[8];
	struct htxf_hdr hdr;
	struct sigaction act;
	u_int32_t tot_pos, tot_len;
#if defined(CONFIG_HFS)
	struct hfsinfo fi;
#endif

	ls = make_htxf_sock(&htxf->listen_sockaddr);

	act.sa_flags = 0;
	act.sa_handler = apple_sucks_exit;
	sigfillset(&act.sa_mask);
	sigaction(SIGALRM, &act, 0);
	alarm(10);
next:
	x = sizeof(sa);
	s = accept(ls, (struct sockaddr *)&sa, &x);
	if (s < 0) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		retval = errno;
		goto ret_nobuf;
	}
	inet_ntoa_r(sa.sin_addr, abuf, sizeof(abuf));
	hxd_log("htxf put %s from %s:%u (%s:%s); data_pos %u; rsrc_pos %u", path, abuf, ntohs(sa.sin_port), htlc->name, htlc->login, data_pos, rsrc_pos);
	if (addr != sa.sin_addr.s_addr) {
		hxd_log("WRONG ADDRESS");
		goto next;
	}
	if (read(s, &hdr, 16) != 16) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		goto next;
	}
	if (hdr.ref != ref) {
		hxd_log("wrong ref");
		goto next;
	}
	alarm(0);
	close(ls);

	tot_len = htonl(hdr.len);
	buf = malloc(HTXF_BUFSIZE + 1024);
	if (!buf) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		retval = errno;
		goto ret_nobuf;
	}
	if (read(s, buf, 40) != 40) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		goto ret_err;
	}
	r = (buf[38] ? 0x100 : 0) + buf[39] + 16;
	if (read(s, buf, r) != r) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		goto ret_err;
	}
	L32NTOH(data_size, &buf[r - 4]);
	data_size += data_pos;
	htxf->data_size = data_size;
	tot_pos = 40 + r;
	memcpy(typecrea, &buf[4], 8);

#if defined(CONFIG_HFS)
	memset(&fi, 0, sizeof(struct hfsinfo));
	fi.comlen = buf[73 + buf[71]];
	memcpy(fi.type, "HTftHTLC", 8);
	memcpy(fi.comment, &buf[74 + buf[71]], fi.comlen);
	fi.create_time = hfs_m_to_htime(*((u_int32_t *)(&buf[56])));
	fi.modify_time = hfs_m_to_htime(*((u_int32_t *)(&buf[64])));
	hfsinfo_write(path, &fi);
#endif

	f = open(path, O_WRONLY|O_CREAT, hxd_cfg.permissions.files);
	if (f < 0) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		goto ret_err;
	}
	if (data_size > data_pos) {
		if (fd_lock_write(f) == -1) {
			hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
			goto ret_err;
		}
		if (data_pos)
			if (lseek(f, data_pos, SEEK_SET) != (off_t)data_pos) {
				hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
				goto ret_err;
			}
		for (;;) {
			r = read(s, buf, HTXF_BUFSIZE > (data_size - data_pos) ? (data_size - data_pos) : HTXF_BUFSIZE);
			if (r <= 0) {
				hxd_log("%s:%d: r == %d; %s (%d)", __FILE__, __LINE__, r, strerror(errno), errno);
				goto ret_err;
			}
			if (write(f, buf, r) != r) {
				hxd_log("%s:%d: r == %d; %s (%d)", __FILE__, __LINE__, r, strerror(errno), errno);
				goto ret_err;
			}
			data_pos += r;
			tot_pos += r;
			htxf->data_pos = data_pos;
			if (data_pos >= data_size)
				break;
		}
		fsync(f);
	}
	close(f);
#if defined(CONFIG_HFS)
	if (tot_pos == tot_len)
		goto done;
	if (read(s, buf, 16) != 16) {
		hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
		goto ret_err;
	}
	L32NTOH(rsrc_size, &buf[12]);
	rsrc_size += rsrc_pos;
	htxf->rsrc_size = rsrc_size;
	if (rsrc_size > rsrc_pos) {
		f = resource_open(path, O_WRONLY|O_CREAT, hxd_cfg.permissions.files);
		if (f < 0) {
			hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
			goto ret_err;
		}
		if (fd_lock_write(f) == -1) {
			hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
			goto ret_err;
		}
		if (rsrc_pos)
			if (lseek(f, rsrc_pos, SEEK_CUR) == -1) {
				hxd_log("%s:%d: %s (%d)", __FILE__, __LINE__, strerror(errno), errno);
				goto ret_err;
			}
		tot_pos += (rsrc_size - rsrc_pos);
		for (;;) {
			r = read(s, buf, HTXF_BUFSIZE > (rsrc_size - rsrc_pos) ? (rsrc_size - rsrc_pos) : HTXF_BUFSIZE);
			if (r <= 0) {
				hxd_log("%s:%d: r == %d; %s (%d)", __FILE__, __LINE__, r, strerror(errno), errno);
				goto ret_err;
			}
			if (write(f, buf, r) != r) {
				hxd_log("%s:%d: r == %d; %s (%d)", __FILE__, __LINE__, r, strerror(errno), errno);
				goto ret_err;
			}
			rsrc_pos += r;
			htxf->rsrc_pos = rsrc_pos;
			if (rsrc_pos >= rsrc_size)
				break;
		}
		fsync(f);
		close(f);
	}
done:
	memcpy(fi.type, typecrea, 8);
	hfsinfo_write(path, &fi);
#endif /* CONFIG_HFS */
#ifdef CONFIG_WINDOS_CLIENT_FIX
	fd_blocking(s, 0);
	r = 0;
	while (read(s, buf, HTXF_BUFSIZE > (tot_len - tot_pos) ? (tot_len - tot_pos) : HTXF_BUFSIZE) && errno == EWOULDBLOCK) {
		if (r++ > 3)
			break;
		sleep(1);
	}
#endif
	retval = 0;
	goto ret;
ret_err:
	retval = errno;
ret:
	close(s);
	free(buf);
ret_nobuf:
	hxd_log("closed htxf put %s from %s:%u (%s:%s); data_pos %u; rsrc_pos %u", path, abuf, ntohs(sa.sin_port), htlc->name, htlc->login, data_pos, rsrc_pos);

#if defined(CONFIG_HTXF_PTHREAD)
	{
		u_int16_t i;

		LOCK_HTXF(htlc);
		for (i = 0; i < HTXF_PUT_MAX; i++)
			if (htlc->htxf_in[i] == htxf)
				goto ok;
		goto notok;
ok:
		xfree(htlc->htxf_in[i]);
		htlc->htxf_in[i] = 0;
		htlc->nr_puts--;
		nr_puts--;
notok:
		UNLOCK_HTXF(htlc);
	}
#endif

	return (thread_return_type)retval;
}
#endif /* CONFIG_HTXF_PTHREAD || CONFIG_HTXF_CLONE || CONFIG_HTXF_FORK */

#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
static void
mask_signal (int how, int sig)
{
	sigset_t set;

	sigemptyset(&set);
	sigaddset(&set, sig);
	sigprocmask(how, &set, 0);
}
#endif

void
rcv_file_get (struct htlc_conn *htlc)
{
	u_int16_t fnlen = 0, preview = 0;
	char path[MAXPATHLEN], dir[MAXPATHLEN], filename[NAME_MAX];
	char abuf[16];
	struct stat sb;
	u_int32_t size = 0, data_size = 0, rsrc_size = 0, ref,
		  data_pos = 0, rsrc_pos = 0;
	int err;
	int siz;
	struct htxf_conn *htxf;
	u_int16_t i;
#if defined(CONFIG_HTXF_PTHREAD) || defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
#ifdef CONFIG_HTXF_PTHREAD
	pthread_t tid;
#else
	pid_t pid;
#endif
#ifdef CONFIG_HTXF_CLONE
	void *stack;
#endif
#endif

	dir[0] = 0;

	if (htlc->nr_gets >= htlc->get_limit) {
		snprintf(abuf, 13, "%u at a time", htlc->get_limit);
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 13, abuf);
		return;
	}
	if (nr_gets >= hxd_cfg.limits.total_downloads) {
		char buf[128];
		int len = snprintf(buf, sizeof(buf), "maximum number of total downloads reached (%u >= %d)",
				   nr_gets, hxd_cfg.limits.total_downloads);
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, len, buf);
		return;
	}
	for (i = 0; i < HTXF_GET_MAX; i++)
		if (!htlc->htxf_out[i])
			break;
	if (i == HTXF_GET_MAX) {
		snd_strerror(htlc, EAGAIN);
		return;
	}

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_DIR:
				if ((err = hldir_to_path(dh, ROOTDIR, dir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
			case HTLC_DATA_RFLT:
				if (ntohs(dh->len) >= 50)
					L32NTOH(data_pos, &dh->data[46]);
				if (ntohs(dh->len) >= 66)
					L32NTOH(rsrc_pos, &dh->data[62]);
				break;
			case HTLC_DATA_FILE_PREVIEW:
				if (ntohs(dh->len) >= 2)
					preview = ntohs(*((u_int16_t *)dh->data));
				break;
		}
	dh_end()
	if (!fnlen && !dir[0]) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}

	if (dir[0]) {
		if (fnlen)
			snprintf(path, sizeof(path), "%s/%s", dir, filename);
		else
			strcpy(path, dir);
	} else {
		snprintf(path, sizeof(path), "%s/%s", ROOTDIR, filename);
	}
	if (!htlc->access.view_drop_boxes && strcasestr(path, "DROP BOX") && strcmp(path, htlc->dropbox)) {
		snd_strerror(htlc, EPERM);
		return;
	}

#ifdef CONFIG_HTXF_PREVIEW
	if (preview) {
		Image *img, *mimg;
		ImageInfo ii;

#if MaxTextExtent < MAXPATHLEN
		if (strlen(path) >= sizeof(ii.filename)) {
			hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 13, "path too long");
			return;
		}
#endif
		GetImageInfo(&ii);
		strcpy(ii.filename, path);
		img = ReadImage(&ii);
		if (!img)
			goto text_preview;
		mimg = MinifyImage(img);
		DestroyImage(img);
		if (!mimg) {
			hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 18, "MinifyImage failed");
			return;
		}
		if ((err = preview_path(path, path, 0))) {
			snd_strerror(htlc, err);
			DestroyImage(mimg);
			return;
		}
		strcpy(mimg->filename, path);
		data_pos = 0;
		rsrc_pos = 0;
		WriteImage(&ii, mimg);
		DestroyImage(mimg);
	}

text_preview:
#endif
	if (stat(path, &sb)) {
		snd_strerror(htlc, errno);
		return;
	}

	if (S_ISDIR(sb.st_mode)) {
		snd_strerror(htlc, EISDIR);
		return;
	}

	data_size = sb.st_size;
#if defined(CONFIG_HFS)
	rsrc_size = resource_len(path);
	size = (data_size - data_pos) + (rsrc_size - rsrc_pos);
	if (!preview)
		size += 133 + ((rsrc_size - rsrc_pos) ?  16 : 0) + comment_len(path);
#else
	size = (data_size - data_pos) + (preview ? 0 : 133);
#endif
	size = htonl(size);

	ref = htlc->sockaddr.sin_addr.s_addr;
	ref *= ref;
	ref *= (htlc->sockaddr.sin_port * htlc->trans);
	ref *= ref;
	ref = htonl(ref);

	htxf = xmalloc(sizeof(struct htxf_conn));
	htxf->data_size = data_size;
	htxf->rsrc_size = rsrc_size;
	htxf->data_pos = data_pos;
	htxf->rsrc_pos = rsrc_pos;
	htxf->ref = ref;
	htxf->preview = preview;
	htxf->sockaddr = htlc->sockaddr;
#if defined(CONFIG_HTXF_PTHREAD) || defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	siz = sizeof(struct sockaddr_in);
	if (getsockname(htlc->fd, (struct sockaddr *)&htxf->listen_sockaddr, &siz)) {
		hxd_log("rcv_file_get: getsockname: %s", strerror(errno));
		snd_strerror(htlc, errno);
		xfree(htxf);
		return;
	}
	htxf->listen_sockaddr.sin_port = htons(ntohs(htxf->listen_sockaddr.sin_port) + 1);
#endif
	strcpy(htxf->path, path);
	htxf->htlc = htlc;
#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	mask_signal(SIG_BLOCK, SIGCHLD);
#endif
#if defined(CONFIG_HTXF_PTHREAD)
	err = pthread_create(&tid, 0, get_thread, htxf);
#elif defined(CONFIG_HTXF_CLONE)
	htxf->stack = xmalloc(CLONE_STACKSIZE);
	stack = (void *)(((u_int8_t *)htxf->stack + CLONE_STACKSIZE) - sizeof(void *));
	pid = clone(get_thread, stack, CLONE_VM|CLONE_FS|SIGCHLD, htxf);
#elif defined(CONFIG_HTXF_FORK)
	pid = fork();
	if (pid == 0)
		_exit(get_thread(htxf));
#endif
#if defined(CONFIG_HTXF_PTHREAD)
	if (err) {
		xfree(htxf);
		snd_strerror(htlc, err);
	} else {
		htxf->tid = tid;
		pthread_detach(tid);
#elif defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	if (pid == -1) {
		xfree(htxf);
		snd_strerror(htlc, errno);
	} else {
		htxf->pid = pid;
#endif
		LOCK_HTXF(htlc);
		for (i = 0; i < HTXF_GET_MAX; i++)
			if (!htlc->htxf_out[i]) {
				htlc->htxf_out[i] = htxf;
				break;
			}
		UNLOCK_HTXF(htlc);
		htlc->nr_gets++;
		nr_gets++;
		if (log_download) {
			inet_ntoa_r(htlc->sockaddr.sin_addr, abuf, sizeof(abuf));
			hxd_log("%s@%s:%u - %s:%u:%u:%s - download %s:%08x", htlc->userid, abuf, ntohs(htlc->sockaddr.sin_port),
				htlc->name, htlc->icon, htlc->uid, htlc->login, htxf->path, htxf->ref);
		}
		hlwrite(htlc, HTLS_HDR_TASK, 0, 2,
			HTLS_DATA_HTXF_REF, sizeof(ref), &ref,
			HTLS_DATA_HTXF_SIZE, sizeof(size), &size);
#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	}
	mask_signal(SIG_UNBLOCK, SIGCHLD);
#endif
}

void
rcv_file_put (struct htlc_conn *htlc)
{
	u_int16_t fnlen = 0, resume = 0;
	char path[MAXPATHLEN], dir[MAXPATHLEN], filename[NAME_MAX];
	char abuf[16];
	struct stat sb;
	int err;
	int siz;
	u_int32_t ref, data_pos = 0, rsrc_pos = 0;
	u_int8_t rflt[74];
	struct htxf_conn *htxf;
	u_int16_t i;
#if defined(CONFIG_HTXF_PTHREAD) || defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
#ifdef CONFIG_HTXF_PTHREAD
	pthread_t tid;
#else
	pid_t pid;
#endif
#ifdef CONFIG_HTXF_CLONE
	void *stack;
#endif
#endif

	dir[0] = 0;

	if (htlc->nr_puts >= htlc->put_limit) {
		snprintf(abuf, 13, "%u at a time", htlc->put_limit);
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 13, abuf);
		return;
	}
	if (nr_puts >= hxd_cfg.limits.total_uploads) {
		char buf[128];
		int len = snprintf(buf, sizeof(buf), "maximum number of total uploads reached (%u >= %d)",
				   nr_gets, hxd_cfg.limits.total_uploads);
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, len, buf);
		return;
	}
	for (i = 0; i < HTXF_PUT_MAX; i++)
		if (!htlc->htxf_in[i])
			break;
	if (i == HTXF_PUT_MAX) {
		snd_strerror(htlc, EAGAIN);
		return;
	}

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_FILE_NAME:
				fnlen = ntohs(dh->len) >= NAME_MAX ? NAME_MAX - 1 : ntohs(dh->len);
				read_filename(filename, dh->data, fnlen);
				break;
			case HTLC_DATA_DIR:
				if ((err = hldir_to_path(dh, ROOTDIR, dir))) {
					snd_strerror(htlc, err);
					return;
				}
				break;
			case HTLC_DATA_FILE_PREVIEW:
				if (ntohs(dh->len) >= 2)
					resume = htons(*((u_int16_t *)dh->data));
				break;
		}
	dh_end()

	if (!htlc->access.upload_anywhere && (!dir[0] || (!strcasestr(dir, "UPLOAD") && !strcasestr(dir, "DROP BOX")))) {
		snd_strerror(htlc, EPERM);
		return;
	}
	if (!fnlen && !dir[0]) {
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
		return;
	}

	if (dir[0]) {
		if (fnlen)
			snprintf(path, sizeof(path), "%s/%s", dir, filename);
		else
			strcpy(path, dir);
	} else {
		snprintf(path, sizeof(path), "%s/%s", ROOTDIR, filename);
	}

	if (!resume) {
		if (!stat(path, &sb)) {
			snd_strerror(htlc, EEXIST);
			return;
		}
		if (errno != ENOENT) {
			snd_strerror(htlc, errno);
			return;
		}
	} else {
		if (stat(path, &sb)) {
			snd_strerror(htlc, errno);
			return;
		}
		data_pos = sb.st_size;
#if defined(CONFIG_HFS)
		rsrc_pos = resource_len(path);
#endif
		memcpy(rflt, "RFLT\0\1\
\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\0\0\0\2DATA\0\0\0\0\0\0\0\0\0\0\0\0MACR\0\0\0\0\0\0\0\0\0\0\0\0", 74);
		S32HTON(data_pos, &rflt[46]);
		S32HTON(rsrc_pos, &rflt[62]);
	}

	ref = htlc->sockaddr.sin_addr.s_addr;
	ref *= ref;
	ref *= (htlc->sockaddr.sin_port * htlc->trans);
	ref *= ref;
	ref = htonl(ref);

	htxf = xmalloc(sizeof(struct htxf_conn));
	htxf->data_pos = data_pos;
	htxf->rsrc_pos = rsrc_pos;
	htxf->ref = ref;
	htxf->sockaddr = htlc->sockaddr;
#if defined(CONFIG_HTXF_PTHREAD) || defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	siz = sizeof(struct sockaddr_in);
	if (getsockname(htlc->fd, (struct sockaddr *)&htxf->listen_sockaddr, &siz)) {
		hxd_log("rcv_file_get: getsockname: %s", strerror(errno));
		snd_strerror(htlc, errno);
		xfree(htxf);
		return;
	}
	htxf->listen_sockaddr.sin_port = htons(ntohs(htxf->listen_sockaddr.sin_port) + 1);
#endif
	strcpy(htxf->path, path);
	htxf->htlc = htlc;
#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	mask_signal(SIG_BLOCK, SIGCHLD);
#endif
#if defined(CONFIG_HTXF_PTHREAD)
	err = pthread_create(&tid, 0, put_thread, htxf);
#elif defined(CONFIG_HTXF_CLONE)
	htxf->stack = xmalloc(CLONE_STACKSIZE);
	stack = (void *)(((u_int8_t *)htxf->stack + CLONE_STACKSIZE) - sizeof(void *));
	pid = clone(put_thread, stack, CLONE_VM|CLONE_FS|SIGCHLD, htxf);
#elif defined(CONFIG_HTXF_FORK)
	pid = fork();
	if (pid == 0)
		_exit(put_thread(htxf));
#endif
#if defined(CONFIG_HTXF_PTHREAD)
	if (err) {
		xfree(htxf);
		snd_strerror(htlc, err);
	} else {
		htxf->tid = tid;
		pthread_detach(tid);
#elif defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	if (pid == -1) {
		xfree(htxf);
		snd_strerror(htlc, errno);
	} else {
		htxf->pid = pid;
#endif
		LOCK_HTXF(htlc);
		for (i = 0; i < HTXF_PUT_MAX; i++)
			if (!htlc->htxf_in[i]) {
				htlc->htxf_in[i] = htxf;
				break;
			}
		UNLOCK_HTXF(htlc);
		htlc->nr_puts++;
		nr_puts++;
		if (log_upload) {
			inet_ntoa_r(htlc->sockaddr.sin_addr, abuf, sizeof(abuf));
			hxd_log("%s@%s:%u - %s:%u:%u:%s - upload %s:%08x", htlc->userid, abuf, ntohs(htlc->sockaddr.sin_port),
				htlc->name, htlc->icon, htlc->uid, htlc->login, htxf->path, htxf->ref);
		}
		if (!resume)
			hlwrite(htlc, HTLS_HDR_TASK, 0, 1,
				HTLS_DATA_HTXF_REF, sizeof(ref), &ref);
		else
			hlwrite(htlc, HTLS_HDR_TASK, 0, 2,
				HTLS_DATA_RFLT, 74, rflt,
				HTLS_DATA_HTXF_REF, sizeof(ref), &ref);
#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	}
	mask_signal(SIG_UNBLOCK, SIGCHLD);
#endif
}
