#include <sys/types.h>
#include <sys/param.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <pwd.h>
#include <locale.h>
#include "hxd.h"
#include "hx.h"
#include "getopt.h"
#include "xmalloc.h"

char last_msg_nick[32];

static char homedir[MAXPATHLEN];

void
expand_tilde (char *buf, char *str)
{
	char *p = str;
	int len;

	if (*p == '~') {
		p++;
		if (*p == '/' || !*p) {
			len = strlen(homedir);
			strcpy(&(buf[len]), p);
			memcpy(buf, homedir, len);
		} else {
			char *rest;
			struct passwd *pwe;

			if ((rest = strchr(p, '/')))
				*rest++ = 0;
			if ((pwe = getpwnam(p))) {
				len = strlen(pwe->pw_dir);
				if (rest) {
					strcpy(&(buf[len + 1]), rest);
					buf[len] = '/';
				}
				memcpy(buf, pwe->pw_dir, len);
			}
		}
	} else
		strcpy(buf, str);
}

void
hotline_client_input (char *str)
{
	if (!str)
		str = rl_line_buffer;

	if (*str) {
		/* history_expand(str, &str); */
		add_history(str);
		if (*str == '/' && *++str && *str != '/')
			hx_command(str);
		else
			hx_send_chat(str);
		/* free(str); */
	}
	rl_point = 0;
	rl_end = 0;
	hx_output.clear_input();
}

void
chrexpand (char *str, int len)
{
	char *p;
	int off;

	p = str;
	if (*p != '\\')
		return;
	off = 1;
	switch (p[1]) {
		case 'r':
			p[1] = '\r';
			break;
		case 'n':
			p[1] = '\n';
			break;
		case 't':
			p[1] = '\t';
			break;
		case 'x':
			while (isxdigit(p[off+1]) && off < 3)
				off++;
			p[off] = (char)strtoul(p+2, 0, 16);
			break;
		default:
			if (!isdigit(p[1]) || p[1] >= '8')
				break;
			while ((isdigit(p[off+1]) && p[off+1] < '8') && off < 3)
				off++;
			p[off] = (char)strtoul(p+2, 0, 8);
			break;
	}
	len -= off;
	memcpy(p, p+off, len);
	p[len] = 0;
}

int
strexpand (char *str, int len)
{
	char *p;
	int off;

	for (p = str; p < str+len; p++) {
		if (*p != '\\')
			continue;
		off = 1;
		switch (p[1]) {
			case 'r':
				p[1] = '\r';
				break;
			case 'n':
				p[1] = '\n';
				break;
			case 't':
				p[1] = '\t';
				break;
			case 'x':
				while (isxdigit(p[off+1]) && off < 3)
					off++;
				p[off] = (char)strtoul(p+2, 0, 16);
				break;
			default:
				if (!isdigit(p[1]) || p[1] >= '8')
					break;
				while ((isdigit(p[off+1]) && p[off+1] < '8') && off < 3)
					off++;
				p[off] = (char)strtoul(p+2, 0, 8);
				break;
		}
		len -= off;
		memcpy(p, p+off, str+len - p);
		str[len] = 0;
	}

	return len;
}

int
strunexpand (char *str, int slen, char *buf, int blen)
{
	char *p, *bp, c;

	for (p = str, bp = buf; p < str+slen && bp < buf+blen; p++, bp++) {
		if (isgraph(*p) && (*p != '\'' && *p != '"' && *p != '?' && *p != '*')) {
			*bp = *p;
			continue;
		}
		*bp++ = '\\';
		switch (*p) {
			case '\r':
				*bp = 'r';
				break;
			case '\n':
				*bp = 'n';
				break;
			case '\t':
				*bp = 't';
				break;
			case ' ':
			case '\'':
			case '"':
			case '*':
			case '?':
				*bp = *p;
				break;
			default:
				*bp++ = 'x';
				c = (*p >> 4) & 0xf;
				c = c > 9 ? c - 0xa + 'a' : c + '0';
				*bp++ = c;
				c = *p & 0xf;
				c = c > 9 ? c - 0xa + 'a' : c + '0';
				*bp = c;
		}
	}
	if (bp == buf+blen)
		bp--;
	*bp = 0;
	blen = bp - buf;

	return blen;
}

void
hotline_client_init (int argc, char **argv)
{
	char *home, *user;
	struct passwd *pwe;
	int load_hxrc = 1;
	char *load_this = 0;
	struct opt_r opt;
	int i;

	opt.err_printf = hxd_log;
	opt.ind = 0;
	while ((i = getopt_r(argc, argv, "ql:", &opt)) != EOF) {
		switch (i) {
			case 'l':
				load_this = opt.arg;
				break;
			case 'q':
				load_hxrc = 0;
				break;
		}
	}

	(void)setlocale(LC_ALL, "");

	home = getenv("HOME");
	user = getenv("USER");
	if (!home || !user) {
		pwe = getpwuid(getuid());
		if (!pwe) {
			hxd_log("getpwuid: %s", strerror(errno));
		} else {
			if (!home)
				home = pwe->pw_dir;
			if (!user)	
				user = pwe->pw_name;
		}
	}
	strncpy(homedir, home ? home : ".", MAXPATHLEN - 1);

	hx_output.init(argc, argv);

	memset(&hx_htlc, 0, sizeof(struct htlc_conn));
	hx_htlc.icon = 500;
	if (user)
		strncpy(hx_htlc.name, user, 31);
	else
		strcpy(hx_htlc.name, "Evaluation 0wn3r");

	gen_command_hash();

	last_msg_nick[0] = 0;

	hx_output.status();

	if (load_hxrc)
		hx_load(load_this ? load_this : "~/.hxrc");
}
