#ifndef __hxd_HX_H
#define __hxd_HX_H 1

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <sys/types.h>
#include <stdio.h>
#include <limits.h>
#include <sys/param.h>

#define USE_READLINE	1

#if USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#ifndef MAXPATHLEN
#ifdef PATH_MAX
#define MAXPATHLEN PATH_MAX
#else
#define MAXPATHLEN 4095
#endif
#endif

#if MAXPATHLEN > 4095
#undef MAXPATHLEN
#define MAXPATHLEN 4095
#endif

extern struct htlc_conn hx_htlc;

extern int LI, CO;

extern char last_msg_nick[32];

#define INFOPREFIX	" \033[36m[\033[0m\033[32mhx\033[0m\033[36m]\033[0m "
#define RED		"\033[0;31m"
#define RED_BOLD	"\033[1;31m"
#define GREEN		"\033[0;32m"
#define GREEN_BOLD	"\033[1;32m"
#define DEFAULT		"\033[0m"

extern void hx_printf (const char *fmt, ...);
extern void hx_save (char *filename);
extern void hx_load (char *filename);

extern void gen_command_hash (void);
extern void hx_command (char *str);
extern void hx_send_chat (char *str);
extern void hotline_client_input (char *str);

extern void expand_tilde (char *buf, char *str);
extern char *expand_user_name (char *name, unsigned int nlen);
extern int expand_path (char *path, int len, char *pathbuf);
extern u_int32_t cmd_arg(int argc, char *str);

extern void chrexpand (char *str, int len);
extern int strexpand (char *str, int len);
extern int strunexpand (char *str, int slen, char *buf, int blen);

#if !defined(__va_copy)
#define __va_copy(_dst, _src) ((_dst) = (_src))
#endif

static inline void
strip_ansi (char *buf, int len)
{
	register char *p, *end = buf + len;

	for (p = buf; p < end; p++)
		if (*p < 31 && *p > 13 && *p != 15 && *p != 22)
				*p = (*p & 127) | 64;
}

struct user {
	struct user *next, *prev;
	u_int32_t uid;
	u_int16_t icon;
	u_int16_t color;
	u_int8_t name[32];
};

extern struct user *user_new (struct user **utailp);
extern void user_delete (struct user **utailp, struct user *user);
extern struct user *user_with_uid (struct user *ulist, u_int32_t uid);
extern struct user *user_with_name (struct user *ulist, u_int8_t *name);

struct chat {
	struct chat *next, *prev;
	u_int32_t cid;
	u_int32_t nusers;
	struct user __user_list;
	struct user *user_list;
	struct user *user_tail;
	char subject[256];
};

extern struct chat *chat_list, *chat_front;

extern struct chat *chat_new (u_int32_t cid);
extern void chat_delete (struct chat *chat);
extern struct chat *chat_with_cid (u_int32_t cid);

struct output_functions {
	void (*init)(int argc, char **argv);
	void (*cleanup)(void);
	void (*status)(void);
	void (*clear)(void);
	void (*clear_input)(void);
	void (*mode_underline)(void);
	void (*mode_clear)(void);
	void (*chat)(u_int32_t cid, char *chat, u_int16_t len);
	void (*msg)(u_int32_t uid, char *name, char *msgbuf, u_int16_t msglen);
	void (*agreement)(char *agreement, u_int16_t len);
};

extern struct output_functions hx_output;

#endif /* ndef __hxd_HX_H */
