#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/wait.h>
#include <fnmatch.h>
#include <glob.h>
#include <ctype.h>
#include "hxd.h"
#include "hx.h"
#include "xmalloc.h"
#include "getopt.h"

#ifdef SOCKS
#include "socks.h"
#endif

#ifdef CONFIG_HOPE
#include "sha.h"
#include "md5.h"
#endif

#if defined(CONFIG_HFS)
#include "hfs.h"
#endif

#if defined(CONFIG_HTXF_PTHREAD)
#include <pthread.h>
#elif defined(CONFIG_HTXF_CLONE)
#include <sched.h>
#define CLONE_STACKSIZE	0x10000
#endif

#define MAX_HOTLINE_PACKET_LEN 0x100000

#if defined(CONFIG_HAL)
extern int hal_active;
extern void cmd_hal (int argc, char **argv, char *str);
extern void hal_rcv (char *input, char *nick);
#endif

static void away_log (const char *fmt, ...);

struct htlc_conn hx_htlc;

static char hx_hostname[128];

static char *hx_timeformat = "%c";

static u_int8_t dir_char = '/';

#define DIRCHAR			'/'
#define UNKNOWN_TYPECREA	"TEXTR*ch"

typedef enum { snd_login, snd_message, snd_chat, snd_news, snd_join, snd_part,
               snd_chat_invite, snd_file_done, snd_error } snd_events;
static void play_sound(snd_events event);

static void
hx_rcv_chat (struct htlc_conn *htlc)
{
	u_int32_t cid = 0;
	u_int16_t len = 0;
	char chatbuf[8192 + 1], *chat;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_CHAT:
				len = (ntohs(dh->len) > sizeof(chatbuf) - 1 ? sizeof(chatbuf) - 1 : ntohs(dh->len));
				memcpy(chatbuf, dh->data, len);
				break;
			case HTLS_DATA_CHAT_ID:
				dh_getint(cid);
				break;
		}
	dh_end()
	CR2LF(chatbuf, len);
	strip_ansi(chatbuf, len);
	
	if (chatbuf[0] == '\n') {
		chat = chatbuf+1;
		len--;
	} else {
		chat = chatbuf;
	}
#if defined(CONFIG_HAL)
	if (hal_active) {
		char nickname[32], *user, *sep, *lf, ss;

		chatbuf[len] = 0;
		sep = chat + 13;
		if ((sep[0] == ':' && sep[1] == ' ' && sep[2] == ' ')
		    || (sep[0] == ' ' && sep[1] == '-' && sep[2] == ' ')) {
			ss = sep[0];
			sep[0] = 0;
			user = chat;
			while (*user == ' ' && *user)
				user++;
			strcpy(nickname, user);
			sep[0] = ss;
			if (strcmp(nickname, htlc->name)) {
				sep += 3;
				lf = strchr(sep, '\n');
				if (lf)
					*lf = 0;
				hal_rcv(sep, nickname);
				if (lf)
					*lf = '\n';
			}
		}
	}
#endif
	hx_output.chat(cid, chat, len);
	play_sound(snd_chat);
}

char *
expand_user_name (char *name, unsigned int nlen)
{
	struct user *userp, *user = 0;

	for (userp = chat_list->user_list->next; userp; userp = userp->next) {
		if (!strncmp(userp->name, name, nlen)) {
			if (user)
				return 0;
			user = userp;
		}
	}

	return user ? user->name : 0;
}

static void
hx_rcv_msg (struct htlc_conn *htlc)
{
	u_int32_t uid = 0;
	u_int16_t msglen = 0, nlen = 0;
	char msgbuf[4096 + 1], name[64 + 1];

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
			case HTLS_DATA_MSG:
				msglen = ntohs(dh->len) > 4096 ? 4096 : ntohs(dh->len);
				memcpy(msgbuf, dh->data, msglen);
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 64 ? 64 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				strip_ansi(name, nlen);
				break;
		}
	dh_end()
	CR2LF(msgbuf, msglen);
	strip_ansi(msgbuf, msglen);
	name[nlen] = 0;
	hx_output.msg(uid, name, msgbuf, msglen);
	msgbuf[msglen] = 0;
	away_log("[%s(%u)]: %s\n", name, uid, msgbuf);
	play_sound(snd_message);
	if (!*last_msg_nick) {
		strncpy(last_msg_nick, name, 31);
		last_msg_nick[31] = 0;
	}
}

static void
hx_rcv_agreement_file (struct htlc_conn *htlc)
{
	dh_start(htlc)
		if (ntohs(dh->type) != HTLS_DATA_AGREEMENT)
			continue;
		CR2LF(dh->data, ntohs(dh->len));
		strip_ansi(dh->data, ntohs(dh->len));
		hx_output.agreement(dh->data, ntohs(dh->len));
	dh_end()
}

static size_t news_len = 0;
static u_int8_t *news_buf = 0;

static void
hx_rcv_news_post (struct htlc_conn *htlc)
{
	dh_start(htlc)
		if (ntohs(dh->type) != HTLS_DATA_NEWS)
			continue;
		news_len += ntohs(dh->len);
		news_buf = xrealloc(news_buf, news_len + 1);
		memmove(&(news_buf[ntohs(dh->len)]), news_buf, news_len - ntohs(dh->len));
		memcpy(news_buf, dh->data, ntohs(dh->len));
		CR2LF(news_buf, ntohs(dh->len));
		strip_ansi(news_buf, ntohs(dh->len));
		news_buf[news_len] = 0;
		hx_printf(INFOPREFIX "news posted:\n%.*s\n", ntohs(dh->len), news_buf);
		play_sound(snd_news);
	dh_end()
}

struct task {
	struct task *next, *prev;
	u_int32_t trans;
	void (*rcv)();
	void *ptr;
};

static struct task __task_list = {0, 0, 0, 0, 0}, *task_list = &__task_list, *task_tail = &__task_list;

static struct task *
task_new (struct htlc_conn *htlc, void (*rcv)(), void *ptr)
{
	struct task *tsk;

	tsk = xmalloc(sizeof(struct task));
	tsk->trans = htlc->trans;
	tsk->rcv = rcv;
	tsk->ptr = ptr;

	tsk->next = 0;
	tsk->prev = task_tail;
	task_tail->next = tsk;
	task_tail = tsk;

	return tsk;
}

static void
task_delete (struct task *tsk)
{
	if (tsk->next)
		tsk->next->prev = tsk->prev;
	if (tsk->prev)
		tsk->prev->next = tsk->next;
	if (task_tail == tsk)
		task_tail = tsk->prev;
	xfree(tsk);
}

static struct task *
task_with_trans (u_int32_t trans)
{
	struct task *tsk;

	for (tsk = task_list->next; tsk; tsk = tsk->next)
		if (tsk->trans == trans)
			return tsk;

	return 0;
}

static void
task_error (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;

	dh_start(htlc)
		if (ntohs(dh->type) == HTLS_DATA_TASKERROR) {
			CR2LF(dh->data, ntohs(dh->len));
			strip_ansi(dh->data, ntohs(dh->len));
			hx_printf(INFOPREFIX "task 0x%08x error: %.*s\n",
				  ntohl(h->trans), ntohs(dh->len), dh->data);
			play_sound(snd_error);
		}
	dh_end()
}

static void
hx_rcv_task (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	u_int32_t trans = ntohl(h->trans);
	struct task *tsk = task_with_trans(trans);

	if ((ntohs(h->flag) & 1))
		task_error(htlc);
	if (tsk) {
		/* XXX tsk->rcv might call task_delete */
		int fd = htlc->fd;
		if (tsk->rcv)
			tsk->rcv(htlc, tsk->ptr);
		if (hxd_files[fd].conn.htlc)
			task_delete(tsk);
	} else {
		hx_printf(INFOPREFIX "got task 0x%08x\n", trans);
	}
}

#if defined(CONFIG_HOPE)
static void
hash2str (char *str, u_int8_t *hash, int len)
{
	int i, c;

	for (i = 0; i < len; i++) {
		c = hash[i] >> 4;
		c = c > 9 ? c - 0xa + 'a' : c + '0';
		str[i*2 + 0] = c;
		c = hash[i] & 0xf;
		c = c > 9 ? c - 0xa + 'a' : c + '0';
		str[i*2 + 1] = c;
	}
}
#endif

static void
user_print (struct user *ulist, char *str)
{
	struct user *userp;

	hx_output.mode_underline();
	hx_printf("   uid | nickname                        |  icon | level | stat |\n");
	hx_output.mode_clear();
	for (userp = ulist->next; userp; userp = userp->next) {
		if (str && !strstr(userp->name, str))
			continue;
		hx_printf(" %5u | %s%-31s" DEFAULT " | %5u | %5s | %4s |\n",
			  userp->uid,
			  userp->color % 4 > 1 ? (userp->color % 2 ? RED : RED_BOLD)
						: (userp->color % 2 ? GREEN : GREEN_BOLD),
			  userp->name, userp->icon,
			  userp->color % 4 > 1 ? "ADMIN" : "USER", userp->color % 2 ? "AWAY" : "\0");
	}
}

static void
hx_rcv_user_change (struct htlc_conn *htlc)
{
	u_int32_t uid = 0, cid = 0, icon = 0, color = 0;
	u_int16_t nlen = 0, got_color = 0;
	u_int8_t name[32];
	struct chat *chat;
	struct user *user;
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;

	if ((ntohs(h->flag) & 1))
		return;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
			case HTLS_DATA_ICON:
				dh_getint(icon);
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				strip_ansi(name, nlen);
				break;
			case HTLS_DATA_COLOUR:
				dh_getint(color);
				got_color = 1;
				break;
			case HTLS_DATA_CHAT_ID:
				dh_getint(cid);
				break;
		}
	dh_end()
	chat = chat_with_cid(cid);
	if (!chat) {
		chat = chat_new(cid);
		chat_front = chat;
	}
	user = user_with_uid(chat->user_list, uid);
	if (!user) {
		char *col;
		user = user_new(&chat->user_tail);
		chat->nusers++;
		user->uid = uid;
		col = color % 4 > 1 ? (color % 2 ? RED : RED_BOLD)
				     : (color % 2 ? GREEN : GREEN_BOLD);
		if (chat->cid)
			hx_printf(INFOPREFIX "chat 0x%08x join: %s%.*s%s [%u:%u:%u]\n",
					chat->cid, col, nlen, name, DEFAULT, uid, icon, color);
		else
			hx_printf(INFOPREFIX "join: %s%.*s%s [%u:%u:%u]\n",
					col, nlen, name, DEFAULT, uid, icon, color);
		play_sound(snd_join);
	} else {
		char *col0, *col1;
		col0 = user->color % 4 > 1 ? (user->color % 2 ? RED : RED_BOLD)
					 : (user->color % 2 ? GREEN : GREEN_BOLD);
		if (!got_color)
			col1 = col0;
		else
			col1 = color % 4 > 1 ? (color % 2 ? RED : RED_BOLD)
					      : (color % 2 ? GREEN : GREEN_BOLD);
		if (chat->cid)
			hx_printf(INFOPREFIX "chat 0x%08x %s%s%s [%u:%u:%u] is now known as %s%.*s%s [%u:%u:%u]\n",
					chat->cid, col0, user->name, DEFAULT, user->uid, user->icon, user->color,
					col1, nlen, name, DEFAULT, uid, icon, color);
		else
			hx_printf(INFOPREFIX "%s%s%s [%u:%u:%u] is now known as %s%.*s%s [%u:%u:%u]\n",
					col0, user->name, DEFAULT, user->uid, user->icon, user->color,
					col1, nlen, name, DEFAULT, uid, icon, color);
	}
	if (nlen) {
		memcpy(user->name, name, nlen);
		user->name[nlen] = 0;
	}
	if (icon)
		user->icon = icon;
	if (got_color)
		user->color = color;
	if (uid && uid == htlc->uid) {
		htlc->uid = user->uid;
		htlc->icon = user->icon;
		htlc->color = user->color;
		strcpy(htlc->name, user->name);
		hx_output.status();
	}
}

static void
hx_rcv_user_part (struct htlc_conn *htlc)
{
	u_int32_t uid = 0, cid = 0;
	struct chat *chat;
	struct user *user;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
			case HTLS_DATA_CHAT_ID:
				dh_getint(cid);
				break;
		}
	dh_end()
	chat = chat_with_cid(cid);
	if (!chat)
		return;
	user = user_with_uid(chat->user_list, uid);
	if (user) {
		char *col = user->color % 4 > 1 ? (user->color % 2 ? RED : RED_BOLD)
					      : (user->color % 2 ? GREEN : GREEN_BOLD);
		if (chat->cid)
			hx_printf(INFOPREFIX "chat 0x%08x parts: %s%s%s [%u:%u:%u]\n",
				  chat->cid, col, user->name, DEFAULT, user->uid, user->icon, user->color);
		else
			hx_printf(INFOPREFIX "parts: %s%s%s [%u:%u:%u]\n",
				  col, user->name, DEFAULT, user->uid, user->icon, user->color);
		user_delete(&chat->user_tail, user);
		chat->nusers--;
		play_sound(snd_part);
	}
}

static void
hx_rcv_chat_subject (struct htlc_conn *htlc)
{
	u_int32_t cid = 0;
	u_int16_t slen = 0;
	u_int8_t subject[256];
	struct chat *chat;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_CHAT_ID:
				dh_getint(cid);
				break;
			case HTLS_DATA_CHAT_SUBJECT:
				slen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
				memcpy(subject, dh->data, slen);
				break;
		}
	dh_end()
	if (slen) {
		chat = chat_with_cid(cid);
		if (!chat)
			return;
		memcpy(chat->subject, subject, slen);
		chat->subject[slen] = 0;
		hx_printf(INFOPREFIX "chat 0x%08x subject: %s\n", cid, chat->subject);
	}
}

static void
hx_rcv_chat_invite (struct htlc_conn *htlc)
{
	u_int32_t uid = 0, cid = 0;
	u_int16_t nlen;
	u_int8_t name[32];

	name[0] = 0;
	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
			case HTLS_DATA_CHAT_ID:
				dh_getint(cid);
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				strip_ansi(name, nlen);
				break;
		}
	dh_end()
	hx_printf(INFOPREFIX "[%s(%u)] invites you to chat 0x%08x\n", name, uid, cid);
	play_sound(snd_chat_invite);
}

static void
hx_rcv_user_selfinfo (struct htlc_conn *htlc)
{
	struct hl_userlist_hdr *uh;
	u_int16_t nlen;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_ACCESS:
				if (ntohs(dh->len) != 8)
					break;
				memcpy(&htlc->access, dh->data, 8);
				break;
			case HTLS_DATA_USER_LIST:
				if (ntohs(dh->len) < (SIZEOF_HL_USERLIST_HDR - SIZEOF_HL_DATA_HDR))
					break;
				uh = (struct hl_userlist_hdr *)dh;
				htlc->uid = ntohs(uh->uid);
				htlc->icon = ntohs(uh->icon);
				htlc->color = ntohs(uh->color);
				nlen = ntohs(uh->nlen) > 31 ? 31 : ntohs(uh->nlen);
				memcpy(htlc->name, uh->name, nlen);
				htlc->name[nlen] = 0;
				hx_output.status();
				break;
		}
	dh_end()
}

static void
hx_rcv_dump (struct htlc_conn *htlc)
{
	int fd;

	fd = open("hx.dump", O_WRONLY|O_APPEND|O_CREAT, 0644);
	if (fd < 0)
		return;
	write(fd, htlc->in.buf, htlc->in.pos);
	fsync(fd);
	close(fd);
}

static void
hx_rcv_hdr (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	u_int32_t len;

	if (ntohl(h->len) < 2)
		len = 0;
	else
		len = (ntohl(h->len) > MAX_HOTLINE_PACKET_LEN ? MAX_HOTLINE_PACKET_LEN : ntohl(h->len)) - 2;

	/* htlc->trans = ntohl(h->trans); */
	htlc->rcv = 0;
	switch (ntohl(h->type)) {
		case HTLS_HDR_CHAT:
			htlc->rcv = hx_rcv_chat;
			break;
		case HTLS_HDR_MSG:
			htlc->rcv = hx_rcv_msg;
			break;
		case HTLS_HDR_USER_CHANGE:
		case HTLS_HDR_CHAT_USER_CHANGE:
			htlc->rcv = hx_rcv_user_change;
			break;
		case HTLS_HDR_USER_PART:
		case HTLS_HDR_CHAT_USER_PART:
			htlc->rcv = hx_rcv_user_part;
			break;
		case HTLS_HDR_NEWS_POST:
			htlc->rcv = hx_rcv_news_post;
			break;
		case HTLS_HDR_TASK:
			htlc->rcv = hx_rcv_task;
			break;
		case HTLS_HDR_CHAT_SUBJECT:
			htlc->rcv = hx_rcv_chat_subject;
			break;
		case HTLS_HDR_CHAT_INVITE:
			htlc->rcv = hx_rcv_chat_invite;
			break;
		case HTLS_HDR_MSG_BROADCAST:
			htlc->rcv = hx_rcv_msg;
			break;
		case HTLS_HDR_USER_SELFINFO:
			htlc->rcv = hx_rcv_user_selfinfo;
			break;
		case HTLS_HDR_AGREEMENT:
			htlc->rcv = hx_rcv_agreement_file;
			break;
		case HTLS_HDR_POLITEQUIT:
			hx_printf(INFOPREFIX, "polite quit\n");
			htlc->rcv = hx_rcv_msg;
			break;
		default:
			hx_printf(INFOPREFIX "unknown header type 0x%08x\n", ntohl(h->type));
			htlc->rcv = hx_rcv_dump;
			break;
	}

	if (len) {
		qbuf_set(&htlc->in, htlc->in.pos, len);
	} else {
		if (htlc->rcv)
			htlc->rcv(htlc);
		htlc->rcv = hx_rcv_hdr;
		qbuf_set(&htlc->in, 0, SIZEOF_HL_HDR);
	}
}

static void
hx_rcv_magic (struct htlc_conn *htlc)
{
	htlc->rcv = hx_rcv_hdr;
	qbuf_set(&htlc->in, 0, SIZEOF_HL_HDR);
}

static void
hx_htlc_close (struct htlc_conn *htlc)
{
	int fd = htlc->fd;
	struct chat *chat, *cnext;
	struct user *user, *unext;
	struct task *tsk, *tsknext;
	char buf[16];

	inet_ntoa_r(htlc->sockaddr.sin_addr, buf, sizeof(buf));
	hx_printf(INFOPREFIX "%s:%u: connection closed\n", buf, ntohs(htlc->sockaddr.sin_port));

	close(fd);
	FD_CLR(fd, &hxd_rfds);
	FD_CLR(fd, &hxd_wfds);
	if (htlc->in.buf) {
		xfree(htlc->in.buf);
		htlc->in.buf = 0;
	}
	if (htlc->out.buf) {
		xfree(htlc->out.buf);
		htlc->out.buf = 0;
	}
	memset(&hxd_files[fd], 0, sizeof(struct hxd_file));
	htlc->fd = 0;
	htlc->uid = 0;
	htlc->color = 0;
#ifdef CONFIG_HOPE
	memset(htlc->sessionkey, 0, sizeof(htlc->sessionkey));
#endif
	memset(htlc->login, 0, sizeof(htlc->login));

	for (chat = chat_list; chat; chat = cnext) {
		cnext = chat->next;
		for (user = chat->user_list->next; user; user = unext) {
			unext = user->next;
			user_delete(&chat->user_tail, user);
		}
		if (chat != chat_list)
			chat_delete(chat);
	}
	memset(chat_list, 0, sizeof(struct chat));
	chat_list->user_list = chat_list->user_tail = &chat_list->__user_list;
	for (tsk = task_list->next; tsk; tsk = tsknext) {
		tsknext = tsk->next;
		task_delete(tsk);
	}
}

static void
htlc_read (int fd)
{
	ssize_t r;
	struct htlc_conn *htlc = hxd_files[fd].conn.htlc;

	r = read(fd, &htlc->in.buf[htlc->in.pos], htlc->in.len);
	if (r == 0 || (r < 0 && errno != EWOULDBLOCK && errno != EINTR)) {
		hx_htlc_close(htlc);
	} else {
		htlc->in.pos += r;
		htlc->in.len -= r;
		if (!htlc->in.len) {
			if (htlc->rcv) {
				if (htlc->rcv == hx_rcv_hdr)
					hx_rcv_hdr(htlc);
				else {
					htlc->rcv(htlc);
					goto reset;
				}
			} else {
reset:
				htlc->rcv = hx_rcv_hdr;
				qbuf_set(&htlc->in, 0, SIZEOF_HL_HDR);
			}
		}
	}
}

static void
htlc_write (int fd)
{
	ssize_t r;
	struct htlc_conn *htlc = hxd_files[fd].conn.htlc;

	r = write(fd, &htlc->out.buf[htlc->out.pos], htlc->out.len);
	if (r == 0 || (r < 0 && errno != EWOULDBLOCK && errno != EINTR)) {
		hx_htlc_close(htlc);
	} else {
		htlc->out.pos += r;
		htlc->out.len -= r;
		if (!htlc->out.len) {
			htlc->out.pos = 0;
			htlc->out.len = 0;
			FD_CLR(fd, &hxd_wfds);
		}
	}
}

static void
rcv_task_msg (struct htlc_conn *htlc __attribute__((__unused__)), char *msg_buf)
{
	hx_printf("%s\n", msg_buf);
	xfree(msg_buf);
}

#define COMMAND(x) static void			\
cmd_##x (int argc __attribute__((__unused__)),	\
	char **argv __attribute__((__unused__)),\
	char *str __attribute__((__unused__)),	\
	struct htlc_conn *htlc __attribute__((__unused__)))

COMMAND(msg)
{
	u_int32_t uid;
	char *name, *msg, *buf;
	struct user *user = 0;
	size_t buflen;

	name = argv[1];
	if (!name)
		goto usage;
	msg = str + (cmd_arg(2, str) >> 16);
	if (!*msg) {
usage:		hx_printf(INFOPREFIX "usage %s <uid> <msg>\n", argv[0]);
		return;
	}
	uid = atou32(name);
	if (!uid) {
		user = user_with_name(chat_list->user_list, name);
		if (!user) {
			hx_printf(INFOPREFIX "%s: no such nickname %s\n", argv[0], name);
			return;
		}
		uid = user->uid;
	}
	strncpy(last_msg_nick, name, 31);
	last_msg_nick[31] = 0;
	if (!user)
		user = user_with_uid(chat_list->user_list, uid);
	if (user) {
		char *col = user->color % 4 > 1 ? (user->color % 2 ? RED : RED_BOLD)
					      : (user->color % 2 ? GREEN : GREEN_BOLD);
		buflen = strlen(msg) + strlen(user->name) + 20 + strlen(col) + strlen(DEFAULT);
		buf = xmalloc(buflen);
		snprintf(buf, buflen, "[%s%s%s(%u)]-> %s", col, user->name, DEFAULT, uid, msg);
	} else {
		buflen = strlen(msg) + 20;
		buf = xmalloc(buflen);
		snprintf(buf, buflen, "[(%u)]-> %s", uid, msg);
	}

	uid = htonl(uid);
	task_new(htlc, rcv_task_msg, buf);
	hlwrite(htlc, HTLC_HDR_MSG, 0, 2,
		HTLC_DATA_UID, 4, &uid,
		HTLC_DATA_MSG, strlen(msg), msg);
}

COMMAND(nick)
{
	u_int16_t icon16, icon = 0;
	char *name;
	struct opt_r opt;
	int i;

	opt.err_printf = hx_printf;
	opt.ind = 0;
	while ((i = getopt_r(argc, argv, "i:", &opt)) != EOF) {
		if (i == 'i')
			icon = atou32(opt.arg);
	}
	name = argv[opt.ind];
	if (!name) {
		hx_printf(INFOPREFIX "usage: %s <nickname> [-i icon]\n", argv[0]);
		return;
	}
	i = strlen(name) > 31 ? 31 : strlen(name);
	memcpy(htlc->name, name, i);
	htlc->name[i] = 0;
	if (icon)
		htlc->icon = icon;

	icon16 = htons(htlc->icon);
	hlwrite(htlc, HTLC_HDR_USER_CHANGE, 0, 2,
		HTLC_DATA_ICON, 2, &icon16,
		HTLC_DATA_NAME, i, name);
	hx_output.status();
}

COMMAND(icon)
{
	u_int16_t icon, icon16;

	if (!argv[1]) {
		hx_printf(INFOPREFIX "usage: %s <icon>\n", argv[0]);
		return;
	}
	icon = atou16(argv[1]);
	htlc->icon = icon;
	icon16 = htons(icon);
	hlwrite(htlc, HTLC_HDR_USER_CHANGE, 0, 2,
		HTLC_DATA_ICON, 2, &icon16,
		HTLC_DATA_NAME, strlen(htlc->name), htlc->name);
	hx_output.status();
}

COMMAND(quit)
{
	hx_output.cleanup();
	exit(0);
}

#ifdef CONFIG_HOPE
u_int16_t
hmac_xxx (u_int8_t *md, u_int8_t *key, u_int32_t keylen, u_int8_t *text, u_int32_t textlen, u_int8_t *macalg)
{
	u_int8_t mdkey[20];
	u_int8_t k_ipad[64], k_opad[64];
	unsigned int i;

	if (!strcmp(macalg, "SHA1")) {
		struct sha_ctx ctx;

		sha_init(&ctx);
		sha_update(&ctx, key, keylen);
		sha_update(&ctx, text, textlen);
		sha_final(md, &ctx);

		return 20;
	} else if (!strcmp(macalg, "MD5")) {
		struct md5_ctx ctx;

		md5_init_ctx(&ctx);
		md5_process_bytes(key, keylen, &ctx);
		md5_process_bytes(text, textlen, &ctx);
		md5_finish_ctx(&ctx, md);

		return 16;
	}

	if (keylen > 64) {
		if (!strcmp(macalg, "HMAC-SHA1")) {
			struct sha_ctx ctx;

			sha_init(&ctx);
			sha_update(&ctx, key, keylen);
			sha_final(mdkey, &ctx);
			keylen = 20;
		} else if (!strcmp(macalg, "HMAC-MD5")) {
			struct md5_ctx ctx;

			md5_init_ctx(&ctx);
			md5_process_bytes(key, keylen, &ctx);
			md5_finish_ctx(&ctx, mdkey);
			keylen = 16;
		} else if (!strncmp(macalg, "HMAC-HAVAL", 10)) {
			return 0;
		} else {
			return 0;
		}
		key = mdkey;
	}

	memcpy(k_ipad, key, keylen);
	memcpy(k_opad, key, keylen);
	memset(k_ipad+keylen, 0, 64 - keylen);
	memset(k_opad+keylen, 0, 64 - keylen);

	for (i = 0; i < 64; i++) {
		k_ipad[i] ^= 0x36;
		k_opad[i] ^= 0x5c;
	}

	if (!strcmp(macalg, "HMAC-SHA1")) {
		struct sha_ctx ctx;

		sha_init(&ctx);
		sha_update(&ctx, k_ipad, 64);
		sha_update(&ctx, text, textlen);
		sha_final(md, &ctx);

		sha_init(&ctx);
		sha_update(&ctx, k_opad, 64);
		sha_update(&ctx, md, 20);
		sha_final(md, &ctx);

		return 20;
	} else if (!strcmp(macalg, "HMAC-MD5")) {
		struct md5_ctx ctx;

		md5_init_ctx(&ctx);
		md5_process_bytes(k_ipad, 64, &ctx);
		md5_process_bytes(text, textlen, &ctx);
		md5_finish_ctx(&ctx, md);

		md5_init_ctx(&ctx);
		md5_process_bytes(k_opad, 64, &ctx);
		md5_process_bytes(md, 16, &ctx);
		md5_finish_ctx(&ctx, md);

		return 16;
	} else if (!strncmp(macalg, "HMAC-HAVAL", 10)) {
		return 0;
	} else {
		return 0;
	}
}
#endif

static void rcv_task_user_list (struct htlc_conn *htlc, struct chat *chat);

static void
rcv_task_login (struct htlc_conn *htlc, char *pass __attribute__((__unused__)))
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	char buf[16];
	u_int32_t uid;
#ifdef CONFIG_HOPE
	u_int16_t icon16;
	u_int16_t sklen = 0, macalglen = 0, secure_login = 0, secure_password = 0;

	if (pass) {
		dh_start(htlc)
			switch (ntohs(dh->type)) {
				case HTLS_DATA_LOGIN:
					if (ntohs(dh->len) && ntohs(dh->len) == strlen(htlc->macalg) && !memcmp(htlc->macalg, dh->data, ntohs(dh->len)))
						secure_login = 1;
					break;
				case HTLS_DATA_PASSWORD:
					if (ntohs(dh->len) && ntohs(dh->len) == strlen(htlc->macalg) && !memcmp(htlc->macalg, dh->data, ntohs(dh->len)))
						secure_password = 1;
					break;
				case HTLS_DATA_MAC_ALG:
					macalglen = ntohs(dh->len);
					if (macalglen >= 4) {
						macalglen = dh->data[2] > 31 ? 31 : dh->data[2];
						if (macalglen <= ntohs(dh->len) - 3) {
							memcpy(htlc->macalg, dh->data+3, macalglen);
							htlc->macalg[macalglen] = 0;
						} else
							macalglen = 0;
					} else
						macalglen = 0;
					break;
				case HTLS_DATA_SESSIONKEY:
					sklen = ntohs(dh->len) > sizeof(htlc->sessionkey) ? sizeof(htlc->sessionkey) : ntohs(dh->len);
					memcpy(htlc->sessionkey, dh->data, sklen);
					htlc->sklen = sklen;
					break;
			}
		dh_end()

		if (!macalglen) {
			hx_printf(INFOPREFIX "bad macalg from server\n");
			hx_htlc_close(htlc);
			return;
		}
		if (sklen < 6) {
			hx_printf(INFOPREFIX "sessionkey not big enough\n");
			hx_htlc_close(htlc);
			return;
		}
		if (*((u_int32_t *)(htlc->sessionkey)) != htlc->sockaddr.sin_addr.s_addr
		    || *((u_int16_t *)(htlc->sessionkey + 4)) != htlc->sockaddr.sin_port) {
			char fakeabuf[16], realabuf[16];
			struct in_addr fakeinaddr;

			fakeinaddr.s_addr = *((u_int32_t *)(htlc->sessionkey));
			inet_ntoa_r(fakeinaddr, fakeabuf, sizeof(fakeabuf));
			inet_ntoa_r(htlc->sockaddr.sin_addr, realabuf, sizeof(realabuf));
			hx_printf(INFOPREFIX "%s:%u != %s:%u\n", fakeabuf, ntohs(*((u_int16_t *)(htlc->sessionkey + 4))),
								 realabuf, ntohs(htlc->sockaddr.sin_port));
			hx_htlc_close(htlc);
			return;
		}

		task_new(htlc, rcv_task_login, 0);
		icon16 = htons(htlc->icon);
		if (!(ntohs(h->flag) & 1) && sklen) {
			u_int8_t password_mac[20];
			u_int8_t login[32];
			u_int16_t llen, pmaclen;

			if (secure_login) {
				llen = hmac_xxx(login, htlc->login, strlen(htlc->login), htlc->sessionkey, sklen, htlc->macalg);
				if (!llen) {
					hx_printf("%s:%d: bad mac alg\n", __FILE__, __LINE__);
					return;
				}
			} else {
				llen = strlen(htlc->login);
				hl_encode(login, htlc->login, llen);
				login[llen] = 0;
			}
			pmaclen = hmac_xxx(password_mac, pass, strlen(pass), htlc->sessionkey, sklen, htlc->macalg);
			if (!pmaclen) {	
				hx_printf("%s:%d: bad mac alg\n", __FILE__, __LINE__);
				return;
			}
			hlwrite(htlc, HTLC_HDR_LOGIN, 0, 4,
				HTLC_DATA_LOGIN, llen, login,
				HTLC_DATA_PASSWORD, pmaclen, password_mac,
				HTLC_DATA_NAME, strlen(htlc->name), htlc->name,
				HTLC_DATA_ICON, 2, &icon16);
		}
		xfree(pass);
	} else {
#endif
		inet_ntoa_r(htlc->sockaddr.sin_addr, buf, sizeof(buf));
		hx_printf(INFOPREFIX "%s:%u: login %s\n",
			  buf, ntohs(htlc->sockaddr.sin_port), (ntohs(h->flag) & 1) ? "failed?" : "successful");
		if (!(ntohs(h->flag) & 1)) {
			play_sound(snd_login);
			dh_start(htlc)
				switch (ntohs(dh->type)) {
					case HTLS_DATA_UID:
						dh_getint(uid);
						htlc->uid = uid;
						break;
				}
			dh_end()
			task_new(htlc, rcv_task_user_list, chat_list);
			hlwrite(htlc, HTLC_HDR_USER_GETLIST, 0, 0);
		}
#ifdef CONFIG_HOPE
	}
#endif
}

static struct option server_opts[] = {
	{"login",	1, 0,	'l'},
	{"password",	1, 0,	'p'},
	{"nickname",	1, 0,	'n'},
	{"icon",	1, 0,	'i'},
	{"old", 	0, 0,	'o'},
	{"secure",	0, 0,	's'},
	{0, 0, 0, 0}
};

COMMAND(server)
{
	int s;
	struct sockaddr_in saddr;
	u_int16_t port = 0, icon = 0;
	char buf[16];
	char *serverstr = 0, *portstr = 0, *login = 0, *pass = 0, *name = 0;
	struct opt_r opt;
	int o, longind;
	u_int16_t icon16;
	int secure = 1;

	opt.err_printf = hx_printf;
	opt.ind = 0;
	while ((o = getopt_long_r(argc, argv, "l:p:n:i:os", server_opts, &longind, &opt)) != EOF) {
		if (o == 0)
			o = server_opts[longind].val;
		switch (o) {
			case 'l':
				login = opt.arg;
				break;
			case 'p':
				pass = opt.arg;
				break;
			case 'n':
				name = opt.arg;
				break;
			case 'i':
				icon = atou16(opt.arg);
				break;
			case 'o':
				secure = 0;
				break;
			case 's':
				secure = 1;
				break;
			default:
				goto usage;
		}
	}
	if (opt.ind < argc) {
		serverstr = argv[opt.ind];
		if (opt.ind + 1 < argc)
			portstr = argv[opt.ind + 1];
		else {
			char *p = strchr(serverstr, ':');
			if (p) {
				*p = 0;
				portstr = p + 1;
			}
		}
	}
	if (!serverstr) {
usage:
		hx_printf(INFOPREFIX "usage: %s [OPTIONS] <server address>[:][port]\n"
				"  -l, --login <login>\n"
				"  -p, --password <password>\n"
				"  -n, --nickname <nickname>\n"
				"  -i, --icon <icon>\n"
				"  -o, --old  [unsecure]\n"
				"  -s, --secure\n"
				, argv[0]);
		return;
	}

	if (portstr)
		port = atou16(portstr);
	if (!port)
		port = HTLS_TCPPORT;

	s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (s < 0) {
		hx_printf(INFOPREFIX "socket: %s\n", strerror(errno));
		return;
	}
	if (s >= hxd_open_max) {
		hx_printf(INFOPREFIX "%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, s, hxd_open_max);
		close(s);
		return;
	}

	if (hx_hostname[0]) {
		if (!inet_aton(hx_hostname, &saddr.sin_addr)) {
			struct hostent *he;

			if ((he = gethostbyname(hx_hostname))) {
				size_t len = (unsigned)he->h_length > sizeof(struct in_addr)
					     ? sizeof(struct in_addr) : he->h_length;
				memcpy(&saddr.sin_addr, he->h_addr, len);
			} else {
#ifndef HAVE_HSTRERROR
				hx_printf(INFOPREFIX "DNS lookup for %s failed\n", hx_hostname);
#else
				hx_printf(INFOPREFIX "DNS lookup for %s failed: %s\n", hx_hostname, hstrerror(h_errno));
#endif
				return;
			}
		}
		saddr.sin_port = 0;
		saddr.sin_family = AF_INET;
		if (bind(s, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
			inet_ntoa_r(saddr.sin_addr, buf, sizeof(buf));
			hx_printf(INFOPREFIX "bind %s (%s): %s\n", hx_hostname, buf, strerror(errno));
			close(s);
			return;
		}
	}

	if (!inet_aton(serverstr, &saddr.sin_addr)) {
		struct hostent *he;

		if ((he = gethostbyname(serverstr))) {
			size_t len = (unsigned)he->h_length > sizeof(struct in_addr)
				     ? sizeof(struct in_addr) : he->h_length;
			memcpy(&saddr.sin_addr, he->h_addr, len);
		} else {
#ifndef HAVE_HSTRERROR
			hx_printf(INFOPREFIX "DNS lookup for %s failed\n", serverstr);
#else
			hx_printf(INFOPREFIX "DNS lookup for %s failed: %s\n", serverstr, hstrerror(h_errno));
#endif
			return;
		}
	}
	saddr.sin_port = htons(port);
	saddr.sin_family = AF_INET;

	if (htlc->fd)
		hx_htlc_close(htlc);
	if (connect(s, (struct sockaddr *)&saddr, sizeof(saddr))) {
		hx_printf(INFOPREFIX "connect: %s\n", strerror(errno));
		close(s);
		return;
	}
	fd_blocking(s, 0);
	fd_closeonexec(s, 1);

	inet_ntoa_r(saddr.sin_addr, buf, sizeof(buf));
	hx_printf(INFOPREFIX "connected to %s:%u\n", buf, ntohs(saddr.sin_port));
	htlc->sockaddr = saddr;
	hx_output.status();

	hxd_files[s].ready_read = htlc_read;
	hxd_files[s].ready_write = htlc_write;
	hxd_files[s].conn.htlc = htlc;

	if (name)
		strcpy(htlc->name, name);
	if (icon)
		htlc->icon = icon;
	htlc->fd = s;
	htlc->rcv = hx_rcv_magic;
	htlc->trans = 1;
	memset(&htlc->in, 0, sizeof(struct qbuf));
	memset(&htlc->out, 0, sizeof(struct qbuf));
	qbuf_set(&htlc->in, 0, HTLS_MAGIC_LEN);
	{
		int r = write(s, HTLC_MAGIC, HTLC_MAGIC_LEN);
		if (r != HTLC_MAGIC_LEN) {
			if (r < 0)
				r = 0;
			qbuf_add(&htlc->out, HTLC_MAGIC + r, HTLC_MAGIC_LEN - r);
			FD_SET(s, &hxd_wfds);
		}
	}
	FD_SET(s, &hxd_rfds);
	if (high_fd < s)
		high_fd = s;

	if (login)
		strcpy(htlc->login, login);

#ifdef CONFIG_HOPE
	if (secure) {
		u_int8_t macalglist[64];
		u_int16_t macalglistlen;

		buf[0] = 0;
		task_new(htlc, rcv_task_login, pass ? xstrdup(pass) : xstrdup(buf));
		strcpy(htlc->macalg, "HMAC-SHA1");
		S16HTON(2, macalglist);
		macalglistlen = 2;
		macalglist[macalglistlen] = 9;
		macalglistlen++;
		memcpy(macalglist+macalglistlen, htlc->macalg, 9);
		macalglistlen += 9;
		macalglist[macalglistlen] = 8;
		macalglistlen++;
		memcpy(macalglist+macalglistlen, "HMAC-MD5", 8);
		macalglistlen += 8;
		hlwrite(htlc, HTLC_HDR_LOGIN, 0, 3,
			HTLC_DATA_LOGIN, 1, buf,
			HTLC_DATA_PASSWORD, 1, buf,
			HTLC_DATA_MAC_ALG, macalglistlen, macalglist,
			HTLC_DATA_SESSIONKEY, 0, 0);
		return;
	}
#endif

	task_new(htlc, rcv_task_login, 0);

	icon16 = htons(htlc->icon);
	if (login)
		hl_encode(login, login, strlen(login));
	if (pass) {
		hl_encode(pass, pass, strlen(pass));
		hlwrite(htlc, HTLC_HDR_LOGIN, 0, 4,
			HTLC_DATA_ICON, 2, &icon16,
			HTLC_DATA_LOGIN, login ? strlen(login) : 0, login,
			HTLC_DATA_PASSWORD, strlen(pass), pass,
			HTLC_DATA_NAME, strlen(htlc->name), htlc->name);
	} else {
		hlwrite(htlc, HTLC_HDR_LOGIN, 0, 3,
			HTLC_DATA_ICON, 2, &icon16,
			HTLC_DATA_LOGIN, login ? strlen(login) : 0, login,
			HTLC_DATA_NAME, strlen(htlc->name), htlc->name);
	}
}

static void
news_display (void)
{
	if (news_buf)
		hx_printf("%s\n", news_buf);
}

static void
rcv_task_news_file (struct htlc_conn *htlc)
{
	dh_start(htlc)
		if (ntohs(dh->type) != HTLS_DATA_NEWS)
			continue;
		news_len = ntohs(dh->len);
		news_buf = xrealloc(news_buf, news_len + 1);
		memcpy(news_buf, dh->data, news_len);
		CR2LF(news_buf, news_len);
		strip_ansi(news_buf, news_len);
		news_buf[news_len] = 0;
	dh_end()
	news_display();
}

COMMAND(news)
{
	if ((argv[1] && ((argv[1][0] == '-' && argv[1][1] == 'r') || !strcmp(argv[1], "--reload")))
	    || !news_len) {
		task_new(htlc, rcv_task_news_file, 0);
		hlwrite(htlc, HTLC_HDR_NEWS_GETFILE, 0, 0);
	} else {
		news_display();
	}
}

static void
rcv_task_user_list (struct htlc_conn *htlc, struct chat *chat)
{
	struct hl_userlist_hdr *uh;
	struct user *user;
	u_int16_t nlen;

	dh_start(htlc)
		if (ntohs(dh->type) == HTLS_DATA_USER_LIST) {
			uh = (struct hl_userlist_hdr *)dh;
			user = user_with_uid(chat->user_list, ntohs(uh->uid));
			if (!user) {
				user = user_new(&chat->user_tail);
				chat->nusers++;
			}
			user->uid = ntohs(uh->uid);
			user->icon = ntohs(uh->icon);
			user->color = ntohs(uh->color);
			nlen = ntohs(uh->nlen) > 31 ? 31 : ntohs(uh->nlen);
			memcpy(user->name, uh->name, nlen);
			user->name[nlen] = 0;
			if (!htlc->uid && !strcmp(user->name, htlc->name) && user->icon == htlc->icon) {
				htlc->uid = user->uid;
				htlc->color = user->color;
				hx_output.status();
			}
		} else if (ntohs(dh->type) == HTLS_DATA_CHAT_SUBJECT) {
			u_int16_t slen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
			memcpy(chat->subject, dh->data, slen);
			chat->subject[slen] = 0;
			hx_printf(INFOPREFIX "chat 0x%08x subject: %s\n", chat->cid, chat->subject);
		}
	dh_end()
	user_print(chat->user_list, 0);
}

static void
rcv_task_user_list_switch (struct htlc_conn *htlc, struct chat *chat)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;

	if ((ntohs(h->flag) & 1)) {
		chat_delete(chat);
		return;
	}
	rcv_task_user_list(htlc, chat);
	chat_front = chat;
}

static struct option who_opts[] = {
	{"chat",	0, 0,	'c'},
	{"reload",	0, 0,	'r'},
	{0, 0, 0, 0}
};

COMMAND(who)
{
	struct chat *chat;
	u_int32_t cid = 0;
	int reload = 0;
	struct opt_r opt;
	int o, longind;

	opt.err_printf = hx_printf;
	opt.ind = 0;
	while ((o = getopt_long_r(argc, argv, "c:r", who_opts, &longind, &opt)) != EOF) {
		if (o == 0)
			o = server_opts[longind].val;
		switch (o) {
			case 'c':
				cid = atou32(opt.arg);
				break;
			case 'r':
				reload = 1;
				break;
		}
	}
	if (!cid && !reload)
		reload = chat_list->user_tail == chat_list->user_list;
	if (reload) {
		task_new(htlc, rcv_task_user_list, chat_list);
		hlwrite(htlc, HTLC_HDR_USER_GETLIST, 0, 0);
	} else {
		chat = chat_with_cid(cid);
		if (!chat)
			hx_printf(INFOPREFIX "no such chat 0x%08x\n", cid);
		else
			user_print(chat->user_list, argv[1]);
	}
}

COMMAND(me)
{
	char *p;
	u_int16_t style;

	for (p = str; *p && *p != ' '; p++) ;
	if (!*p || !(*++p))
		return;
	style = htons(1);
	if (chat_front->cid) {
		u_int32_t cid = htonl(chat_front->cid);
		hlwrite(htlc, HTLC_HDR_CHAT, 0, 3,
			HTLC_DATA_STYLE, 2, &style,
			HTLC_DATA_CHAT, strlen(p), p,
			HTLC_DATA_CHAT_ID, 4, &cid);
	} else
		hlwrite(htlc, HTLC_HDR_CHAT, 0, 2,
			HTLC_DATA_STYLE, 2, &style,
			HTLC_DATA_CHAT, strlen(p), p);
}

COMMAND(chats)
{
	struct chat *chatp;

	hx_output.mode_underline();
	hx_printf("        cid | nusers | subject\n");
	hx_output.mode_clear();
	for (chatp = chat_list; chatp; chatp = chatp->next)
		hx_printf(" 0x%08x | %6u | %s\n", chatp->cid, chatp->nusers, chatp->subject);
}

COMMAND(chat)
{
	u_int32_t uid;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <uid>\n", argv[0]);
		return;
	}
	uid = atou32(argv[1]);
	if (!uid) {
		struct user *user = user_with_name(chat_list->user_list, argv[1]);
		if (!user) {
			hx_printf(INFOPREFIX "%s: no such nickname %s\n", argv[0], argv[1]);
			return;
		}
		uid = user->uid;
	}
	uid = htonl(uid);
	task_new(htlc, hx_rcv_user_change, 0);
	hlwrite(htlc, HTLC_HDR_CHAT_CREATE, 0, 1,
		HTLC_DATA_UID, 4, &uid);
}

COMMAND(invite)
{
	u_int32_t uid, cid;

	if (argc < 3) {
		hx_printf(INFOPREFIX "usage: %s <chat> <uid>\n", argv[0]);
		return;
	}
	cid = atou32(argv[1]);
	uid = atou32(argv[2]);
	if (!uid) {
		struct user *user = user_with_name(chat_list->user_list, argv[2]);
		if (!user) {
			hx_printf(INFOPREFIX "%s: no such nickname\n", argv[2]);
			return;
		}
		uid = user->uid;
	}
	cid = htonl(cid);
	uid = htonl(uid);
	task_new(htlc, 0, 0);
	hlwrite(htlc, HTLC_HDR_CHAT_INVITE, 0, 2,
		HTLC_DATA_CHAT_ID, 4, &cid,
		HTLC_DATA_UID, 4, &uid);
}

COMMAND(join)
{
	u_int32_t cid;
	struct chat *chat;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <chat>\n", argv[0]);
		return;
	}
	cid = atou32(argv[1]);
	chat = chat_with_cid(cid);
	if (chat)
		chat_front = chat;
	else {
		chat = chat_new(cid);
		cid = htonl(cid);
		task_new(htlc, rcv_task_user_list_switch, chat);
		hlwrite(htlc, HTLC_HDR_CHAT_JOIN, 0, 1,
			HTLC_DATA_CHAT_ID, 4, &cid);
	}
}

COMMAND(part)
{
	u_int32_t cid;
	struct chat *chat;

	if (argv[1])
		chat = chat_with_cid(atou32(argv[1]));
	else
		chat = chat_front;
	if (!chat) {
		hx_printf(INFOPREFIX "%s: no such chat %s\n", argv[0], argv[1] ? argv[1] : "(front)");
		return;
	}
	cid = htonl(chat->cid);
	hlwrite(htlc, HTLC_HDR_CHAT_PART, 0, 1,
		HTLC_DATA_CHAT_ID, 4, &cid);
	if (chat_front == chat_list) {
		hx_printf(INFOPREFIX "not can part from server chat\n");
		return;
	}
	chat_delete(chat);
}

COMMAND(subject)
{
	u_int32_t cid;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <subject>\n", argv[0]);
		return;
	}
	if (chat_front) {
		cid = htonl(chat_front->cid);
		hlwrite(htlc, HTLC_HDR_CHAT_SUBJECT, 0, 2,
			HTLC_DATA_CHAT_ID, 4, &cid,
			HTLC_DATA_CHAT_SUBJECT, strlen(argv[1]), argv[1]);
	}
}

static void
rcv_task_kick (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;

	if ((ntohs(h->flag) & 1))
		return;
	hx_printf(INFOPREFIX "kick successful\n");
}

COMMAND(kick)
{
	u_int32_t uid;
	u_int16_t ban_his_ass;
	struct opt_r opt;
	int i, ban = 0;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <uid>\n", argv[0]);
		return;
	}
	opt.err_printf = hx_printf;
	opt.ind = 0;
	while ((i = getopt_r(argc, argv, "b", &opt)) != EOF) {
		if (i == 'b')
			ban = 1;
	}
	for (i = opt.ind; i < argc; i++) {
		uid = atou32(argv[i]);
		if (!uid) {
			struct user *user = user_with_name(chat_list->user_list, argv[i]);
			if (!user) {
				hx_printf(INFOPREFIX "%s: no such nickname %s\n", argv[0], argv[i]);
				return;
			}
			uid = user->uid;
		}
		uid = htonl(uid);
		ban_his_ass = htons(1);
		task_new(htlc, rcv_task_kick, 0);
		if (ban)
			hlwrite(htlc, HTLC_HDR_USER_KICK, 0, 2,
				HTLC_DATA_BAN, 2, &ban_his_ass,
				HTLC_DATA_UID, 4, &uid);
		else
			hlwrite(htlc, HTLC_HDR_USER_KICK, 0, 1,
				HTLC_DATA_UID, 4, &uid);
	}
}

COMMAND(post)
{
	char *p;

	for (p = str; *p && *p != ' '; p++) ;
	if (!*p || !(*++p))
		return;
	task_new(htlc, 0, 0);
	hlwrite(htlc, HTLC_HDR_NEWS_POST, 0, 1,
		HTLC_DATA_NEWS_POST, strlen(p), p);
}

COMMAND(close)
{
	if (htlc->fd)
		hx_htlc_close(htlc);
}

static void
rcv_task_user_info (struct htlc_conn *htlc)
{
	u_int16_t ilen = 0, nlen = 0;
	u_int8_t info[4096 + 1], name[32];

	name[0] = 0;
	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_USER_INFO:
				ilen = ntohs(dh->len) > 4096 ? 4096 : ntohs(dh->len);
				memcpy(info, dh->data, ilen);
				info[ilen] = 0;
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				strip_ansi(name, nlen);
				break;
		}
	dh_end()
	if (nlen)
		hx_printf(INFOPREFIX "info for %s\n", name);
	if (ilen) {
		CR2LF(info, ilen);
		strip_ansi(info, ilen);
		hx_printf("%s\n", info);
	}
}

COMMAND(info)
{
	u_int32_t uid;
	int i;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <uid>\n", argv[0]);
		return;
	}
	for (i = 1; i < argc; i++) { 
		uid = atou32(argv[i]);
		if (!uid) {
			struct user *user = user_with_name(chat_list->user_list, argv[i]);
			if (!user) {
				hx_printf(INFOPREFIX "%s: no such nickname %s\n", argv[0], argv[i]);
				return;
			}
			uid = user->uid;
		}
		uid = htonl(uid);
		task_new(htlc, rcv_task_user_info, 0);
		hlwrite(htlc, HTLC_HDR_USER_GETINFO, 0, 1,
			HTLC_DATA_UID, 4, &uid);
	}
}

COMMAND(load)
{
	char path[MAXPATHLEN], buf[4096];
	int i, fd;
	size_t j;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <file1> [file2...]\n", argv[0]);
		return;
	}
	for (i = 1; i < argc; i++) {
		expand_tilde(path, argv[i]);
		if ((fd = open(path, O_RDONLY)) < 0) {
			hx_printf(INFOPREFIX "%s: %s: %s\n", argv[0], argv[i], strerror(errno));
			continue;
		}
		j = 0;
		while (j < sizeof(buf) && read(fd, &buf[j], 1) > 0) {
			if (buf[j] == '\n') {
				if (j) {
					buf[j] = 0;
					j = 0;
					hx_command(buf);
				}
			} else
				j++;
		}
		close(fd);
	}
}

void
hx_load (char *file)
{
	int argc = 2;
	char *argv[3];

	argv[0] = "load";
	argv[1] = file;
	argv[2] = 0;
	cmd_load(argc, argv, 0, &hx_htlc);
}

COMMAND(type)
{
#define MAXCHATSIZ	2048
	char path[MAXPATHLEN], buf[4096];
	struct opt_r opt;
	int i, fd;
	size_t j;
	u_int32_t cid = 0, uid = 0, news = 0;
	struct user *user = 0;
	char *agin_buf;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s [-n] [-c cid] [-u uid] <file1> [file2...]\n", argv[0]);
		return;
	}
	opt.err_printf = hx_printf;
	opt.ind = 0;
	cid = htonl(chat_front->cid);
	while ((i = getopt_r(argc, argv, "c:u:n", &opt)) != EOF) {
		switch (i) {
			case 'c':
				cid = htonl(atou32(opt.arg));
				break;
			case 'n':
				news = 1;
				break;
			case 'u':
				uid = atou32(opt.arg);
				if (!uid) {
					user = user_with_name(chat_list->user_list, opt.arg);
					if (!user) {
						hx_printf(INFOPREFIX "%s: no such nickname %s\n", argv[0], opt.arg);
						return;
					}
					uid = user->uid;
				}
				break;
		}
	}
	for (i = opt.ind; i < argc; i++) {
		expand_tilde(path, argv[i]);
		if ((fd = open(path, O_RDONLY)) < 0) {
			hx_printf(INFOPREFIX "%s: %s: %s\n", argv[0], argv[i], strerror(errno));
			continue;
		}
		agin_buf = 0;
agin:
		if (agin_buf) {
			j = MAXCHATSIZ - (agin_buf - buf);
			memcpy(buf, agin_buf, j);
			agin_buf = 0;
		} else
			j = 0;
		while (j < MAXCHATSIZ && read(fd, &buf[j], 1) > 0) {
			if (buf[j] == '\n')
				buf[j] = '\r';
			j++;
		}
		if (j == 0)
			continue;
		if (j == MAXCHATSIZ) {
			char *p = strrchr(buf, '\r');
			if (p) {
				j = p - buf;
				p++;
				agin_buf = p;
			} else {
				agin_buf = 0;
			}
		}
		if (buf[j - 1] == '\r')
			j--;
		if (news) {
			task_new(htlc, 0, 0);
			hlwrite(htlc, HTLC_HDR_NEWS_POST, 0, 1,
				HTLC_DATA_NEWS_POST, j, buf);
		} else if (uid) {
			size_t buflen;
			char *msg_buf;

			if (!user)
				user = user_with_uid(chat_list->user_list, uid);
			if (user) {
				char *col = user->color % 4 > 1 ? (user->color % 2 ? RED : RED_BOLD)
							      : (user->color % 2 ? GREEN : GREEN_BOLD);
				buflen = strlen(path) + strlen(user->name) + 32 + strlen(col) + strlen(DEFAULT);
				msg_buf = xmalloc(buflen);
				snprintf(msg_buf, buflen, "[%s%s%s(%u)]->file> %s", col, user->name, DEFAULT, uid, path);
			} else {
				buflen = strlen(path) + 32;
				msg_buf = xmalloc(buflen);
				snprintf(msg_buf, buflen, "[(%u)]->file> %s", uid, path);
			}
			uid = htonl(uid);
			task_new(htlc, rcv_task_msg, msg_buf);
			hlwrite(htlc, HTLC_HDR_MSG, 0, 2,
				HTLC_DATA_UID, 4, &uid,
				HTLC_DATA_MSG, j, buf);
		} else if (cid)
			hlwrite(htlc, HTLC_HDR_CHAT, 0, 2,
				HTLC_DATA_CHAT, j, buf,
				HTLC_DATA_CHAT_ID, 4, &cid);
		else
			hlwrite(htlc, HTLC_HDR_CHAT, 0, 1,
				HTLC_DATA_CHAT, j, buf);
		if (agin_buf)
			goto agin;
		close(fd);
	}
}

struct x_fhdr {
	u_int16_t enc;
	u_int8_t len, name[1];
};

static u_int8_t *
path_to_hldir (const char *path, u_int16_t *hldirlen, int is_file)
{
	u_int8_t *hldir;
	struct x_fhdr *fh;
	char const *p, *p2;
	u_int16_t pos = 2, dc = 0;
	u_int8_t nlen;

	hldir = xmalloc(2);
	p = path;
	while ((p2 = strchr(p, dir_char))) {
		if (!(p2 - p)) {
			p++;
			continue;
		}
		nlen = (u_int8_t)(p2 - p);
		pos += 3 + nlen;
		hldir = xrealloc(hldir, pos);
		fh = (struct x_fhdr *)(&(hldir[pos - (3 + nlen)]));
		memset(&fh->enc, 0, 2);
		fh->len = nlen;
		memcpy(fh->name, p, nlen);
		dc++;
		p = p2 + 1;
	}
	if (!is_file && *p) {
		nlen = (u_int8_t)strlen(p);
		pos += 3 + nlen;
		hldir = xrealloc(hldir, pos);
		fh = (struct x_fhdr *)(&(hldir[pos - (3 + nlen)]));
		memset(&fh->enc, 0, 2);
		fh->len = nlen;
		memcpy(fh->name, p, nlen);
		dc++;
	}
	*((u_int16_t *)hldir) = htons(dc);

	*hldirlen = pos;
	return hldir;
}

struct cached_filelist {
	struct cached_filelist *next, *prev;
	char *path;
	struct hl_filelist_hdr *fh;
	u_int32_t fhlen;
	int completing;
	char **filter_argv;
};

static struct cached_filelist __cfl_list = {0, 0, 0, 0, 0, 0, 0}, *cfl_list = &__cfl_list, *cfl_tail = &__cfl_list;

struct cached_filelist *
cfl_lookup (char *path)
{
	struct cached_filelist *cfl;

	for (cfl = cfl_list->next; cfl; cfl = cfl->next)
		if (!strcmp(cfl->path, path))
			return cfl;

	cfl = xmalloc(sizeof(struct cached_filelist));
	memset(cfl, 0, sizeof(struct cached_filelist));

	cfl->next = 0;
	cfl->prev = cfl_tail;
	cfl_tail->next = cfl;
	cfl_tail = cfl;

	return cfl;
}

static void
cfl_delete (struct cached_filelist *cfl)
{
	if (cfl->path)
		xfree(cfl->path);
	if (cfl->fh)
		xfree(cfl->fh);
	if (cfl->next)
		cfl->next->prev = cfl->prev;
	if (cfl->prev)
		cfl->prev->next = cfl->next;
	if (cfl_tail == cfl)
		cfl_tail = cfl->prev;
	xfree(cfl);
}

COMMAND(dirchar)
{
	struct cached_filelist *cfl;
	char *p, old_dirchar = dir_char;

	if (!argv[1]) {
		hx_printf(INFOPREFIX "%c\n", dir_char);
		return;
	}
	dir_char = argv[1][0];
	for (p = htlc->rootdir; *p; p++)
		if (*p == old_dirchar)
			*p = dir_char;
	for (cfl = cfl_list->next; cfl; cfl = cfl->next)
		for (p = cfl->path; p && *p; p++)
			if (*p == old_dirchar)
				*p = dir_char;
}

static char *
dirchar_basename (char *path)
{
	size_t len;

	len = strlen(path);
	while (len--) {
		if (path[len] == dir_char)
			return path+len+1;
	}

	return path;
}

static void
fh_print (struct hl_filelist_hdr *fh)
{
	char buf[4096];
	u_int16_t i, bpos;
	u_int32_t fnlen;

	fnlen = ntohl(fh->fnlen);
	for (i = 0, bpos = 0; i < fnlen && bpos < sizeof(buf); i++) {
		if (!isgraph(fh->fname[i])) {
			buf[bpos++] = '\\';
			switch (fh->fname[i]) {
				case ' ':
					buf[bpos++] = ' ';
					break;
				case '\r':
					buf[bpos++] = 'r';
					break;
				case '\n':
					buf[bpos++] = 'n';
					break;
				case '\t':
					buf[bpos++] = 't';
					break;
				default:
					sprintf(&(buf[bpos]), "x%2.2x", fh->fname[i]);
					bpos += 3;
			}
		} else
			buf[bpos++] = fh->fname[i];
	}
	buf[bpos] = 0;
	hx_printf("%4.4s/%4.4s %10u %s\n",
		(char *)&fh->fcreator, (char *)&fh->ftype, ntohl(fh->fsize), buf);
}

#define XFER_GET	0
#define XFER_PUT	1

static struct htxf_conn *xfer_new (const char *path, const char *remotepath, u_int16_t type);

static inline void
dirchar_fix (char *lpath)
{
	char *p;

	for (p = lpath; *p; p++)
		if (*p == '/')
			*p = (dir_char == '/' ? ':' : dir_char);
}

static inline void
dirmask (char *dst, char *src, char *mask)
{
	while (*mask && *src && *mask++ == *src++) ;

	strcpy(dst, src);
}

#define COMPLETE_NONE	0
#define COMPLETE_EXPAND	1
#define COMPLETE_LS_R	2
#define COMPLETE_GET_R	3

static void
rcv_task_file_list (struct htlc_conn *htlc, struct cached_filelist *cfl)
{
	struct hl_filelist_hdr *fh = 0;
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	u_int32_t fh_len;

	if ((ntohs(h->flag) & 1)) {
		cfl_delete(cfl);
		return;
	}
	if (cfl->completing == COMPLETE_NONE || cfl->completing == COMPLETE_LS_R)
		hx_printf(INFOPREFIX "%s:\n", cfl->path);
	dh_start(htlc)
		if (ntohs(dh->type) != HTLS_DATA_FILE_LIST)
			continue;
		fh = (struct hl_filelist_hdr *)dh;
		if (cfl->completing == COMPLETE_NONE || cfl->completing == COMPLETE_LS_R)
			fh_print(fh);
		if (cfl->completing > 1) {
			char *pathbuf;
			int fnlen, len;

			fnlen = ntohl(fh->fnlen);
			len = strlen(cfl->path) + 1 + fnlen + 1;
			pathbuf = xmalloc(len);
			snprintf(pathbuf, len, "%s%c%.*s", cfl->path[1] ? cfl->path : "", dir_char, (int)fnlen, fh->fname);
			if (ntohl(fh->ftype) == 0x666c6472) {
				struct cached_filelist *ncfl;
				u_int16_t hldirlen;
				u_int8_t *hldir;

				ncfl = cfl_lookup(pathbuf);
				ncfl->completing = cfl->completing;
				ncfl->filter_argv = cfl->filter_argv;
				if (!ncfl->path)
					ncfl->path = xstrdup(pathbuf);
				hldir = path_to_hldir(pathbuf, &hldirlen, 0);
				task_new(&hx_htlc, rcv_task_file_list, ncfl);
				hlwrite(&hx_htlc, HTLC_HDR_FILE_LIST, 0, 1,
					HTLC_DATA_DIR, hldirlen, hldir);
				xfree(hldir);
			} else if (cfl->completing == COMPLETE_GET_R) {
				struct htxf_conn *htxf;
				char *lpath, *p;

				lpath = xmalloc(len);
				dirmask(lpath, pathbuf, htlc->rootdir);
				p = lpath+1;
				while ((p = strchr(p, dir_char))) {
					*p = 0;
					if (mkdir(lpath+1, S_IRUSR|S_IWUSR|S_IXUSR)) {
						if (errno != EEXIST)
							hx_printf(INFOPREFIX "mkdir(%s): %s\n", lpath+1, strerror(errno));
					}
					*p++ = '/';
					while (*p == dir_char)
						*p++ = '/';
				}
				p = basename(lpath+1);
				if (p)
					dirchar_fix(p);
				htxf = xfer_new(lpath+1, pathbuf, XFER_GET);
				htxf->filter_argv = cfl->filter_argv;
				xfree(lpath);
			}
			xfree(pathbuf);
		}
		fh_len = SIZEOF_HL_DATA_HDR + ntohs(fh->len);
		fh_len += 4 - (fh_len % 4);
		cfl->fh = xrealloc(cfl->fh, cfl->fhlen + fh_len);
		memcpy((char *)cfl->fh + cfl->fhlen, fh, SIZEOF_HL_DATA_HDR + ntohs(fh->len));
		*((u_int16_t *)((char *)cfl->fh + cfl->fhlen + 2)) = htons(fh_len - SIZEOF_HL_DATA_HDR);
		cfl->fhlen += fh_len;
	dh_end()
	cfl->completing = COMPLETE_NONE;
}

static struct cached_filelist *last_cfl = 0;

int
expand_path (char *path, int len, char *pathbuf)
{
	struct cached_filelist *cfl;
	struct hl_filelist_hdr *fh;
	int fnlen, blen = 0, eplen, epchr, ambig, r, ambigi = 0;
	char *p, *ent, *ep, buf[MAXPATHLEN], ambigbuf[4096], *ambigbufp = ambigbuf;

	if (*path != dir_char) {
		len = snprintf(buf, MAXPATHLEN, "%s%c%.*s", hx_htlc.rootdir, dir_char, len, path);
		if (len < 0)
			len = MAXPATHLEN;
		path = buf;
	}
	ent = path;
	for (p = path + len - 1; p >= path; p--)
		if (*p == dir_char) {
			ent = p+1;
			while (p > path && *p == dir_char)
				p--;
			blen = (p+1) - path;
			len -= ent - path;
			break;
		}
	if (!*ent)
		return -1;

	for (cfl = cfl_list->next; cfl; cfl = cfl->next) {
		if (!strncmp(cfl->path, path, blen))
			break;
	}
	if (!cfl) {
		u_int16_t hldirlen;
		u_int8_t *hldir;
		char xxxbuf[MAXPATHLEN];
		int i;

		snprintf(xxxbuf, MAXPATHLEN, "%.*s", blen, path);
		path = xxxbuf;
		i = strlen(path);
		while (i > 1 && path[--i] == dir_char)
			path[i] = 0;
		cfl = cfl_lookup(path);
		cfl->completing = COMPLETE_EXPAND;
		cfl->path = xstrdup(path);
		hldir = path_to_hldir(path, &hldirlen, 0);
		task_new(&hx_htlc, rcv_task_file_list, cfl);
		hlwrite(&hx_htlc, HTLC_HDR_FILE_LIST, 0, 1,
			HTLC_DATA_DIR, hldirlen, hldir);
		xfree(hldir);
		return -2;
	}
	if (!cfl->fh)
		return -1;
	ep = 0;
	eplen = 0;
	epchr = 0;
	ambig = 0;
	for (fh = cfl->fh; (u_int32_t)((char *)fh - (char *)cfl->fh) < cfl->fhlen; (char *)fh += ntohs(fh->len) + SIZEOF_HL_DATA_HDR) {
		fnlen = ntohl(fh->fnlen);
		if (fnlen >= len && !strncmp(fh->fname, ent, len)) {
			if (ep) {
				int i;
	
				if (fnlen < eplen)
					eplen = fnlen;
				for (i = 0; i < eplen; i++)
					if (fh->fname[i] != ep[i])
						break;
				eplen = i;
				ambig = 1;
				epchr = 0;
xxx:
				if (last_cfl == cfl) {
					r = snprintf(ambigbufp, sizeof(ambigbuf) - (ambigbufp - ambigbuf), "  %-16.*s", fnlen, fh->fname);
					if (!(r == -1 || sizeof(ambigbuf) <= (unsigned)(ambigbufp + r - ambigbuf))) {
						ambigi++;
						ambigbufp += r;
						if (!(ambigi % 5))
							*ambigbufp++ = '\n';
					}
				}
			} else {
				eplen = fnlen;
				epchr = ntohl(fh->ftype) == 0x666c6472 ? dir_char : 0;
				if (epchr)
					ambig = 1;
				goto xxx;
			}
			ep = fh->fname;
		}
	}
	if (!ep)
		return -1;
	if (ambigi > 1)
		hx_printf("%s%c", ambigbuf, (ambigi % 5) ? '\n' : 0);
	else
		last_cfl = cfl;

	snprintf(pathbuf, MAXPATHLEN, "%.*s%c%.*s%c", path[blen-1] == dir_char ? blen - 1 : blen, path, dir_char, eplen, ep, epchr);

	return ambig;
}

char **
glob_remote (char *path, int *npaths)
{
	struct cached_filelist *cfl;
	struct hl_filelist_hdr *fh;
	char *p, *ent, *patternbuf, *pathbuf, **paths;
	int n, len, blen = 0;

	ent = path;
	len = strlen(path);
	for (p = path + len - 1; p >= path; p--)
		if (*p == dir_char) {
			ent = p+1;
			while (p > path && *p == dir_char)
				p--;
			blen = (p+1) - path;
			break;
		}

	patternbuf = xmalloc(blen + 1);
	memcpy(patternbuf, path, blen);
	patternbuf[blen] = 0;

	paths = 0;
	n = 0;
	for (cfl = cfl_list->next; cfl; cfl = cfl->next) {
		u_int32_t fnlen;
		if (fnmatch(patternbuf, cfl->path, FNM_NOESCAPE))
			continue;

		for (fh = cfl->fh; (u_int32_t)((char *)fh - (char *)cfl->fh) < cfl->fhlen; (char *)fh += ntohs(fh->len) + SIZEOF_HL_DATA_HDR) {
			fnlen = ntohl(fh->fnlen);
			len = strlen(cfl->path) + 1 + fnlen + 1;
			pathbuf = xmalloc(len);
			snprintf(pathbuf, len, "%s%c%.*s", cfl->path[1] ? cfl->path : "", dir_char, (int)fnlen, fh->fname);
			if (!fnmatch(path, pathbuf, FNM_NOESCAPE)) {
				paths = xrealloc(paths, (n + 1) * sizeof(char *));
				paths[n] = pathbuf;
				n++;
			} else {
				xfree(pathbuf);
			}
		}
	}
	if (n)
		goto ret;

	paths = xmalloc(sizeof(char *));
	n = 1;
	paths[0] = xstrdup(path);

ret:
	*npaths = n;

	return paths;
}

int
exists_remote (char *path)
{
	struct cached_filelist *cfl;
	struct hl_filelist_hdr *fh;
	char *p, *ent, buf[MAXPATHLEN];
	u_int32_t fnlen;
	int blen = 0, len;

	len = strlen(path);
	if (*path != dir_char) {
		len = snprintf(buf, MAXPATHLEN, "%s%c%.*s", hx_htlc.rootdir, dir_char, len, path);
		if (len < 0)
			len = MAXPATHLEN;
		path = buf;
	}
	ent = path;
	for (p = path + len - 1; p >= path; p--)
		if (*p == dir_char) {
			ent = p+1;
			while (p > path && *p == dir_char)
				p--;
			blen = (p+1) - path;
			len -= ent - path;
			break;
		}
	if (!*ent)
		return -1;

	for (cfl = cfl_list->next; cfl; cfl = cfl->next)
		if (!strncmp(cfl->path, path, blen))
			break;
	if (!cfl) {
		u_int16_t hldirlen;
		u_int8_t *hldir;

		snprintf(buf, MAXPATHLEN, "%.*s", blen, path);
		path = buf;
		len = strlen(path);
		while (len > 1 && path[--len] == dir_char)
			path[len] = 0;
		cfl = cfl_lookup(path);
		cfl->completing = COMPLETE_EXPAND;
		cfl->path = xstrdup(path);
		hldir = path_to_hldir(path, &hldirlen, 0);
		task_new(&hx_htlc, rcv_task_file_list, cfl);
		hlwrite(&hx_htlc, HTLC_HDR_FILE_LIST, 0, 1,
			HTLC_DATA_DIR, hldirlen, hldir);
		xfree(hldir);
		return 0;
	}
	if (!cfl->fh)
		return 0;

	for (fh = cfl->fh; (u_int32_t)((char *)fh - (char *)cfl->fh) < cfl->fhlen; (char *)fh += ntohs(fh->len) + SIZEOF_HL_DATA_HDR) {
		fnlen = ntohl(fh->fnlen);
		if ((int)fnlen == len && !strncmp(fh->fname, ent, len))
			return 1;
	}

	return 0;
}

COMMAND(ls)
{
	char *path, **paths, buf[MAXPATHLEN];
	u_int16_t hldirlen;
	u_int8_t *hldir;
	int i, j, npaths, reload = 0, recurs = 0;
	struct cached_filelist *cfl;

	npaths = 0;
	for (i = 1; i < argc; i++)
		if (argv[i][0] != '-')
			npaths++;
	if (!npaths) {
		argv[i] = htlc->rootdir;
		argc++;
	}
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			if ((argv[i][1] == 'r' && !argv[i][2])
			    || !strcmp(argv[i], "--reload")) {
				reload = 1;
				if (i == 1 && !argv[2])
					argv[1] = htlc->rootdir;
				else
					continue;
			} else if (argv[i][1] == 'R' && !argv[i][2]) {
				recurs = 1;
				continue;
			}
		}
		path = argv[i];
		if (*path == dir_char)
			strcpy(buf, path);
		else
			snprintf(buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, path);
		path = buf;
		j = strlen(path);
		while (j > 1 && path[--j] == dir_char)
			path[j] = 0;
		paths = glob_remote(path, &npaths);
		for (j = 0; j < npaths; j++) {
			path = paths[j];
			cfl = cfl_lookup(path);
			if (cfl->fh) {
				if (reload) {
					xfree(cfl->fh);
					cfl->fh = 0;
					cfl->fhlen = 0;
				} else {
					struct hl_filelist_hdr *fh;

					hx_printf(INFOPREFIX "%s:\n", cfl->path);
					for (fh = cfl->fh; (u_int32_t)((char *)fh - (char *)cfl->fh) < cfl->fhlen;
							   (char *)fh += ntohs(fh->len) + SIZEOF_HL_DATA_HDR)
						fh_print(fh);
					goto next;
				}
			}
			if (recurs)
				cfl->completing = COMPLETE_LS_R;
			if (!cfl->path)
				cfl->path = xstrdup(path);
			hldir = path_to_hldir(path, &hldirlen, 0);
			task_new(htlc, rcv_task_file_list, cfl);
			hlwrite(htlc, HTLC_HDR_FILE_LIST, 0, 1,
				HTLC_DATA_DIR, hldirlen, hldir);
			xfree(hldir);
next:
			xfree(path);
		}
		xfree(paths);
	}
}

COMMAND(cd)
{
	int len;

	if (argc < 2) {
		hx_printf("usage: %s <directory>\n", argv[0]);
		return;
	}
	if (argv[1][0] == dir_char) {
		len = strlen(argv[1]) > MAXPATHLEN - 1 ? MAXPATHLEN - 1 : strlen(argv[1]);
		memcpy(htlc->rootdir, argv[1], len);
		htlc->rootdir[len] = 0;
	} else {
		char buf[MAXPATHLEN];

		len = snprintf(buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, argv[1]);
		strcpy(htlc->rootdir, buf);
	}
	if (htlc->rootdir[len-1] == dir_char)
		htlc->rootdir[len-1] = 0;
}

COMMAND(pwd)
{
	hx_printf(INFOPREFIX "%s%c\n", htlc->rootdir, dir_char);
}

COMMAND(lcd)
{
	char buf[MAXPATHLEN];

	expand_tilde(buf, argv[1] ? argv[1] : "~");
	if (chdir(buf))
		hx_printf(INFOPREFIX "%s: %s: %s\n", argv[0], buf, strerror(errno));
}

COMMAND(lpwd)
{
	char buf[MAXPATHLEN];

	if (!getcwd(buf, sizeof(buf) - 1))
		hx_printf(INFOPREFIX "%s: getcwd: %s\n", argv[0], strerror(errno));
	else {
		buf[sizeof(buf) - 1] = 0;
		hx_printf(INFOPREFIX "%s\n", buf);
	}
}

COMMAND(mkdir)
{
	char *path, **paths, buf[MAXPATHLEN];
	u_int16_t hldirlen;
	u_int8_t *hldir;
	int i, j, npaths;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <directory> [dir2...]\n", argv[0]);
		return;
	}
	for (i = 1; i < argc; i++) {
		path = argv[i];
		if (*path == dir_char)
			strcpy(buf, path);
		else
			snprintf(buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, path);
		path = buf;
		paths = glob_remote(path, &npaths);
		for (j = 0; j < npaths; j++) {
			path = paths[j];
			hldir = path_to_hldir(path, &hldirlen, 0);
			task_new(htlc, 0, 0);
			hlwrite(htlc, HTLC_HDR_FILE_MKDIR, 0, 1,
				HTLC_DATA_DIR, hldirlen, hldir);
			xfree(hldir);
			xfree(path);
		}
		xfree(paths);
	}
}

COMMAND(rm)
{
	char *path, **paths, *file, buf[MAXPATHLEN];
	u_int16_t hldirlen;
	u_int8_t *hldir;
	int i, j, npaths;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <file> [file2...]\n", argv[0]);
		return;
	}
	for (i = 1; i < argc; i++) {
		path = argv[i];
		if (*path == dir_char)
			strcpy(buf, path);
		else
			snprintf(buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, path);
		path = buf;
		paths = glob_remote(path, &npaths);
		for (j = 0; j < npaths; j++) {
			path = paths[j];
			task_new(htlc, 0, 0);
			file = dirchar_basename(path);
			if (file != path) {
				hldir = path_to_hldir(path, &hldirlen, 1);
				hlwrite(htlc, HTLC_HDR_FILE_DELETE, 0, 2,
					HTLC_DATA_FILE_NAME, strlen(file), file,
					HTLC_DATA_DIR, hldirlen, hldir);
				xfree(hldir);
			} else {
				hlwrite(htlc, HTLC_HDR_FILE_DELETE, 0, 1,
					HTLC_DATA_FILE_NAME, strlen(file), file);
			}
			xfree(path);
		}
		xfree(paths);
	}
}

COMMAND(ln)
{
	char *src_path, *dst_path, *src_file, *dst_file, **paths;
	char src_buf[MAXPATHLEN], dst_buf[MAXPATHLEN];
	u_int16_t hldirlen, rnhldirlen;
	u_int8_t *hldir, *rnhldir;
	int i, j, npaths;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <source file> <dest file>\n", argv[0]);
		return;
	}
	dst_path = argv[--argc];
	if (*dst_path == dir_char)
		strcpy(dst_buf, dst_path);
	else
		snprintf(dst_buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, dst_path);
	dst_path = dst_buf;
	dst_file = dirchar_basename(dst_path);
	for (i = 1; i < argc; i++) {
		src_path = argv[i];
		if (*src_path == dir_char)
			strcpy(src_buf, src_path);
		else
			snprintf(src_buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, src_path);
		src_path = src_buf;
		paths = glob_remote(src_path, &npaths);
		for (j = 0; j < npaths; j++) {
			src_path = paths[j];
			src_file = dirchar_basename(src_path);
			hldir = path_to_hldir(src_path, &hldirlen, 1);
			rnhldir = path_to_hldir(dst_path, &rnhldirlen, 1);
			task_new(htlc, 0, 0);
			hlwrite(htlc, HTLC_HDR_FILE_SYMLINK, 0, 4,
				HTLC_DATA_FILE_NAME, strlen(src_file), src_file,
				HTLC_DATA_DIR, hldirlen, hldir,
				HTLC_DATA_DIR_RENAME, rnhldirlen, rnhldir,
				HTLC_DATA_FILE_RENAME, strlen(dst_file), dst_file);
			xfree(rnhldir);
			xfree(hldir);
			xfree(src_path);
		}
		xfree(paths);
	}
}

COMMAND(mv)
{
	char *src_path, *dst_path, *src_file, *dst_file, **paths;
	char src_buf[MAXPATHLEN], dst_buf[MAXPATHLEN];
	u_int16_t hldirlen, rnhldirlen;
	u_int8_t *hldir, *rnhldir;
	int i, j, npaths;
	size_t len;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <source file> <dest file>\n", argv[0]);
		return;
	}
	dst_path = argv[--argc];
	if (*dst_path == dir_char)
		strcpy(dst_buf, dst_path);
	else
		snprintf(dst_buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, dst_path);
	dst_path = dst_buf;
	dst_file = dirchar_basename(dst_path);
	for (i = 1; i < argc; i++) {
		src_path = argv[i];
		if (*src_path == dir_char)
			strcpy(src_buf, src_path);
		else
			snprintf(src_buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, src_path);
		src_path = src_buf;
		paths = glob_remote(src_path, &npaths);
		for (j = 0; j < npaths; j++) {
			src_path = paths[j];
			src_file = dirchar_basename(src_path);
			hldir = path_to_hldir(src_path, &hldirlen, 1);
			len = strlen(dst_path) - (strlen(dst_path) - (dst_file - dst_path));
			if (len && (len != strlen(src_path) - (strlen(src_path) - (src_file - src_path)) || memcmp(dst_path, src_path, len))) {
				rnhldir = path_to_hldir(dst_path, &rnhldirlen, 1);
				task_new(htlc, 0, 0);
				hlwrite(htlc, HTLC_HDR_FILE_MOVE, 0, 3,
					HTLC_DATA_FILE_NAME, strlen(src_file), src_file,
					HTLC_DATA_DIR, hldirlen, hldir,
					HTLC_DATA_DIR_RENAME, rnhldirlen, rnhldir);
				xfree(rnhldir);
			}
			if (*dst_file && strcmp(src_file, dst_file)) {
				task_new(htlc, 0, 0);
				hlwrite(htlc, HTLC_HDR_FILE_SETINFO, 0, 3,
					HTLC_DATA_FILE_NAME, strlen(src_file), src_file,
					HTLC_DATA_FILE_RENAME, strlen(dst_file), dst_file,
					HTLC_DATA_DIR, hldirlen, hldir);
			}
			xfree(hldir);
			xfree(src_path);
		}
		xfree(paths);
	}
}

static void
rcv_task_file_getinfo (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	u_int8_t icon[4], type[32], crea[32], date_create[8], date_modify[8];
	u_int8_t name[256], comment[256];
	u_int16_t nlen, clen, tlen;
	u_int32_t size = 0;
	char created[32], modified[32];
	time_t t;
	struct tm tm;

	if ((ntohs(h->flag) & 1))
		return;
	name[0] = comment[0] = type[0] = crea[0] = 0;
	dh_start(htlc)
		switch(ntohs(dh->type)) {
			case HTLS_DATA_FILE_ICON:
				if (ntohs(dh->len) >= 4)
					memcpy(icon, dh->data, 4);
				break;
			case HTLS_DATA_FILE_TYPE:
				tlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(type, dh->data, tlen);
				type[tlen] = 0;
				break;
			case HTLS_DATA_FILE_CREATOR:
				clen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(crea, dh->data, clen);
				crea[clen] = 0;
				break;
			case HTLS_DATA_FILE_SIZE:
				if (ntohs(dh->len) >= 4)
					size = ntohl(*((u_int32_t *)dh->data));
				break;
			case HTLS_DATA_FILE_NAME:
				nlen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				strip_ansi(name, nlen);
				break;
			case HTLS_DATA_FILE_DATE_CREATE:
				if (ntohs(dh->len) >= 8)
					memcpy(date_create, dh->data, 8);
				break;
			case HTLS_DATA_FILE_DATE_MODIFY:
				if (ntohs(dh->len) >= 8)
					memcpy(date_modify, dh->data, 8);
				break;
			case HTLS_DATA_FILE_COMMENT:
				clen = ntohs(dh->len) > 255 ? 255 : ntohs(dh->len);
				memcpy(comment, dh->data, clen);
				comment[clen] = 0;
				CR2LF(comment, clen);
				strip_ansi(name, clen);
				break;
		}
	dh_end()

	t = ntohl(*((u_int32_t *)&date_create[4])) - 0x7c25b080;
	localtime_r(&t, &tm);
	strftime(created, 31, hx_timeformat, &tm);
	t = ntohl(*((u_int32_t *)&date_modify[4])) - 0x7c25b080;
	localtime_r(&t, &tm);
	strftime(modified, 31, hx_timeformat, &tm);

	hx_printf("\
    icon: %4.4s\n\
    type: %s\n\
 creator: %s\n\
    size: %u\n\
    name: %s\n\
 created: %s\n\
modified: %s\n\
 comment: %s\n", icon, type, crea, size, name, created, modified, comment);
}

COMMAND(finfo)
{
	char *path, **paths, *file, buf[MAXPATHLEN];
	u_int16_t hldirlen;
	u_int8_t *hldir;
	int i, j, npaths;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <file> [file2...]\n", argv[0]);
		return;
	}
	for (i = 1; i < argc; i++) {
		path = argv[i];
		if (*path == dir_char)
			strcpy(buf, path);
		else
			snprintf(buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, path);
		path = buf;
		paths = glob_remote(path, &npaths);
		for (j = 0; j < npaths; j++) {
			path = paths[j];
			task_new(htlc, rcv_task_file_getinfo, 0);
			file = dirchar_basename(path);
			if (file != path) {
				hldir = path_to_hldir(path, &hldirlen, 1);
				hlwrite(htlc, HTLC_HDR_FILE_GETINFO, 0, 2,
					HTLC_DATA_FILE_NAME, strlen(file), file,
					HTLC_DATA_DIR, hldirlen, hldir);
				xfree(hldir);
			} else {
				hlwrite(htlc, HTLC_HDR_FILE_GETINFO, 0, 1,
					HTLC_DATA_FILE_NAME, strlen(file), file);
			}
			xfree(path);
		}
		xfree(paths);
	}
}

static void
exec_close (int fd)
{
	close(fd);
	memset(&hxd_files[fd], 0, sizeof(struct hxd_file));
	FD_CLR(fd, &hxd_rfds);
	if (high_fd == fd) {
		for (fd--; fd && !FD_ISSET(fd, &hxd_rfds); fd--)
			;
		high_fd = fd;
	}
}

static void
exec_ready_read (int fd)
{
	ssize_t r;
	u_int8_t buf[0x4000];

	r = read(fd, buf, sizeof(buf) - 1);
	if (r == 0 || (r < 0 && errno != EINTR)) {
		if (hxd_files[fd].conn.htxf && (hxd_files[fd].conn.htlc != &hx_htlc))
			play_sound(snd_file_done);
		exec_close(fd);
	} else {
		buf[r] = 0;
		if (hxd_files[fd].conn.htlc == &hx_htlc) {
			LF2CR(buf, r);
			if (buf[r - 1] == '\r')
				buf[r - 1] = 0;
			hx_send_chat(buf);
		} else {
			hx_printf("%s", buf);
		}
	}
}

COMMAND(exec)
{
	int pfds[2];
	char *p, *av[4];
	u_int32_t output_to = 0;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s [-o] <command>\n", argv[0]);
		return;
	}
	p = str;
find_cmd_arg:
	for (; *p && *p != ' '; p++) ;
	if (!*p || !(*++p))
		return;
	if (*p == '-' && *(p + 1) == 'o') {
		output_to = 1;
		goto find_cmd_arg;
	}
	if (pipe(pfds)) {
		hx_printf(INFOPREFIX "%s: pipe: %s\n", argv[0], strerror(errno));
		return;
	}
	if (pfds[0] >= hxd_open_max) {
		hx_printf(INFOPREFIX "%s:%d: %d >= hxd_open_max (%d)\n", __FILE__, __LINE__, pfds[0], hxd_open_max);
		close(pfds[0]);
		close(pfds[1]);
		return;
	}
	switch (fork()) {
		case -1:
			hx_printf(INFOPREFIX "%s: fork: %s\n", argv[0], strerror(errno));
			close(pfds[0]);
			close(pfds[1]);
			return;
		case 0:
			close(pfds[0]);
			av[0] = "/bin/sh";
			av[1] = "-c";
			av[2] = p;
			av[3] = 0;
			if (dup2(pfds[1], 1) == -1 || dup2(pfds[1], 2) == -1) {
				hx_printf(INFOPREFIX "%s: dup2: %s\n", argv[0], strerror(errno));
				_exit(1);
			}
			close(0);
			execve("/bin/sh", av, hxd_environ);
			hx_printf(INFOPREFIX "%s: execve: %s\n", argv[0], strerror(errno));
			_exit(127);
		default:
			close(pfds[1]);
			if (output_to)
				hxd_files[pfds[0]].conn.htlc = htlc;
			else
				hxd_files[pfds[0]].conn.htlc = 0;
			hxd_files[pfds[0]].ready_read = exec_ready_read;
			FD_SET(pfds[0], &hxd_rfds);
			if (high_fd < pfds[0])
				high_fd = pfds[0];
			break;
	}
}

#if defined(CONFIG_HFS)
static int
resource_path (char *rsrcpath, const char *path, struct stat *statbuf)
{
	int i, len;
	char pathbuf[MAXPATHLEN];
	struct stat sb;

	len = strlen(path);
	if (len + 16 >= MAXPATHLEN)
		return ENAMETOOLONG;
	strcpy(pathbuf, path);
	for (i = len - 1; i > 0; i--)
		if (pathbuf[i] == DIRCHAR) {
			pathbuf[i++] = 0;
			break;
		}
	snprintf(rsrcpath, MAXPATHLEN, "%s/.resource", i > 0 ? pathbuf : ".");
	if (stat(rsrcpath, &sb)) {
		if (statbuf)
			return ENOENT;
		if (mkdir(rsrcpath, S_IRUSR|S_IWUSR|S_IXUSR))
			return errno;
	}
	strcat(rsrcpath, "/");
	strcat(rsrcpath, &pathbuf[i]);
	if (statbuf && stat(rsrcpath, statbuf))
		return errno;

	return 0;
}

static int
resource_open (const char *path, int mode, int perm)
{
	int f = -1;

	if (1) {
		char rsrcpath[MAXPATHLEN];

		if (resource_path(rsrcpath, path, 0))
			return -1;
		f = open(rsrcpath, mode, perm);
	}

	return f;
}

static size_t
resource_len (const char *path)
{
	if (1) {
		char rsrcpath[MAXPATHLEN];
		struct stat sb;

		if (resource_path(rsrcpath, path, &sb))
			return 0;

		return sb.st_size;
	}

	return 0;
}

static int
finderinfo_path (char *infopath, const char *path, struct stat *statbuf)
{
	int i, len;
	char pathbuf[MAXPATHLEN];
	struct stat sb;

	len = strlen(path);
	if (len + 16 >= MAXPATHLEN)
		return ENAMETOOLONG;
	strcpy(pathbuf, path);
	for (i = len - 1; i > 0; i--)
		if (pathbuf[i] == DIRCHAR) {
			pathbuf[i++] = 0;
			break;
		}
	snprintf(infopath, MAXPATHLEN, "%s/.finderinfo", i > 0 ? pathbuf : ".");
	if (stat(infopath, &sb)) {
		if (statbuf)
			return ENOENT;
		if (mkdir(infopath, S_IRUSR|S_IWUSR|S_IXUSR))
			return errno;
	}
	strcat(infopath, "/");
	strcat(infopath, &pathbuf[i]);
	if (statbuf && stat(infopath, statbuf))
		return errno;

	return 0;
}

static int
hfsinfo_read (const char *path, struct hfs_cap_info *fi)
{
	char infopath[MAXPATHLEN];
	int f, r = 0;
	struct stat sb;

	if (!finderinfo_path(infopath, path, &sb)) {
		f = open(infopath, O_RDONLY);
		if (f >= 0) {
			r = read(f, fi, SIZEOF_HFS_CAP_INFO);
			close(f);
		}
	}

	return r;
}

static void
hfsinfo_write (const char *path, struct hfs_cap_info *fi)
{
	char infopath[MAXPATHLEN];
	int f;

	if (!finderinfo_path(infopath, path, 0)) {
		f = open(infopath, O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR);
		if (f >= 0) {
			fi->fi_magic1 = HFS_CAP_MAGIC1;
			fi->fi_version = HFS_CAP_VERSION;
			fi->fi_magic = HFS_CAP_MAGIC;
			fi->fi_datemagic = HFS_CAP_DMAGIC;
			fi->fi_datevalid = HFS_CAP_MDATE|HFS_CAP_CDATE;
			write(f, fi, SIZEOF_HFS_CAP_INFO);
			fsync(f);
			close(f);
		}
	}
}

static inline int
comment_len (const char *path)
{
	char infopath[MAXPATHLEN];
	int f, len = 0;
	struct stat sb;
	struct hfs_cap_info cap;

	if (!finderinfo_path(infopath, path, &sb)) {
		f = open(infopath, O_RDONLY);
		if (f >= 0) {
			if (read(f, &cap, SIZEOF_HFS_CAP_INFO) == SIZEOF_HFS_CAP_INFO)
				len = cap.fi_comln > 200 ? 200 : cap.fi_comln;
			close(f);
		}
	}

	return len;
}

static inline char *
suffix (const char *path)
{
	char *p = (char *)path + strlen(path);
	while (p-- > path)
		if (*p == '.')
			return p + 1;
	return 0;
}

static void
type_creator (u_int8_t *buf, const char *path)
{
	u_int8_t *tc;
	char *suff;
	char infopath[MAXPATHLEN];
	struct stat sb;

	if (!finderinfo_path(infopath, path, &sb)) {
		int r, fd = open(infopath, O_RDONLY);
		if (fd < 0)
			goto use_suffix;
		r = read(fd, buf, 8);
		close(fd);
		if (r != 8)
			goto use_suffix;
		return;
	}

use_suffix:
	tc = UNKNOWN_TYPECREA;
	if (!(suff = suffix(path)))
		goto dunno;

	if (!strcmp(suff, "jpg") || !strcmp(suff, "jpeg"))
		tc = "JPEGGKON";
	else if (!strcmp(suff, "png"))
		tc = "PNGfGKON";
	else if (!strcmp(suff, "gif"))
		tc = "GIFfGKON";
	else if (!strcmp(suff, "mp3") || !strcmp(suff, "mp2"))
		tc = "MP3 MAmp";
	else if (!strcmp(suff, "mpg") || !strcmp(suff, "mpeg"))
		tc = "MPEGTVOD";
	else if (!strcmp(suff, "mov"))
		tc = "MooVTVOD";
	else if (!strcmp(suff, "sit"))
		tc = "SITDSIT!";
	else if (!strcmp(suff, "zip"))
		tc = "ZIP ZIP ";
	else if (!strcmp(suff, "app") || !strcmp(suff, "sea"))
		tc = "APPLpeff";
	else if (!strcmp(suff, "img"))
		tc = "rohdWrap";
	else if (!strcmp(suff, "pict"))
		tc = "PICTGKON";

dunno:
	memcpy(buf, tc, 8);
}
#endif /* CONFIG_HFS */

static void
ignore_signals (sigset_t *oldset)
{
	sigset_t set;

	sigfillset(&set);
	sigprocmask(SIG_BLOCK, &set, oldset);
}

static void
unignore_signals (sigset_t *oldset)
{
	sigprocmask(SIG_SETMASK, oldset, 0);
}

static struct htxf_conn **xfers = 0;
static int nxfers = 0;

static void rcv_task_file_get (struct htlc_conn *htlc, struct htxf_conn *htxf);
static void rcv_task_file_put (struct htlc_conn *htlc, struct htxf_conn *htxf);

static void
xfer_go (struct htxf_conn *htxf)
{
	char *rfile;
	u_int16_t hldirlen;
	u_int8_t *hldir;
	u_int32_t data_size = 0, rsrc_size = 0;
	u_int8_t rflt[74];
	struct stat sb;

	if (htxf->gone)
		return;
	htxf->gone = 1;
	if (htxf->type == XFER_GET) {
		if (!stat(htxf->path, &sb))
			data_size = sb.st_size;
#if defined(CONFIG_HFS)
		rsrc_size = resource_len(htxf->path);
#else
		rsrc_size = 0;
#endif
		rfile = dirchar_basename(htxf->remotepath);
		if (data_size || rsrc_size) {
			memcpy(rflt, "\
RFLT\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\2\
DATA\0\0\0\0\0\0\0\0\0\0\0\0\
MACR\0\0\0\0\0\0\0\0\0\0\0\0", 74);
			S32HTON(data_size, &rflt[46]);
			S32HTON(rsrc_size, &rflt[62]);
			htxf->data_pos = data_size;
			htxf->rsrc_pos = rsrc_size;
		}
		task_new(&hx_htlc, rcv_task_file_get, htxf);
		if (rfile != htxf->remotepath) {
			hldir = path_to_hldir(htxf->remotepath, &hldirlen, 1);
			hlwrite(&hx_htlc, HTLC_HDR_FILE_GET, 0, (data_size || rsrc_size) ? 3 : 2,
				HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
				HTLC_DATA_DIR, hldirlen, hldir,
				HTLC_DATA_RFLT, 74, rflt);
			xfree(hldir);
		} else {
			hlwrite(&hx_htlc, HTLC_HDR_FILE_GET, 0, (data_size || rsrc_size) ? 2 : 1,
				HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
				HTLC_DATA_RFLT, 74, rflt);
		}
	} else {
		rfile = basename(htxf->path);
		hldir = path_to_hldir(htxf->remotepath, &hldirlen, 1);
		if (exists_remote(htxf->remotepath)) {
			task_new(&hx_htlc, rcv_task_file_put, htxf);
			hlwrite(&hx_htlc, HTLC_HDR_FILE_PUT, 0, 3,
				HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
				HTLC_DATA_DIR, hldirlen, hldir,
				HTLC_DATA_FILE_PREVIEW, 2, "\0\1");
		} else {
			task_new(&hx_htlc, rcv_task_file_put, htxf);
			hlwrite(&hx_htlc, HTLC_HDR_FILE_PUT, 0, 2,
				HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
				HTLC_DATA_DIR, hldirlen, hldir);
		}
		xfree(hldir);
	}
}

static struct htxf_conn *
xfer_new (const char *path, const char *remotepath, u_int16_t type)
{
	struct htxf_conn *htxf;

	htxf = xmalloc(sizeof(struct htxf_conn));
	memset(htxf, 0, sizeof(struct htxf_conn));
	strcpy(htxf->remotepath, remotepath);
	strcpy(htxf->path, path);
	htxf->type = type;

	xfers = xrealloc(xfers, (nxfers + 1) * sizeof(struct htxf_conn *));
	xfers[nxfers] = htxf;
	nxfers++;
	if (nxfers == 1)
		xfer_go(htxf);

	return htxf;
}

static void
xfer_delete (struct htxf_conn *htxf)
{
	int i;

	for (i = 0; i < nxfers; i++)
		if (xfers[i] == htxf) {
#if defined(CONFIG_HTXF_PTHREAD)
			if (htxf->tid)
				pthread_cancel(htxf->tid);
#else
			if (htxf->pid)
				kill(htxf->pid, SIGKILL);
#endif
			else {
#if defined(CONFIG_HTXF_CLONE)
				xfree(htxf->stack);
#endif
				timer_delete_ptr(htxf);
				xfree(htxf);
				if (nxfers > (i+1))
					memcpy(xfers+i, xfers+i+1, (nxfers-(i+1)) * sizeof(struct htxf_conn *));
				nxfers--;
				if (nxfers)
					xfer_go(xfers[0]);
			}
			break;
		}
}

struct sound_event {
	pid_t pid;
	char *player;
	char *sound;
	short on, beep;
};

static struct sound_event sounds[] = { 
	{ 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0 },
};
#define nsounds	(int)(sizeof(sounds) / sizeof(struct sound_event))

#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
static void
free_htxf (void *__arg)
{
	int i;
	pid_t pid = (pid_t)__arg;
	struct htxf_conn *htxfp;

	for (i = 0; i < nxfers; i++) {
		htxfp = xfers[i];
		if (htxfp->pid == pid) {
			htxfp->pid = 0;
			xfer_delete(htxfp);
			return;
		}
	}
}
#endif

void
hlclient_reap_pid (pid_t pid, int status)
{
	int i;
#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
	struct htxf_conn *htxfp;

	for (i = 0; i < nxfers; i++) {
		htxfp = xfers[i];
		if (htxfp->pid == pid) {
			timer_add_secs(0, free_htxf, (void *)pid);
			return;
		}
	}
#endif

	for (i = 0; i < nsounds; i++) {
		if (sounds[i].pid == pid) {
			if ((WIFEXITED(status) && WEXITSTATUS(status)) && sounds[i].beep)
				write(1, "\a", 1);
		}
	}
}

COMMAND(xfers)
{
	struct htxf_conn *htxfp;
	int i, j;

	hx_output.mode_underline();
	hx_printf(" #   type   pid   file\n");
	hx_output.mode_clear();
	for (i = 0; i < nxfers; i++) {
		struct timeval now;
		time_t sdiff, usdiff, Bps, eta;

		if (argc > 1) {
			for (j = 1; j < argc; j++)
				if (i == (int)atou32(argv[j]))
					goto want_this;
			continue;
		}
want_this:
		htxfp = xfers[i];
		gettimeofday(&now, 0);
		sdiff = now.tv_sec - htxfp->start.tv_sec;
		usdiff = now.tv_usec - htxfp->start.tv_usec;
		if (!sdiff)
			sdiff = 1;
		Bps = htxfp->total_pos / sdiff;
		if (!Bps)	
			Bps = 1;
		eta = (htxfp->total_size - htxfp->total_pos) / Bps
		    + ((htxfp->total_size - htxfp->total_pos) % Bps) / Bps;
		hx_printf("[%d]  %s  %5u  %u/%u  %lu Bps  ETA: %lu s  %s\n", i,
			  htxfp->type == XFER_GET ? "get" : "put",
#if defined(CONFIG_HTXF_PTHREAD)
			  htxfp->tid,
#else
			  htxfp->pid,
#endif
			  htxfp->total_pos, htxfp->total_size, Bps, eta, htxfp->path);
	}
}

COMMAND(xfkill)
{
	int i;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <xfer number> | *\n", argv[0]);
		return;
	}
	if (argv[1][0] == '*') {
		while (nxfers) {
			xfers[0]->gone = 1;
			xfer_delete(xfers[0]);
		}
		return;
	}
	for (i = 1; i < argc; i++) {
		int x = atou32(argv[i]);
		if (x < nxfers)
			xfer_delete(xfers[x]);
	}
}

COMMAND(xfgo)
{
	int i;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <xfer number> | *\n", argv[0]);
		return;
	}
	if (argv[1][0] == '*') {
		for (i = 0; i < nxfers; i++)
			xfer_go(xfers[i]);
		return;
	}
	for (i = 1; i < argc; i++) {
		int x = atou32(argv[i]);
		if (x < nxfers)
			xfer_go(xfers[x]);
	}
}

static int
htxf_connect (struct htxf_conn *htxf)
{
	struct htxf_hdr h;
	int s;

	s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (s < 0) {
		return -1;
	}
	if (connect(s, (struct sockaddr *)&htxf->listen_sockaddr, sizeof(struct sockaddr_in))) {
		close(s);
		return -1;
	}
	h.magic = htonl(HTXF_MAGIC_INT);
	h.ref = htonl(htxf->ref);
	h.unknown = 0;
	h.len = htonl(htxf->total_size);
	if (write(s, &h, SIZEOF_HTXF_HDR) != SIZEOF_HTXF_HDR) {
		return -1;
	}

	return s;
}

static int
rd_wr (int rd_fd, int wr_fd, u_int32_t data_len, u_int32_t *total_pos)
{
	int r, pos, len;
	u_int8_t *buf;
	size_t bufsiz;

	bufsiz = 0xf000;
	buf = malloc(bufsiz);
	if (!buf)
		return 111;
	while (data_len) {
		if ((len = read(rd_fd, buf, (bufsiz < data_len) ? bufsiz : data_len)) < 1)
			return len ? errno : EIO;
		pos = 0;
		while (len) {
			if ((r = write(wr_fd, &(buf[pos]), len)) < 1)
				return errno;
			pos += r;
			len -= r;
			*(total_pos) += r;
		}
		data_len -= pos;
	}
	free(buf);

	return 0;
}

#if defined(CONFIG_HTXF_PTHREAD)
typedef void * thread_return_type;
#else
typedef int thread_return_type;
#endif

static thread_return_type
get_thread (void *__arg)
{
	struct htxf_conn *htxf = (struct htxf_conn *)__arg;
	u_int32_t pos, len, tot_len;
	int s, f, r, retval = 0;
	u_int8_t typecrea[8], buf[1024];
	struct hfs_cap_info fi;

	s = htxf_connect(htxf);
	if (s < 0) {
		retval = s;
		goto ret;
	}

	len = 40;
	pos = 0;
	while (len) {
		if ((r = read(s, &(buf[pos]), len)) < 1) {
			retval = errno;
			goto ret;
		}
		pos += r;
		len -= r;
		htxf->total_pos += r;
	}
	pos = 0;
	len = (buf[38] ? 0x100 : 0) + buf[39];
	len += 16;
	tot_len = 40 + len;
	while (len) {
		if ((r = read(s, &(buf[pos]), len)) < 1) {
			retval = errno;
			goto ret;
		}
		pos += r;
		len -= r;
		htxf->total_pos += r;
	}
#if defined(CONFIG_HFS)
	memcpy(typecrea, &buf[4], 8);
	memset(&fi, 0, sizeof(struct hfs_cap_info));
	fi.fi_comln = buf[73 + buf[71]];
	memcpy(fi.fi_fndr, "HTftHTLC", 8);
	memcpy(fi.fi_comnt, &buf[74 + buf[71]], fi.fi_comln);
	*((u_int32_t *)(&buf[56])) = hfs_m_to_htime(*((u_int32_t *)(&buf[56])));
	*((u_int32_t *)(&buf[64])) = hfs_m_to_htime(*((u_int32_t *)(&buf[64])));
	memcpy(&fi.fi_ctime, &buf[56], 4);
	memcpy(&fi.fi_mtime, &buf[64], 4);
	hfsinfo_write(htxf->path, &fi);
#endif /* CONFIG_HFS */

	L32NTOH(len, &buf[pos - 4]);
	tot_len += len;
	if ((f = open(htxf->path, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR)) < 0) {
		retval = errno;
		goto ret;
	}
	if (!len)
		goto get_rsrc;
	if (htxf->data_pos)
		lseek(f, htxf->data_pos, SEEK_SET);
	if (htxf->filter_argv) {
		pid_t pid;
		int pfds[2];

		if (pipe(pfds)) {
			hx_printf(INFOPREFIX "get_thread: pipe: %s\n", strerror(errno));
			goto no_filter;
		}
		pid = fork();
		switch (pid) {
			case -1:
				hx_printf(INFOPREFIX "get_thread: fork: %s\n", strerror(errno));
				close(pfds[0]);
				close(pfds[1]);
				goto no_filter;
			case 0:
				close(pfds[1]);
				if (dup2(pfds[0], 0) == -1 || dup2(f, 1) == -1) {
					hx_printf(INFOPREFIX "get_thread: dup2: %s\n", strerror(errno));
					_exit(1);
				}
				close(2);
				if (htxf->data_pos)
					lseek(0, htxf->data_pos, SEEK_SET);
				execve(htxf->filter_argv[0], htxf->filter_argv, hxd_environ);
				hx_printf(INFOPREFIX "%s: execve: %s\n", htxf->filter_argv[0], strerror(errno));
				_exit(127);
			default:
				close(pfds[0]);
				close(f);
				f = pfds[1];
		}
	}
no_filter:
	retval = rd_wr(s, f, len, &htxf->total_pos);
	fsync(f);
	close(f);
	if (retval)
		goto ret;
get_rsrc:
#if defined(CONFIG_HFS)
	if (tot_len >= htxf->total_size)
		goto done;
	pos = 0;
	len = 16;
	while (len) {
		if ((r = read(s, &(buf[pos]), len)) < 1) {
			retval = errno;
			goto ret;
		}
		pos += r;
		len -= r;
		htxf->total_pos += r;
	}
	L32NTOH(len, &buf[12]);
	if (!len)
		goto done;
	if ((f = resource_open(htxf->path, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR)) < 0) {
		retval = errno;
		goto ret;
	}
	if (htxf->rsrc_pos)
		lseek(f, htxf->rsrc_pos, SEEK_SET);
	retval = rd_wr(s, f, len, &htxf->total_pos);
	fsync(f);
	close(f);
	if (retval)
		goto ret;

done:
	memcpy(fi.fi_fndr, typecrea, 8);
	hfsinfo_write(htxf->path, &fi);
#endif /* CONFIG_HFS */

	{
#if defined(CONFIG_HTXF_FORK)
		char outbuf[MAXPATHLEN + 128];
		int len;
#endif
		struct timeval now;
		time_t sdiff, usdiff, Bps;

		gettimeofday(&now, 0);
		sdiff = now.tv_sec - htxf->start.tv_sec;
		usdiff = now.tv_usec - htxf->start.tv_usec;
		if (!sdiff)
			sdiff = 1;
		Bps = htxf->total_pos / sdiff;
#if defined(CONFIG_HTXF_FORK)
		len = snprintf(outbuf, sizeof(outbuf), INFOPREFIX "%s: %lu second%c, %d bytes, %lu Bps\n",
			       htxf->path, sdiff, sdiff == 1 ? 0 : 's', htxf->total_pos, Bps);
		write(htxf->pipe, outbuf, len);
#else
		hx_printf(INFOPREFIX "%s: %lu second%c, %d bytes, %lu Bps\n",
			  htxf->path, sdiff, sdiff == 1 ? 0 : 's', htxf->total_pos, Bps);
#endif
	}

ret:
	close(s);
#if defined(CONFIG_HTXF_PTHREAD)
	htxf->tid = 0;
	xfer_delete(htxf);
#endif

	return (thread_return_type)retval;
}

static thread_return_type
put_thread (void *__arg)
{
	struct htxf_conn *htxf = (struct htxf_conn *)__arg;
	int s, f, retval = 0;
	u_int8_t buf[512];
	struct hfs_cap_info fi;

	s = htxf_connect(htxf);
	if (s < 0) {
		retval = s;
		goto ret;
	}

	memcpy(buf, "\
FILP\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\2INFO\0\0\0\0\0\0\0\0\0\0\0^AMAC\
TYPECREA\
\0\0\0\0\0\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\
\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\7\160\0\0\0\0\0\0\7\160\0\0\0\0\0\0\0\0\0\3hxd", 115);
#if defined(CONFIG_HFS)
	hfsinfo_read(htxf->path, &fi);
	if (htxf->rsrc_size - htxf->rsrc_pos)
		buf[23] = 3;
	if (65 + fi.fi_comln + 12 > 0xff)
		buf[38] = 1;
	buf[39] = 65 + fi.fi_comln + 12;
	type_creator(&buf[44], htxf->path);
	*((u_int32_t *)(&buf[96])) = hfs_h_to_mtime(*((u_int32_t *)(&fi.fi_ctime)));
	*((u_int32_t *)(&buf[104])) = hfs_h_to_mtime(*((u_int32_t *)(fi.fi_mtime)));
	buf[116] = fi.fi_comln;
	memcpy(buf+117, fi.fi_comnt, fi.fi_comln);
	memcpy(&buf[117] + fi.fi_comln, "DATA\0\0\0\0\0\0\0\0", 12);
	S32HTON((htxf->data_size - htxf->data_pos), &buf[129 + fi.fi_comln]);
	if (write(s, buf, 133 + fi.fi_comln) != 133 + fi.fi_comln) {
		retval = errno;
		goto ret;
	}
	htxf->total_pos += 133 + fi.fi_comln;
#else
	buf[39] = 65 + 12;
	memcpy(&buf[117], "DATA\0\0\0\0\0\0\0\0", 12);
	S32HTON((htxf->data_size - htxf->data_pos), &buf[129]);
	if (write(s, buf, 133) != 133) {
		retval = errno;
		goto ret;
	}
	htxf->total_pos += 133;
#endif /* CONFIG_HFS */

	if (!(htxf->data_size - htxf->data_pos))
		goto put_rsrc;
	if ((f = open(htxf->path, O_RDONLY)) < 0) {
		retval = errno;
		goto ret;
	}
	if (htxf->data_pos)
		lseek(f, htxf->data_pos, SEEK_SET);
	retval = rd_wr(f, s, htxf->data_size, &htxf->total_pos);
	if (retval)
		goto ret;
	close(f);

put_rsrc:
#if defined(CONFIG_HFS)
	if (!(htxf->rsrc_size - htxf->rsrc_pos))
		goto done;
	memcpy(buf, "MACR\0\0\0\0\0\0\0\0", 12);
	S32HTON(htxf->rsrc_size, &buf[12]);
	if (write(s, buf, 16) != 16) {
		retval = 0;
		goto ret;
	}
	htxf->total_pos += 16;
	if ((f = resource_open(htxf->path, O_RDONLY, 0)) < 0) {
		retval = errno;
		goto ret;
	}
	if (htxf->rsrc_pos)
		lseek(f, htxf->rsrc_pos, SEEK_SET);
	retval = rd_wr(f, s, htxf->rsrc_size, &htxf->total_pos);
	if (retval)
		goto ret;
	close(f);

done:
#endif /* CONFIG_HFS */

	{
#if defined(CONFIG_HTXF_FORK)
		char outbuf[MAXPATHLEN + 128];
		int len;
#endif
		struct timeval now;
		time_t sdiff, usdiff, Bps;

		gettimeofday(&now, 0);
		sdiff = now.tv_sec - htxf->start.tv_sec;
		usdiff = now.tv_usec - htxf->start.tv_usec;
		if (!sdiff)
			sdiff = 1;
		Bps = htxf->total_pos / sdiff;
#if defined(CONFIG_HTXF_FORK)
		len = snprintf(outbuf, sizeof(outbuf), INFOPREFIX "%s: %lu second%c, %d bytes, %lu Bps\n",
			       htxf->path, sdiff, sdiff == 1 ? 0 : 's', htxf->total_pos, Bps);
		write(htxf->pipe, outbuf, len);
#else
		hx_printf(INFOPREFIX "%s: %lu second%c, %d bytes, %lu Bps\n",
			  htxf->path, sdiff, sdiff == 1 ? 0 : 's', htxf->total_pos, Bps);
#endif
	}

ret:
	close(s);
#if defined(CONFIG_HTXF_PTHREAD)
	htxf->tid = 0;
	xfer_delete(htxf);
#endif

	return (thread_return_type)retval;
}

static void
xfer_ready_write (struct htxf_conn *htxf)
{
	sigset_t oldset;
	struct sigaction act, tstpact, contact;
#if defined(CONFIG_HTXF_PTHREAD) || defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
#ifdef CONFIG_HTXF_PTHREAD
	pthread_t tid;
	int err;
#else
	pid_t pid;
#ifdef CONFIG_HTXF_CLONE
	void *stack;
#endif
#ifdef CONFIG_HTXF_FORK
	int pfds[2];
#endif
#endif
#endif

#if defined(CONFIG_HTXF_FORK)
	if (pipe(pfds)) {
		hx_printf(INFOPREFIX "xfer: pipe: %s\n", strerror(errno));
		goto err_fd;
	}
	if (pfds[0] >= hxd_open_max) {
		hx_printf(INFOPREFIX "%s:%d: %d >= hxd_open_max (%d)\n", __FILE__, __LINE__, pfds[0], hxd_open_max);
		goto err_pfds;
	}
	htxf->pipe = pfds[1];
#endif
	ignore_signals(&oldset);
	act.sa_flags = 0;
	act.sa_handler = SIG_DFL;
	sigfillset(&act.sa_mask);
	sigaction(SIGTSTP, &act, &tstpact);
	sigaction(SIGCONT, &act, &contact);
#if defined(CONFIG_HTXF_PTHREAD)
	err = pthread_create(&tid, 0, ((htxf->type == XFER_GET) ? get_thread : put_thread), htxf);
#elif defined(CONFIG_HTXF_CLONE)
	htxf->stack = xmalloc(CLONE_STACKSIZE);
	stack = (void *)(((u_int8_t *)htxf->stack + CLONE_STACKSIZE) - sizeof(void *));
	pid = clone(htxf->type == XFER_GET ? get_thread : put_thread, stack, CLONE_VM|CLONE_FS|SIGCHLD, htxf);
#elif defined(CONFIG_HTXF_FORK)
	pid = fork();
	if (pid == 0)
		_exit(htxf->type == XFER_GET ? get_thread(htxf) : put_thread(htxf));
#endif
	sigaction(SIGTSTP, &tstpact, 0);
	sigaction(SIGCONT, &contact, 0);
	unignore_signals(&oldset);
#if defined(CONFIG_HTXF_PTHREAD)
	if (err) {
		hx_printf(INFOPREFIX "xfer: pthread_create: %s\n", strerror(err));
		goto err_fd;
	}
	htxf->tid = tid;
	pthread_detach(tid);
#else
	if (pid == -1) {
#if defined(CONFIG_HTXF_CLONE)
		hx_printf(INFOPREFIX "xfer: clone: %s\n", strerror(errno));
		goto err_fd;
#else
		hx_printf(INFOPREFIX "xfer: fork: %s\n", strerror(errno));
		goto err_pfds;
#endif
	}
	htxf->pid = pid;
#endif
#if defined(CONFIG_HTXF_FORK)
	close(pfds[1]);
	hxd_files[pfds[0]].ready_read = exec_ready_read;
	hxd_files[pfds[0]].conn.htxf = htxf;
	FD_SET(pfds[0], &hxd_rfds);
	if (high_fd < pfds[0])
		high_fd = pfds[0];
#endif
	return;

#if defined(CONFIG_HTXF_FORK)
err_pfds:
	close(pfds[0]);
	close(pfds[1]);
#endif
err_fd:
	xfer_delete(htxf);
}

static void
rcv_task_file_get (struct htlc_conn *htlc, struct htxf_conn *htxf)
{
	u_int32_t ref = 0, size = 0;
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;

	if ((ntohs(h->flag) & 1)) {
		if (htxf->opt.retry) {
			htxf->gone = 0;
			timer_add_secs(1, xfer_go, htxf);
		} else
			xfer_delete(htxf);
		return;
	}

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_HTXF_SIZE:
				dh_getint(size);
				break;
			case HTLS_DATA_HTXF_REF:
				dh_getint(ref);
				break;
		}
	dh_end()
	if (!size || !ref)
		return;
	hx_printf(INFOPREFIX "get: %s; %u bytes\n",
		  htxf->path, size);
	htxf->ref = ref;
	htxf->total_size = size;
	gettimeofday(&htxf->start, 0);
	htxf->listen_sockaddr = htlc->sockaddr;
	htxf->listen_sockaddr.sin_port = htons(ntohs(htxf->listen_sockaddr.sin_port) + 1);
	xfer_ready_write(htxf);
}

static char *tr_argv[] = { PATH_TR, "\r", "\n", 0 };

COMMAND(get)
{
	char *lpath, *rpath, *rfile, **paths, buf[MAXPATHLEN];
	int i, j, npaths, recurs = 0, ckopts = 1, retry = 0;
	char **filter = 0;
	struct htxf_conn *htxf;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s [-Rkt] <file> [file2...]\n", argv[0]);
		return;
	}
	for (i = 1; i < argc; i++) {
		rpath = argv[i];
		if (ckopts && *rpath == '-') {
			rpath++;
			if (*rpath == '-' && !*(rpath+1)) {
				ckopts = 0;
			} else {
				while (*rpath) {
					if (*rpath == 't')
						filter = tr_argv;
					else if (*rpath == 'R')
						recurs = 1;
					else if (*rpath == 'k')
						retry = 1;
					else
						hx_printf(INFOPREFIX "%s: unknown option '%c'\n", argv[0], *rpath);
					rpath++;
				}
			}
			continue;
		}
		if (*rpath == dir_char)
			strcpy(buf, rpath);
		else
			snprintf(buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, rpath);
		rpath = buf;
		paths = glob_remote(rpath, &npaths);
		for (j = 0; j < npaths; j++) {
			rpath = paths[j];
			if (recurs) {
				struct cached_filelist *ncfl;
				u_int16_t hldirlen;
				u_int8_t *hldir;

				ncfl = cfl_lookup(rpath);
				ncfl->completing = COMPLETE_GET_R;
				ncfl->filter_argv = filter;
				if (!ncfl->path)
					ncfl->path = xstrdup(rpath);
				hldir = path_to_hldir(rpath, &hldirlen, 0);
				task_new(&hx_htlc, rcv_task_file_list, ncfl);
				hlwrite(&hx_htlc, HTLC_HDR_FILE_LIST, 0, 1,
					HTLC_DATA_DIR, hldirlen, hldir);
				xfree(hldir);
				xfree(rpath);
				continue;
			}
			rfile = dirchar_basename(rpath);
			lpath = xstrdup(rfile);
			dirchar_fix(lpath);
			htxf = xfer_new(lpath, rpath, XFER_GET);
			htxf->filter_argv = filter;
			htxf->opt.retry = retry;
			xfree(lpath);
			xfree(rpath);
		}
		xfree(paths);
	}
}

static void
rcv_task_file_put (struct htlc_conn *htlc, struct htxf_conn *htxf)
{
	u_int32_t ref = 0, data_pos = 0, rsrc_pos = 0;
	struct stat sb;
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;

	if ((ntohs(h->flag) & 1)) {
		xfer_delete(htxf);
		return;
	}

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_HTXF_REF:
				dh_getint(ref);
				break;
			case HTLS_DATA_RFLT:
				if (ntohs(dh->len) >= 66) {
					L32NTOH(data_pos, &dh->data[46]);
					L32NTOH(rsrc_pos, &dh->data[62]);
				}
				break;
		}
	dh_end()
	if (!ref)
		return;
	htxf->data_pos = data_pos;
	htxf->rsrc_pos = rsrc_pos;
	if (!stat(htxf->path, &sb))
		htxf->data_size = sb.st_size;
	htxf->rsrc_size = resource_len(htxf->path);
	htxf->total_size = 133 + ((htxf->rsrc_size - htxf->rsrc_pos) ? 16 : 0) + comment_len(htxf->path)
			 + (htxf->data_size - htxf->data_pos) + (htxf->rsrc_size - htxf->rsrc_pos);
	htxf->ref = ref;
	hx_printf(INFOPREFIX "put: %s; %u bytes\n",
		  htxf->path, htxf->total_size);
	gettimeofday(&htxf->start, 0);
	htxf->listen_sockaddr = htlc->sockaddr;
	htxf->listen_sockaddr.sin_port = htons(ntohs(htxf->listen_sockaddr.sin_port) + 1);
	xfer_ready_write(htxf);
}

static int
glob_error (const char *epath, int eerrno)
{
	hx_printf(INFOPREFIX "glob: %s: %s\n", epath, strerror(eerrno));

	return 0;
}

COMMAND(put)
{
	char *lpath, *rpath, buf[MAXPATHLEN];
	int i;
	size_t j;
	glob_t g;
	struct htxf_conn *htxf;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <file> [file2...]\n", argv[0]);
		return;
	}
	for (i = 1; i < argc; i++) {
		lpath = argv[i];
#ifndef GLOB_NOESCAPE	/* BSD has GLOB_QUOTE */
#define GLOB_NOESCAPE	0
#endif
#ifndef GLOB_TILDE	/* GNU extension */
#define GLOB_TILDE	0
#endif
		if (glob(argv[i], GLOB_TILDE|GLOB_NOESCAPE|GLOB_NOCHECK, glob_error, &g))
			continue;
		for (j = 0; j < (size_t)g.gl_pathc; j++) {
			lpath = g.gl_pathv[j];
			snprintf(buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, basename(lpath));
			rpath = buf;
			htxf = xfer_new(lpath, rpath, XFER_PUT);
		}
		globfree(&g);
	}
}

COMMAND(clear)
{
	hx_output.clear();
}

COMMAND(save)
{
	if (!argv[1])
		hx_printf(INFOPREFIX "usage: %s <file>\n", argv[0]);
	else
		hx_save(argv[1]);
}

static int away_log_fd = -1;

static void
away_log (const char *fmt, ...)
{
	va_list ap;
	va_list save;
	char *buf;
	size_t mal_len, stamp_len;
	int len;
	time_t t;
	struct tm tm;

	if (away_log_fd == -1)
		return;
	__va_copy(save, ap);
	mal_len = 256;
	buf = xmalloc(mal_len);
	time(&t);
	localtime_r(&t, &tm);
	stamp_len = strftime(buf, mal_len, "[%H:%M:%S %m/%d/%Y] ", &tm);
	for (;;) {
		va_start(ap, fmt);
		len = vsnprintf(buf + stamp_len, mal_len - stamp_len, fmt, ap);
		va_end(ap);
		if (len != -1)
			break;
		mal_len <<= 1;
		buf = xrealloc(buf, mal_len);
		__va_copy(ap, save);
	}
	write(away_log_fd, buf, stamp_len + len);
	xfree(buf);
}

COMMAND(away)
{
	int f;
	char *file = "hx.away";
	char *msg;
	char buf[4096];
	u_int16_t style, away;
	int len;

	for (msg = str; *msg && *msg != ' '; msg++) ;
	if (!*msg || !*++msg)
		msg = away_log_fd == -1 ? "asf" : "fsa";
	len = snprintf(buf, sizeof(buf), away_log_fd == -1 ? "is away : %s" : "is back : %s", msg);
	style = htons(1);
	away = htons(away_log_fd == -1 ? 1 : 2);
	if (chat_front->cid) {
		u_int32_t cid = htonl(chat_front->cid);
		hlwrite(htlc, HTLC_HDR_CHAT, 0, 4,
			HTLC_DATA_STYLE, 2, &style,
			HTLC_DATA_CHAT_AWAY, 2, &away,
			HTLC_DATA_CHAT, len, buf,
			HTLC_DATA_CHAT_ID, 4, &cid);
	} else
		hlwrite(htlc, HTLC_HDR_CHAT, 0, 3,
			HTLC_DATA_STYLE, 2, &style,
			HTLC_DATA_CHAT_AWAY, 2, &away,
			HTLC_DATA_CHAT, len, buf);
	if (away_log_fd == -1) {
		f = open(file, O_WRONLY|O_CREAT|O_APPEND, S_IRUSR|S_IWUSR);
		away_log_fd = f;
	} else {
		fsync(away_log_fd);
		close(away_log_fd);
		away_log_fd = -1;
	}
}

#ifdef CONFIG_HOPE
static void
rcv_task_file_hash (struct htlc_conn *htlc)
{
	u_int8_t md5[32], haval[64], sha1[40];
	char buf[128];

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_HASH_MD5:
				memcpy(md5, dh->data, ntohs(dh->len));
				break;
			case HTLS_DATA_HASH_HAVAL:
				memcpy(haval, dh->data, ntohs(dh->len));
				break;
			case HTLS_DATA_HASH_SHA1:
				memcpy(sha1, dh->data, ntohs(dh->len));
				break;
		}
	dh_end()
	hash2str(buf, md5, 16);
	hx_printf(INFOPREFIX "md5: %32.32s\n", buf);
	hash2str(buf, haval, 16);
	hx_printf(INFOPREFIX "haval: %32.32s\n", buf);
	hash2str(buf, sha1, 20);
	hx_printf(INFOPREFIX "sha1: %40.40s\n", buf);
}

COMMAND(hash)
{
	char *lpath, *rpath, *rfile, buf[MAXPATHLEN];
	u_int16_t hldirlen;
	u_int8_t *hldir;
	int i;
	u_int32_t data_size = 0xffffffff, rsrc_size = 0xffffffff;
	u_int8_t rflt[74];
	struct stat sb;

	if (argc < 2) {
		hx_printf(INFOPREFIX "usage: %s <file> [file2...]\n", argv[0]);
		return;
	}
	for (i = 1; i < argc; i++) {
		rpath = argv[i];
		if (rpath[0] != dir_char) {
			snprintf(buf, MAXPATHLEN, "%s%c%s", htlc->rootdir, dir_char, rpath);
			rpath = buf;
		}
		rfile = dirchar_basename(rpath);
		lpath = rfile;
		task_new(htlc, rcv_task_file_hash, 0);
		if (!stat(lpath, &sb))
			data_size = sb.st_size;
		rsrc_size = resource_len(lpath);
		memcpy(rflt, "\
RFLT\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\2\
DATA\0\0\0\0\0\0\0\0\0\0\0\0\
MACR\0\0\0\0\0\0\0\0\0\0\0\0", 74);
		S32HTON(data_size, &rflt[46]);
		S32HTON(rsrc_size, &rflt[62]);
		if (rfile != rpath) {
			hldir = path_to_hldir(rpath, &hldirlen, 1);
			hlwrite(htlc, HTLC_HDR_FILE_HASH, 0, 6,
				HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
				HTLC_DATA_DIR, hldirlen, hldir,
				HTLC_DATA_RFLT, 74, rflt,
				HTLC_DATA_HASH_MD5, 0, 0,
				HTLC_DATA_HASH_HAVAL, 0, 0,
				HTLC_DATA_HASH_SHA1, 0, 0);
			xfree(hldir);
		} else {
			hlwrite(htlc, HTLC_HDR_FILE_HASH, 0, 5,
				HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
				HTLC_DATA_RFLT, 74, rflt,
				HTLC_DATA_HASH_MD5, 0, 0,
				HTLC_DATA_HASH_HAVAL, 0, 0,
				HTLC_DATA_HASH_SHA1, 0, 0);
		}
	}
}
#endif

COMMAND(hostname)
{
	if (argv[1] && argv[1][0] != '#') {
		if (strlen(argv[1]) >= sizeof(hx_hostname)) {
			hx_printf(INFOPREFIX "%s: '%s' is too long\n", argv[0], argv[1]);
		} else {
			strcpy(hx_hostname, argv[1]);
			hx_printf(INFOPREFIX "local hostname is now [%s]\n", hx_hostname);
		}
		return;
	}

	{
#if !defined(__linux__)
		char hostname[256];

#if defined(HAVE_GETHOSTNAME)
		if (gethostname(hostname, sizeof(hostname)))
#endif
			strcpy(hostname, "localhost");
		hx_printf(INFOPREFIX "local hostname is [%s]\n", hx_hostname ? hx_hostname : hostname);
#else
		char comm[200];
		FILE *fp;
		char *p = NULL, *q, *hname;
		unsigned long ip;
		struct hostent *he;
		unsigned int i = 0;
		struct in_addr inaddr;

		if (!(fp = popen("/sbin/ifconfig -a", "r"))) {
			hx_printf(INFOPREFIX "%s: /sbin/ifconfig: %s", argv[0], strerror(errno));
			return;
		}
		hx_printf(INFOPREFIX "looking for hostnames with ifconfig\n");
		while ((fgets(comm, 200, fp))) {
			if ((p = strstr(comm, "inet addr"))) {
				p += 10;
				q = strchr(p, ' ');
				*q = 0;
				if ((p && !*p) || (p && !strcmp(p, "127.0.0.1"))) continue;
				ip = inet_addr(p);
				i++;
				inaddr.s_addr = ip;
				inet_ntoa_r(inaddr, comm, sizeof(comm));
				if ((he = gethostbyaddr((char *)&ip, sizeof(ip), AF_INET)))
					hname = he->h_name;
				else
					hname = comm;
				hx_printf(" [%d] %s (%s)\n", i, hname, comm);
				if (argv[1] && argv[1][0] == '#' && i == atou32(argv[1]+1)) {
					if (strlen(hname) >= sizeof(hx_hostname)) {
						hx_printf(INFOPREFIX "%s: '%s' is too long\n", argv[0], hname);
					} else {
						strcpy(hx_hostname, hname);
						hx_printf(INFOPREFIX "local hostname is now [%s]\n", hx_hostname);
					}
					break;
				}
			}
		}
		fclose(fp);
#endif
	}
}

static int
b_read (int fd, void *bufp, size_t len)
{
	register u_int8_t *buf = (u_int8_t *)bufp;
	register int r, pos = 0;

	while (len) {
		if ((r = read(fd, &(buf[pos]), len)) <= 0)
			return -1;
		pos += r;
		len -= r;
	}

	return pos;
}

static int tracker_break;

static RETSIGTYPE
tracker_sigint (int sig __attribute__((__unused__)))
{
	tracker_break = 1;
}

static void
tracker_ready_write (int s)
{
	FD_CLR(s, &hxd_wfds);
	if (write(s, HTRK_MAGIC, HTRK_MAGIC_LEN) != HTRK_MAGIC_LEN)
		return;
	FD_SET(s, &hxd_rfds);
}

static void
tracker_ready_read (int s)
{
	u_int16_t port, nusers, nservers;
	char buf[16], name[512], desc[512];
	struct in_addr a;
	unsigned int c, j, x;
	struct sigaction act, oldact;

	memset(&act, 0, sizeof(act));
	act.sa_handler = tracker_sigint;
	sigaction(SIGINT, &act, &oldact);

	if (b_read(s, buf, 14) != 14)
		goto funk_dat;
	hx_printf(INFOPREFIX "%u servers:\n", (nservers = ntohs(*((u_int16_t *)(&(buf[10]))))));
	for (; nservers && !tracker_break; nservers--) {
		if (b_read(s, buf, 8) == -1)
			break;
		if (!buf[0]) {	/* assuming an address does not begin with 0, we can skip this */
			nservers++;
			continue;
		}
		if (b_read(s, buf+8, 3) == -1)
			break;
		a.s_addr = *((u_int32_t *)buf);
		port = ntohs(*((u_int16_t *)(&(buf[4]))));
		nusers = ntohs(*((u_int16_t *)(&(buf[6]))));
		if (b_read(s, name, (size_t)buf[10]) == -1)
			break;
		name[(size_t)buf[10]] = 0;
		CR2LF(name, (size_t)buf[10]);
		strip_ansi(name, (size_t)buf[10]);
		if (b_read(s, buf, 1) == -1)
			break;
		memset(desc, 0, sizeof(desc));
		if (b_read(s, desc, (size_t)buf[0]) == -1)
			break;
		desc[(size_t)buf[0]] = 0;
		CR2LF(desc, (size_t)buf[0]);
		strip_ansi(desc, (size_t)buf[0]);
		inet_ntoa_r(a, buf, sizeof(buf));
		hx_printf("%16s:%-5u | %5u | %s\n",
			buf, port, nusers, name);
		for (j = x = 0;;) {
			j += CO - 33;
			if (j < sizeof(desc)) {
				c = desc[j];
				desc[j] = 0;
			} else c = 0;
			hx_printf("                                 %s\n", desc+x);
			if (!c)
				break;
			if (j < sizeof(desc))
				desc[j] = c;
			x = j;
		}
	}
funk_dat:
	FD_CLR(s, &hxd_rfds);
	close(s);
	sigaction(SIGINT, &oldact, 0);
}

COMMAND(tracker)
{
	struct sockaddr_in saddr;
	u_int16_t port;
	int i, s = 0;

	if (argc < 2) {
		hx_printf("usage: %s <server> [port]\n", argv[0]);
		return;
	}

	tracker_break = 0;
	for (i = 1; i < argc; i++) {
		if (!inet_aton(argv[i], &saddr.sin_addr)) {
			struct hostent *he;

			if ((he = gethostbyname(argv[i]))) {
				size_t len = (unsigned)he->h_length > sizeof(struct in_addr)
					     ? sizeof(struct in_addr) : he->h_length;
				memcpy(&saddr.sin_addr, he->h_addr, len);
			} else {
#ifndef HAVE_HSTRERROR
				hx_printf(INFOPREFIX "DNS lookup for %s failed\n", argv[i]);
#else
				hx_printf(INFOPREFIX "DNS lookup for %s failed: %s\n", argv[i], hstrerror(h_errno));
#endif
				return;
			}
		}

		if (argv[i + 1] && isdigit(argv[i + 1][0]))
			port = atou32(argv[++i]);
		else
			port = HTRK_TCPPORT;

		if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
			continue;

		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(port);

		fd_blocking(s, 0);
		if (connect(s, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
			if (errno != EINPROGRESS) {
				hx_printf(INFOPREFIX "%s: %s: %s", argv[0], argv[i], strerror(errno));
				continue;
			}
		}
		fd_blocking(s, 1);
		hxd_files[s].conn.htrk = 0;
		hxd_files[s].ready_read = tracker_ready_read;
		hxd_files[s].ready_write = tracker_ready_write;
		FD_SET(s, &hxd_wfds);
		if (high_fd < s)
			high_fd = s;
	}
}

static int sound_on = 0;
static char *event_name[] = { "login", "message", "chat", "news", "join", "part",
		       "chat_invite", "file_done", "error" };
static char default_player[MAXPATHLEN] =
#ifdef DEFAULT_SND_PLAYER
	DEFAULT_SND_PLAYER;
#else
	"play";
#endif

COMMAND(snd)
{
	struct opt_r opt;
	char *p = 0, *s = 0;
	int i, plen, slen, on = 1, ok = 0;

        opt.err_printf = hx_printf;
        opt.ind = 0;
        while ((i = getopt_r(argc, argv, "d:p:s:01oO", &opt)) != EOF) {
                switch (i) {
			case 'd':
				strncpy(default_player, opt.arg, MAXPATHLEN-2);
				default_player[MAXPATHLEN-1] = 0;
				ok = 1;
				break;
			case 'p':
				p = opt.arg;
				break;
			case 's':
        	                s = opt.arg;
        	                break;
			case '0':
				on = 0;
				break;
			case '1':
				on = 1;
				break;
			case 'o':
				sound_on = 0;
				ok = 1;
				break;
			case 'O':
				sound_on = 1;
				ok = 1;
				break;
			default:
				goto usage;
		}
	}
	if (argc - opt.ind != 2) {
		if (ok && argc - opt.ind == 0)
			return;
usage:		hx_printf(INFOPREFIX "usage: %s [OPTIONS] <event> <beep>\n"
				     "  -d <default player> (%s)\n"
				     "  -p <player>\n"
				     "  -s <sound file>\n"
				     "  -0 turn sound for <event> off\n"
				     "  -1 turn sound for <event> on\n"
				     "  -o turn sound off\n"
				     "  -O turn sound on\n"
				     "  events are: login, message, chat, news, join, part,\n"
				     "              chat_invite, file_done, error\n"
				     "  beep (pc speaker) can be on or off\n"
			  , argv[0], default_player);
                return;
        }
	for (i = 0; i < nsounds; i++) {
		if (!strcmp(argv[opt.ind], event_name[i]))
			break;
	}
	if (i == nsounds)
		goto usage;
	opt.ind++;
	if (!strcmp(argv[opt.ind], "on")) 
		sounds[i].beep = 1;
	else {
		if (!strcmp(argv[opt.ind], "off"))
			sounds[i].beep = 0;
		else 
			goto usage;
	}
	slen = s ? strlen(s) : 0;
	plen = p ? strlen(p) : 0;
	if (slen > MAXPATHLEN || plen > MAXPATHLEN) {
		hx_printf(INFOPREFIX "file path too long.");
		return;
	}
	if (s) {
		if (sounds[i].sound) 
			xfree(sounds[i].sound);
		sounds[i].sound = xmalloc(slen+1);
		strcpy(sounds[i].sound, s);
	}
	if (p) {
		if (sounds[i].player) 
			xfree(sounds[i].player);
		sounds[i].player = xmalloc(plen+1);
		strcpy(sounds[i].player, p);
	}
	sounds[i].on = on;
}

static void
play_sound (snd_events event)
{
	char buff[2 * MAXPATHLEN + 20];
	pid_t pid;

	if (!sound_on || !sounds[event].on)
		return;
	if (sounds[event].sound) {
		sprintf(buff, "%s %s &>/dev/null", 
			sounds[event].player ? sounds[event].player : default_player, 
			sounds[event].sound);
		pid = fork();
		if (pid == -1)
			return;
		sounds[event].pid = pid;
		if (pid == 0) {
			char *argv[4];

			argv[0] = "sh";
			argv[1] = "-c";
			argv[2] = buff;
			argv[3] = 0;
			execve("/bin/sh", argv, hxd_environ);
			_exit(127);
		}
	} else 
		if (sounds[event].beep)
			write(1, "\a", 1);
}

struct hx_command {
	char *name;
	void (*fun)();
};

static struct hx_command *commands, *last_command;

COMMAND(help)
{
	struct hx_command *cmd;
	unsigned int i;
	int r;
	char buf[2048], *bufp;

	for (i = 0, bufp = buf, cmd = commands; cmd <= last_command; cmd++) {
		i++;
		r = snprintf(bufp, sizeof(buf) - (bufp - buf), "  %-8s", cmd->name);
		if (r == -1 || sizeof(buf) <= (unsigned)(bufp + r - buf))
			break;
		bufp += r;
		if (!(i % 5))
			*bufp++ = '\n';
	}
	hx_printf(INFOPREFIX "commands:\n%s%c", buf, (i % 5) ? '\n' : 0);
}

static struct hx_command __commands[] = {
	{ "away",	cmd_away },
	{ "cd",		cmd_cd },
	{ "chat",	cmd_chat },
	{ "chats",	cmd_chats },
	{ "clear",	cmd_clear },
	{ "close",	cmd_close },
	{ "dirchar",	cmd_dirchar },
	{ "exec",	cmd_exec },
	{ "finfo",	cmd_finfo },
	{ "get",	cmd_get },
#if defined(CONFIG_HAL)
	{ "hal",	cmd_hal },
#endif
#if defined(CONFIG_HOPE)
	{ "hash",	cmd_hash },
#endif
	{ "help",	cmd_help },
	{ "hostname",	cmd_hostname },
	{ "icon",	cmd_icon },
	{ "info",	cmd_info },
	{ "invite",	cmd_invite },
	{ "join",	cmd_join },
	{ "kick",	cmd_kick },
	{ "lcd",	cmd_lcd },
	{ "leave",	cmd_part },
	{ "ln",		cmd_ln },
	{ "load",	cmd_load },
	{ "lpwd",	cmd_lpwd },
	{ "ls",		cmd_ls },
	{ "me",		cmd_me },
	{ "mkdir",	cmd_mkdir },
	{ "msg",	cmd_msg },
	{ "mv",		cmd_mv },
	{ "news",	cmd_news },
	{ "nick",	cmd_nick },
	{ "part",	cmd_part },
	{ "post",	cmd_post },
	{ "put",	cmd_put },
	{ "pwd",	cmd_pwd },
	{ "quit",	cmd_quit },
	{ "rm",		cmd_rm },
	{ "save",	cmd_save },
	{ "server",	cmd_server },
	{ "snd",	cmd_snd },
	{ "subject",	cmd_subject },
	{ "tracker",	cmd_tracker },
	{ "type",	cmd_type },
	{ "who",	cmd_who },
	{ "xfers",	cmd_xfers },
	{ "xfgo",	cmd_xfgo },
	{ "xfkill",	cmd_xfkill }
};

static struct hx_command
	*commands = __commands,
	*last_command = __commands + sizeof(__commands) / sizeof(struct hx_command) - 1;

static short command_hash[26];

void
gen_command_hash (void)
{
	int i, n;
	struct hx_command *cmd;

	cmd = commands;
	for (n = 0, i = 0; i < 26; i++) {
		if (cmd->name[0] == i + 'a') {
			command_hash[i] = n;
			do {
				if (++cmd > last_command) {
					for (i++; i < 26; i++)
						command_hash[i] = -1;
					return;
				}
				n++;
			} while (cmd->name[0] == i + 'a');
		} else
			command_hash[i] = -1;
	}
}

int
expand_command (char *cmdname, int len, char *cmdbuf)
{
	struct hx_command *cmd;
	char *ename = 0;
	int elen = 0, ambig = 0;

	for (cmd = commands; cmd <= last_command; cmd++)
		if (!strncmp(cmd->name, cmdname, len)) {
			if (ename) {
				if ((int)strlen(cmd->name) < elen)
					elen = strlen(cmd->name);
				for (elen = len; cmd->name[elen] == ename[elen]; elen++);
				ambig = 1;
			} else {
				ename = cmd->name;
				elen = strlen(ename);
			}
		}

	if (ename)
		strncpy(cmdbuf, ename, elen);
	else
		ambig = -1;
	cmdbuf[elen] = 0;

	return ambig;
}

u_int32_t
cmd_arg (int argn, char *str)
{
	char *p, *cur;
	char c, quote = 0;
	int argc = 0;
	u_int32_t offset = 0, length = 0;

	p = str;
	while (isspace(*p)) p++;
	for (cur = p; (c = *p); ) {
		if (c == '\'' || c == '"') {
			if (quote == c) {
				argc++;
				if (argn == argc) {
					p++;
					while (isspace(*p)) p++;
					offset = p - str;
					p--;
				} else if (argn + 1 == argc) {
					length = p - (str + offset);
					break;
				} else {
					p++;
					while (isspace(*p)) p++;
					p--;
				}
				quote = 0;
				cur = ++p;
			} else if (!quote) {
				quote = c;
				cur = ++p;
			}
		} else if (!quote && isspace(c)) {
			argc++;
			if (argn == argc) {
				p++;
				while (isspace(*p)) p++;
				offset = p - str;
				p--;
			} else if (argn + 1 == argc) {
				length = p - (str + offset);
				break;
			} else {
				p++;
				while (isspace(*p)) p++;
				p--;
			}
			cur = ++p;
		} else if (c == '\\' && *(p+1) == ' ') {
			p += 2;
		} else p++;
	}
	if (p != cur)
		argc++;
	if (argn == argc && 0 && argn != 1) {
		cur--;
		offset = cur - str;
		length = strlen(cur);
	} else if (argn + 1 == argc) {
		length = p - (str + offset);
	}

	return (offset << 16) | (length & 0xffff);
}

#define killspace(s) while (isspace(*(s))) strcpy((s), (s)+1)
#define add_arg(s)									\
	do {										\
		argv[argc++] = s;							\
		if (argc >= 16) {							\
			if (argc == 16) {						\
				argv = xmalloc(sizeof(char *) * 17);			\
				memcpy(argv, auto_argv, sizeof(char *) * 16);		\
			} else								\
				argv = xrealloc(argv, ((sizeof(char *) * argc) + 1));	\
		}									\
	} while (0)

void
hx_command (char *str)
{
	int i;
	struct hx_command *cmd;
	char *p;

	for (p = str; *p && !isspace(*p); p++)
		;
	if (*str < 'a' || *str > 'z')
		goto notfound;
	i = *str - 'a';
	if (command_hash[i] == -1)
		goto notfound;
	cmd = commands + command_hash[i];
	do {
		if (!strncmp(str, cmd->name, p - str) && cmd->fun) {
			char *p, *cur, *s;
			char c, quote = 0;
			char *auto_argv[16], **argv = auto_argv;
			int argc = 0;

			s = xstrdup(str);
			killspace(s);
			for (p = cur = s; (c = *p); p++) {
				if (c == '\'' || c == '"') {
					if (quote == c) {
						*p = 0;
						add_arg(cur);
						killspace(p+1);
						quote = 0;
						cur = p+1;
					} else if (!quote) {
						quote = c;
						cur = p+1;
					}
				} else if (!quote && isspace(c)) {
					*p = 0;
					add_arg(cur);
					killspace(p+1);
					cur = p+1;
				} else if (c == '\\') {
					chrexpand(p, strlen(p));
				}
			}
			if (p != cur)
				add_arg(cur);
			argv[argc] = 0;

			cmd->fun(argc, argv, str, &hx_htlc);
			xfree(s);
			if (argv != auto_argv)
				xfree(argv);
			return;
		}
		cmd++;
	} while (cmd <= last_command && cmd->name[0] == *str);

notfound:
	hx_printf(INFOPREFIX "%.*s: command not found\n", p - str, str);
}

void
hx_send_chat (char *str)
{
	if (chat_front->cid) {
		u_int32_t cid = htonl(chat_front->cid);
		hlwrite(&hx_htlc, HTLC_HDR_CHAT, 0, 2,
			HTLC_DATA_CHAT, strlen(str), str,
			HTLC_DATA_CHAT_ID, 4, &cid);
	} else
		hlwrite(&hx_htlc, HTLC_HDR_CHAT, 0, 1,
			HTLC_DATA_CHAT, strlen(str), str);
}
