#include "hx.h"
#include "hxd.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include "xmalloc.h"
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <signal.h>
#include <pthread.h>

static pid_t parent_pid;
static pthread_t gtk_thread_id;
extern int loopZ_exit;
static int gtk_suck = 0;
#define tell_gtk_that_it_sucks(a)	do { gtk_suck = (a); } while (0)

static GtkWidget *chat_window;
static GtkWidget *chat_output_text;
static GtkWidget *chat_input_text;

int CO = 80;

static void
chat_close (GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
	kill(parent_pid, SIGINT);
}

static char *termed_buf = 0;
static size_t termed_buf_siz = 0;

static void
chat_input_activate (GtkWidget *widget, gpointer data)
{
	GtkText *text;
	guint point, len;
	gchar *chars;

	text = GTK_TEXT(widget);
	point = gtk_text_get_point(text);
	len = gtk_text_get_length(text);
	chars = gtk_editable_get_chars(GTK_EDITABLE(text), 0, len);
	if (termed_buf_siz < len+1) {
		termed_buf_siz = len+1;
		termed_buf = xrealloc(termed_buf, len+1);
	}
	memcpy(termed_buf, chars, len);
	termed_buf[len] = 0;
	g_free(chars);
	gdk_threads_leave();
	hotline_client_input(termed_buf);
	gdk_threads_enter();

	gtk_text_set_point(text, 0);
	gtk_text_forward_delete(text, len);
	gtk_text_set_editable(text, 1);

	/* wake up select in loopZ */
	kill(parent_pid, SIGURG);
}

guint insert_id;
guint tab_point, tab_len;

static void
chat_input_insert (GtkWidget *widget, gchar *str, int len, gpointer ptr)
{
	tab_point = gtk_text_get_point(GTK_TEXT(widget));
	tab_len = len;
	tell_gtk_that_it_sucks(2);
	gtk_signal_disconnect(GTK_OBJECT(widget), insert_id);
}

/*
 * sniff the keys coming to the text widget
 * this is called before the text's key_press function is called
 * gtk will emit an activate signal on return only when control
 * is held or the text is not editable.
 */
static void
chat_input_key_press (GtkWidget *widget, GdkEventKey *event)
{
	GtkText *text;
	guint point, len;
	guint k;

	k = event->keyval;
	if (k == GDK_Return) {
		text = GTK_TEXT(widget);
		point = gtk_text_get_point(text);
		len = gtk_text_get_length(text);
		if (len == point) {
			gtk_text_set_editable(text, 0);
		}
	} else if (k == GDK_Tab && (event->state & GDK_CONTROL_MASK)) {
		gtk_container_focus(GTK_CONTAINER(chat_window), GTK_DIR_TAB_BACKWARD);
		text = GTK_TEXT(widget);
		insert_id = gtk_signal_connect(GTK_OBJECT(text), "insert_text", GTK_SIGNAL_FUNC(chat_input_insert), 0);
	}
}

static GdkColor colors[8];

static void
init_colors (GtkWidget *widget)
{
	int i;

	colors[0].red = 0x0000;
	colors[0].green = 00000;
	colors[0].blue = 0x0000;
	colors[1].red = 0xffff;
	colors[1].green = 0x0000;
	colors[1].blue = 0x0000;
	colors[2].red = 0x0000;
	colors[2].green = 0xffff;
	colors[2].blue = 0x0000;
	colors[3].red = 0xffff;
	colors[3].green = 0xffff;
	colors[3].blue = 0x0000;
	colors[4].red = 0x0000;
	colors[4].green = 0x0000;
	colors[4].blue = 0xffff;
	colors[5].red = 0xffff;
	colors[5].green = 0x0000;
	colors[5].blue = 0xffff;
	colors[6].red = 0x0000;
	colors[6].green = 0xffff;
	colors[6].blue = 0xffff;
	colors[7].red = 0xffff;
	colors[7].green = 0xffff;
	colors[7].blue = 0xffff;

	for (i = 0; i < 8; i++) {
		colors[i].pixel = (gulong)((colors[i].red & 0xff00) * 256
				+ (colors[i].green & 0xff00)
				+ (colors[i].blue & 0xff00) / 256);
		if (!gdk_color_alloc(gtk_widget_get_colormap(widget), &colors[i]))
			hxd_log("alloc color failed");
	}
}

static int
fe_main ()
{
	GtkWidget *window;
	GtkWidget *vbox, *hbox;
	GtkWidget *text;
	GtkWidget *vscrollbar;
	GtkStyle *style;

	gdk_threads_enter();
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	chat_window = window;
	gtk_widget_set_usize(window, 640, 480);
	gtk_window_set_policy(GTK_WINDOW(window), 1, 1, 0);  
	gtk_signal_connect(GTK_OBJECT(window), "destroy",
			   GTK_SIGNAL_FUNC(chat_close), 0);
	gtk_window_set_title(GTK_WINDOW(window), "Chat");
	gtk_container_set_border_width(GTK_CONTAINER(window), 0);

	vbox = gtk_vbox_new(0, 0);
	gtk_container_add(GTK_CONTAINER(window), vbox);
	gtk_widget_show(vbox);

	hbox = gtk_hbox_new(0, 0);
	gtk_widget_set_usize(hbox, 500, 400);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 10);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 1, 1, 0);
	gtk_widget_show(hbox);

	init_colors(window);

	style = gtk_style_new();
	style->base[GTK_STATE_NORMAL] = colors[0];
	style->bg[GTK_STATE_NORMAL] = colors[0];
	style->fg[GTK_STATE_NORMAL] = colors[7];
	style->text[GTK_STATE_NORMAL] = colors[7];
	style->light[GTK_STATE_NORMAL] = colors[7];
	style->dark[GTK_STATE_NORMAL] = colors[0];
	style->font = gdk_font_load ("-misc-fixed-medium-r-*-*-*-140-*-*-*-*-*-*");

	if (0) {
		int i;
		GdkGC *fg_gc, *bg_gc;

		fg_gc = style->fg_gc[GTK_STATE_NORMAL];
		bg_gc = style->bg_gc[GTK_STATE_NORMAL];
		style->black_gc = bg_gc;
		style->white_gc = fg_gc;
		for (i = 0; i < 5; i++) {
			style->fg_gc[i] = bg_gc;
			style->bg_gc[i] = fg_gc;
			style->base_gc[i] = fg_gc;
			style->text_gc[i] = bg_gc;
			style->light_gc[i] = bg_gc;
			style->dark_gc[i] = fg_gc;
		}
	}

	text = gtk_text_new(0, 0);
	gtk_widget_set_style(text, style);
	chat_output_text = text;
	gtk_text_set_editable(GTK_TEXT(text), 0);
	gtk_text_set_word_wrap(GTK_TEXT(text), 1);
	gtk_box_pack_start(GTK_BOX(hbox), text, 1, 1, 0);
	gtk_widget_show(text);

	vscrollbar = gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
	gtk_box_pack_start(GTK_BOX(hbox), vscrollbar, 0, 0, 0);
	gtk_widget_show(vscrollbar);

	hbox = gtk_hbox_new(0, 0);
	gtk_widget_set_usize(hbox, 140, 80);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 10);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 1, 1, 0);
	gtk_widget_show(hbox);

	text = gtk_text_new(0, 0);
	chat_input_text = text;
	gtk_widget_set_style(text, style);
	gtk_signal_connect(GTK_OBJECT(text), "key_press_event", GTK_SIGNAL_FUNC(chat_input_key_press), 0);
	gtk_signal_connect(GTK_OBJECT(text), "activate", GTK_SIGNAL_FUNC(chat_input_activate), 0);
	gtk_text_set_editable(GTK_TEXT(text), 1);
	gtk_text_set_word_wrap(GTK_TEXT(text), 1);
	gtk_box_pack_start(GTK_BOX(hbox), text, 1, 1, 0);
	gtk_widget_show(text);

	gtk_widget_show(window);

	gdk_threads_leave();
	while (!loopZ_exit) {
		gdk_threads_enter();
		gtk_main();
		gdk_threads_leave();
		while (gtk_suck) {
			usleep(10000);
			//sched_yield();
		}
	}

	return 0;
}

static gint
gtk_sucks_idle (void *__asf)
{
	if (gtk_suck) {
		if (gtk_suck == 2) {
			if (GTK_TEXT_INDEX(GTK_TEXT(chat_input_text), tab_point) == '\t' && gtk_text_get_point(GTK_TEXT(chat_input_text)) == tab_point+1) {
				gtk_text_set_point(GTK_TEXT(chat_input_text), tab_point);
				gtk_text_forward_delete(GTK_TEXT(chat_input_text), tab_len);
				gtk_suck = 0;
			}
		} else
			gtk_main_quit();
	} else {
		//sched_yield();
		usleep(10000);
	}

	return 1;
}

static void *
init_thread (void *__arg)
{
	gtk_idle_add((GtkFunction)gtk_sucks_idle, 0);
	fe_main();

	return 0;
}

static void
sig_handler (int sig)
{
	if (sig == SIGINT) {
		loopZ_exit = 1;
	}
}

static void
init (int argc, char **argv)
{
	int err;
	struct sigaction act;
	pthread_t tid;

	act.sa_handler = sig_handler;
	act.sa_flags = 0;
	sigfillset(&act.sa_mask);
	sigaction(SIGINT, &act, 0);
	sigaction(SIGURG, &act, 0);

	parent_pid = getpid();
	g_thread_init(0);
	gtk_init(&argc, &argv);
	err = pthread_create(&tid, 0, init_thread, 0);
	if (err) {
		hxd_log("pthread_create: %s\n", strerror(err));
		exit(err);
	}
	gtk_thread_id = tid;
}

static int do_lf;

void
hx_printf (const char *fmt, ...)
{
	GtkAdjustment *adj;
	GtkWidget *text;
	guint val, scroll;
	char *p, *realtext, numstr[4];
	int i, bright;
	GdkColor *bgcolor, *fgcolor;
	va_list ap;
	va_list save;
	char autobuf[256], *buf;
	size_t mal_len;
	int len;

	__va_copy(save, ap);
	mal_len = 256;
	buf = autobuf;
	for (;;) {
		va_start(ap, fmt);
		len = vsnprintf(buf, mal_len, fmt, ap);
		va_end(ap);
		if (len != -1)
			break;
		__va_copy(ap, save);
		mal_len <<= 1;
		if (buf == autobuf)
			buf = xmalloc(mal_len);
		else
			buf = xrealloc(buf, mal_len);
	}

	tell_gtk_that_it_sucks(1);
	gdk_threads_enter();
	text = chat_output_text;
	adj = (GTK_TEXT(text))->vadj;
	val = adj->upper - adj->lower - adj->page_size;
	scroll = adj->value == val;
	gtk_text_freeze(GTK_TEXT(text));
	fgcolor = &text->style->fg[GTK_STATE_NORMAL];
	bgcolor = &text->style->bg[GTK_STATE_NORMAL];
	if (do_lf)
		gtk_text_insert(GTK_TEXT(text), 0, fgcolor, bgcolor, "\n", 1);
	realtext = buf;
	for (p = buf; *p; p++) {
		if (*p == '\033' && *(p+1) == '[') {
			gtk_text_insert(GTK_TEXT(text), 0, fgcolor, bgcolor, realtext, p - realtext);
			p += 2;
			i = 0;
			while (i < 2 && ((*p >= '0' && *p <= '7') || *p == ';')) {
				if (*p == ';') {
					bright = i;
					i = 0;
					p++;
					continue;
				}
				numstr[i++] = *p;
				p++;
			}
			realtext = p+1;
			if (i) {
				numstr[i--] = 0;
				i = i == 1 ? numstr[i] - '0' : 0;
				i += 8 * (numstr[0] - '0');
				if (!i) {
					fgcolor = &text->style->fg[GTK_STATE_NORMAL];
					bgcolor = &text->style->bg[GTK_STATE_NORMAL];
					continue;
				}
				i &= 0xf;
				if (i >= 8)
					fgcolor = &colors[i-8];
				else
					bgcolor = &colors[i];
			}
		}
	}
	if (*(p-1) == '\n') {
		gtk_text_insert(GTK_TEXT(text), 0, fgcolor, bgcolor, realtext, p - realtext - 1);
		do_lf = 1;
	} else {
		gtk_text_insert(GTK_TEXT(text), 0, fgcolor, bgcolor, realtext, p - realtext);
		do_lf = 0;
	}
	gtk_text_thaw(GTK_TEXT(text));
	if (scroll)
		gtk_adjustment_set_value(adj, adj->upper - adj->lower - adj->page_size);
	gdk_threads_leave();
	tell_gtk_that_it_sucks(0);

	if (buf != autobuf)
		xfree(buf);
}

void
hx_save (char *filename)
{
	GtkWidget *text;
	char *chars;
	int f;
	ssize_t r;
	size_t pos, len;
	char path[MAXPATHLEN];

	expand_tilde(path, filename);
	f = open(path, O_WRONLY|O_CREAT|O_APPEND, S_IRUSR|S_IWUSR);
	if (f < 0) {
		hx_printf("save: %s: %s\n", path, strerror(errno));
		return;
	}
	pos = 0;
	tell_gtk_that_it_sucks(1);
	gdk_threads_enter();
	text = chat_output_text;
	len = gtk_text_get_length(GTK_TEXT(text));
	chars = gtk_editable_get_chars(GTK_EDITABLE(text), 0, len);
	gdk_threads_leave();
	tell_gtk_that_it_sucks(0);
	while (len) {
		r = write(f, chars + pos, len);
		if (r <= 0)
			break;
		pos += r;
		len -= r;
	}
	fsync(f);
	close(f);
	hx_printf(INFOPREFIX "%d bytes written to %s\n", pos, path);
}

static void
cleanup ()
{
	loopZ_exit = 1;
	gtk_main_quit();
	kill(parent_pid, SIGINT);
}

static void
status ()
{
}

static void
clear ()
{
}

static void
clear_input ()
{
}

static void
term_mode_underline ()
{
}

static void
term_mode_clear ()
{
}

static void
output_chat (u_int32_t cid, char *chat, u_int16_t chatlen)
{
	if (cid)
		hx_printf("0x%08x: %.*s\n", cid, chatlen, chat);
	else
		hx_printf("%.*s\n", chatlen, chat);
}

static void
output_msg (u_int32_t uid, char *name, char *msgbuf, u_int16_t msglen)
{
	char *col;
	struct user *user;

	user = user_with_uid(chat_list->user_list, uid);
	col = user ? (user->color % 4 > 1 ? (user->color % 2 ? RED : RED_BOLD)
					  : (user->color % 2 ? GREEN : GREEN_BOLD))
		   : GREEN;
	hx_printf("[%s%s%s(%u)]: %.*s\n", col, name, DEFAULT, uid, msglen, msgbuf);
}

static void
output_agreement (char *agreement, u_int16_t len)
{
	hx_printf(INFOPREFIX "Agreement:\n%.*s\n", len, agreement);
}

struct output_functions hx_output = {
	init,
	cleanup,
	status,
	clear,
	clear_input,
	term_mode_underline,
	term_mode_clear,
	output_chat,
	output_msg,
	output_agreement
};
