#include "hx.h"
#include "xmalloc.h"
#include <string.h>

struct user *
user_new (struct user **utailp)
{
	struct user *user, *tail = *utailp;

	user = xmalloc(sizeof(struct user));
	memset(user, 0, sizeof(struct user));

	user->next = 0;
	user->prev = tail;
	tail->next = user;
	tail = user;
	*utailp = tail;

	return user;
}

void
user_delete (struct user **utailp, struct user *user)
{
	if (user->next)
		user->next->prev = user->prev;
	if (user->prev)
		user->prev->next = user->next;
	if (*utailp == user)
		*utailp = user->prev;
	xfree(user);
}

struct user *
user_with_uid (struct user *ulist, u_int32_t uid)
{
	struct user *userp;

	for (userp = ulist->next; userp; userp = userp->next)
		if (userp->uid == uid)
			return userp;

	return 0;
}

struct user *
user_with_name (struct user *ulist, u_int8_t *name)
{
	int ulen, uplen, nlen;
	struct user *userp, *user = 0;

	nlen = strlen(name);
	for (userp = ulist->next; userp; userp = userp->next) {
		if (!strncmp(userp->name, name, nlen)) {
			if (user) {
				ulen = strlen(user->name);
				uplen = strlen(userp->name);
				if (!((uplen == nlen && ulen != nlen) || (ulen == nlen && uplen != nlen)))
					return 0;
			}
			user = userp;
		}
	}

	return user;
}
