#ifndef _HXD_H
#define _HXD_H

#if defined(HAVE_CONFIG_H)
#include "config.h"
#endif
#if !defined(__GNUC__) || defined(__STRICT_ANSI__) || defined(__APPLE_CC__)
#define __attribute__(x)
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#if defined(HAVE_SYS_SELECT_H)
#include <sys/select.h>
#endif
#include <limits.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <stdlib.h>
#include "hotline.h"

#if defined(CONFIG_HTXF_PTHREAD)
#include <pthread.h>
#endif

#ifndef MAXPATHLEN
#ifdef PATH_MAX
#define MAXPATHLEN PATH_MAX
#else
#define MAXPATHLEN 4095
#endif
#endif

#if MAXPATHLEN > 4095
#undef MAXPATHLEN
#define MAXPATHLEN 4095
#endif

struct hxd_config {
	struct hxd_config_limits {
		int total_downloads;
		int total_uploads;
		int individual_downloads;
		int individual_uploads;
		int out_Bps;
	} limits;
	struct hxd_config_paths {
		char *files;
		char *accounts;
		char *exec;
		char *news;
		char *agreement;
		char *log;
		char *banlist;
		char *tracker_banlist;
	} paths;
	struct hxd_config_tracker {
		char *name;
		char *description;
		char **trackers;
		int interval;
		int nusers;
		char *password;
	} tracker;
	struct hxd_config_options {
		int htls_port;
		int uid;
		int gid;
		int detach;
		char **addresses;
		char *timeformat;
		int away_time;
		int ident;
	} options;
	struct hxd_config_permissions {
		int umask;
		int files;
		int directories;
		int account_files;
		int account_directories;
		int log_files;
		int news_files;
	} permissions;
	struct hxd_config_network {
		char **logins;
		char **connect_to;
	} network;
	struct hxd_config_files {
		char *comment;
		int fork;
	} files;
};

extern struct hxd_config hxd_cfg;

extern void hxd_read_config (char *file, void *mem);
extern void init_hxd_cfg (struct hxd_config *cfg);

struct qbuf {
	u_int32_t pos, len;
	u_int8_t *buf;
};

struct htlc_chat {
	struct htlc_chat *next, *prev;
	u_int32_t ref;
	u_int8_t subject[256];
	fd_set fds;
	u_int32_t nusers;
};

struct extra_access_bits {
	u_int32_t chat_private:1,
		  msg:1,
		  user_getlist:1,
		  file_list:1,
		  file_getinfo:1,
		  file_hash:1,
		  can_login:1,
		  is_server:1,
		  user_visibility:1,
		  user_color:1,
		  __reserved:22;
};

struct htlc_conn;

struct htxf_conn {
	u_int32_t data_size, data_pos, rsrc_size, rsrc_pos;
	u_int32_t total_size, total_pos;
	u_int32_t ref;
	u_int16_t preview;
	u_int8_t gone;
	u_int8_t type;
	int fd;
#if defined(CONFIG_HTXF_FORK)
	int pipe;
#endif
#if defined(CONFIG_HTXF_PTHREAD)
	pthread_t tid;
#else
	pid_t pid;
#endif
	struct sockaddr_in sockaddr;
	struct sockaddr_in listen_sockaddr;
	struct htlc_conn *htlc;
	char path[MAXPATHLEN];
	char remotepath[MAXPATHLEN];
	struct qbuf in;
	void *stack;
	char **filter_argv;
	struct timeval start;
	struct {
		u_int32_t retry:1,
			  reserved:31;
	} opt;
};

struct htlc_conn {
	struct htlc_conn *next, *prev;
	struct htlc_conn *server_htlc;
	void (*rcv)(struct htlc_conn *);
	void (*real_rcv)(struct htlc_conn *);
	struct qbuf in, out;
	struct sockaddr_in sockaddr;
	int fd, identfd;
	u_int32_t trans;
	u_int32_t icon;
	u_int16_t uid;
	u_int16_t color;
	struct {
		u_int32_t visible:1, reserved:31;
	} flags; 
	struct hl_access_bits access;
	struct extra_access_bits access_extra;
#ifdef CONFIG_HOPE
	u_int8_t macalg[32];
	u_int8_t sessionkey[64];
	u_int16_t sklen;
#endif
#define AWAY_PERM		1
#define AWAY_INTERRUPTABLE	2
#define AWAY_INTERRUPTED	3
	u_int16_t away;
	u_int16_t current_uid;
	char rootdir[MAXPATHLEN], newsfile[MAXPATHLEN], dropbox[MAXPATHLEN];
	u_int8_t userid[513];
	u_int8_t name[32];
	u_int8_t login[32];
	u_int16_t nr_puts, put_limit, nr_gets, get_limit;
	u_int32_t limit_out_Bps;
#define HTXF_PUT_MAX	4
#define HTXF_GET_MAX	4
	struct htxf_conn *htxf_in[HTXF_PUT_MAX], *htxf_out[HTXF_GET_MAX];
#if defined(CONFIG_HTXF_PTHREAD)
	pthread_mutex_t htxf_mutex;
#elif defined(CONFIG_HTXF_CLONE)
	int htxf_spinlock;
#endif
};

#if defined(CONFIG_HTXF_PTHREAD)
#define LOCK_HTXF(htlc)		pthread_mutex_lock(&(htlc)->htxf_mutex)
#define UNLOCK_HTXF(htlc)	pthread_mutex_unlock(&(htlc)->htxf_mutex)
#define INITLOCK_HTXF(htlc)	pthread_mutex_init(&(htlc)->htxf_mutex, 0)
#elif defined(CONFIG_HTXF_CLONE)
#include <sched.h>
#define LOCK_HTXF(htlc)		do { volatile int *sl = &(htlc)->htxf_spinlock; while (*sl) ; *sl = 1; } while (0)
#define UNLOCK_HTXF(htlc)	do { (htlc)->htxf_spinlock = 0; sched_yield(); } while (0)
#define INITLOCK_HTXF(htlc)	do { (htlc)->htxf_spinlock = 0; } while (0)
#else
#define LOCK_HTXF(htlc)
#define UNLOCK_HTXF(htlc)
#define INITLOCK_HTXF(htlc)
#endif

struct htrk_conn {
	struct qbuf out;
};

struct hxd_file {
	union {
		void *ptr;
		struct htlc_conn *htlc;
		struct htrk_conn *htrk;
		struct htxf_conn *htxf;
	} conn;
	void (*ready_read)(int fd);
	void (*ready_write)(int fd);
};

extern struct hxd_file *hxd_files;

extern int hxd_open_max;

extern int high_fd;

extern fd_set hxd_rfds, hxd_wfds;

extern char *version;

extern char **hxd_environ;

extern struct in_addr hxd_inaddr;

extern struct htlc_conn *htlc_list, *htlc_tail;
extern struct htlc_conn *server_htlc_list, *server_htlc_tail;

extern u_int16_t nhtlc_conns;

extern u_int16_t nr_gets;
extern u_int16_t nr_puts;

#if (defined(CONFIG_HOTLINE_SERVER) || defined(CONFIG_HOTLINE_CLIENT)) && !defined(HAVE_INET_NTOA_R)
extern int inet_ntoa_r (struct in_addr in, char *buf, size_t buflen);
#endif
#ifndef HAVE_INET_ATON
extern int inet_aton (const char *cp, struct in_addr *ia);
#endif
#ifndef HAVE_LOCALTIME_R
#include <time.h>
extern struct tm *localtime_r (const time_t *t, struct tm *tm);
#endif
#if defined(CONFIG_HOTLINE_SERVER)
#if !defined(HAVE_STRCASESTR)
extern char *strcasestr (char *haystack, char *needle);
#endif
#if !defined(HAVE_REALPATH)
extern char *realpath (const char *pathname, char *result);
#endif
#if !defined(HAVE_STRPTIME)
extern char *strptime(char *, const char *, struct tm *);
#endif
#endif /* CONFIG_HOTLINE_SERVER */
#if !defined(HAVE_SNPRINTF) || defined(__hpux__)
extern int snprintf (char *str, size_t count, const char *fmt, ...);
#endif
#if !defined(HAVE_VSNPRINTF) || defined(__hpux__)
#include <stdarg.h>
extern int vsnprintf (char *str, size_t count, const char *fmt, va_list ap);
#endif
#if defined(CONFIG_HOTLINE_CLIENT) && !defined(HAVE_BASENAME)
extern char *basename (char *path);
#endif

#ifndef RETSIGTYPE
#define RETSIGTYPE void
#endif

#if defined(CONFIG_TRACKER_REGISTER)
extern void tracker_register_timer (void *__arg);
#endif

extern int fd_blocking (int fd, int on);
extern int fd_closeonexec (int fd, int on);
extern int fd_lock_write (int fd);

extern void hxd_log (const char *fmt, ...);

extern void timer_add (struct timeval *tv, void (*fn)(), void *ptr);
extern void timer_delete_ptr (void *ptr);
extern void timer_add_secs (time_t secs, void (*fn)(), void *ptr);

extern void qbuf_set (struct qbuf *q, u_int32_t pos, u_int32_t len);
extern void qbuf_add (struct qbuf *q, void *buf, u_int32_t len);

extern void htlc_close (struct htlc_conn *htlc);

extern void start_ident (struct htlc_conn *htlc);
extern int check_banlist (struct htlc_conn *htlc);
extern void addto_banlist (const char *name, const char *login, const char *user, const char *address, const char *message);

extern void snd_strerror (struct htlc_conn *htlc, int err);

extern void hlwrite (struct htlc_conn *htlc, u_int32_t type, u_int32_t flag, int hc, ...);
extern void hl_code (void *__dst, const void *__src, size_t len);
#define hl_decode(d,s,l) hl_code(d,s,l)
#define hl_encode(d,s,l) hl_code(d,s,l)

extern int account_read (const char *login, char *password, char *name, struct hl_access_bits *acc);
extern int account_write (const char *login, const char *password, const char *name, const struct hl_access_bits *acc);
extern int account_delete (const char *login);
extern int account_trusted (const char *login, const char *userid, const char *addr);
extern void account_getconf (struct htlc_conn *htlc);
extern void account_get_access_extra (struct htlc_conn *htlc);

extern void news_send_file (struct htlc_conn *htlc);
extern void news_save_post (char *newsfile, u_int8_t *buf, u_int16_t len);
extern void agreement_send_file (struct htlc_conn *htlc);

extern int chat_isset (struct htlc_conn *, struct htlc_chat *);
extern struct htlc_chat *chat_lookup_ref (u_int32_t);
extern void chat_remove_from_all (struct htlc_conn *htlc);

extern void command_chat (struct htlc_conn *htlc, char *chatbuf);

extern void toggle_away (struct htlc_conn *htlc);
extern void test_away (struct htlc_conn *htlc);
extern void away_timer (struct htlc_conn *htlc);

#define atou32(_str) ((u_int32_t)strtoul(_str, 0, 0))
#define atou16(_str) ((u_int16_t)strtoul(_str, 0, 0))

static inline void
memory_copy (void *__dst, void *__src, unsigned int len)
{
	u_int8_t *dst = __dst, *src = __src;

	for (; len; len--)
		*dst++ = *src++;
}

#define DHARGS_BUF_SIZE	(0x10000 + SIZEOF_HL_DATA_HDR)
/* extern u_int8_t _dhargs_buf[DHARGS_BUF_SIZE]; */
extern u_int8_t *_dhargs_buf;

#if 1
/* data must be accessed from locations that are aligned on multiples of the data size */
#define dh_start(_htlc)													\
{															\
	struct hl_data_hdr *dh;												\
	u_int32_t _pos, _max;												\
	memory_copy(_dhargs_buf, &((_htlc)->in.buf[SIZEOF_HL_HDR]), SIZEOF_HL_DATA_HDR);				\
	dh = (struct hl_data_hdr *)_dhargs_buf;										\
	for (_pos = SIZEOF_HL_HDR, _max = (_htlc)->in.pos;								\
	     _pos + SIZEOF_HL_DATA_HDR <= _max && (u_int16_t)ntohs(dh->len) <= (_max - _pos) - SIZEOF_HL_DATA_HDR;	\
	     memory_copy(_dhargs_buf, &((_htlc)->in.buf[_pos]), SIZEOF_HL_DATA_HDR)) {					\
		memory_copy(dh->data, &((_htlc)->in.buf[_pos + SIZEOF_HL_DATA_HDR]), ntohs(dh->len));			\
		_pos += SIZEOF_HL_DATA_HDR + ntohs(dh->len);

#define L32NTOH(_word, _addr) \
	do { u_int32_t _x; memory_copy(&_x, (_addr), 4); _word = ntohl(_x); } while (0)
#define S32HTON(_word, _addr) \
	do { u_int32_t _x; _x = htonl(_word); memory_copy((_addr), &_x, 4); } while (0)
#define L16NTOH(_word, _addr) \
	do { u_int16_t _x; memory_copy(&_x, (_addr), 2); _word = ntohs(_x); } while (0)
#define S16HTON(_word, _addr) \
	do { u_int16_t _x; _x = htons(_word); memory_copy((_addr), &_x, 2); } while (0)
#else
#define dh_start(_htlc)													\
{															\
	struct hl_data_hdr *dh = (struct hl_data_hdr *)(&((_htlc)->in.buf[SIZEOF_HL_HDR]));				\
	u_int32_t _pos, _max;												\
	for (_pos = SIZEOF_HL_HDR, _max = (_htlc)->in.pos;								\
	     _pos + SIZEOF_HL_DATA_HDR < _max && (u_int16_t)ntohs(dh->len) <= (_max - _pos) - SIZEOF_HL_DATA_HDR;	\
	     _pos += SIZEOF_HL_DATA_HDR + ntohs(dh->len),								\
	     dh = (struct hl_data_hdr *)(((u_int8_t *)dh) + SIZEOF_HL_DATA_HDR + ntohs(dh->len))) {

#define L32NTOH(_word, _addr) \
	_word = ntohl(*((u_int32_t *)_addr))
#define S32HTON(_word, _addr) \
	*((u_int32_t *)_addr) = htonl(_word)
#define L16NTOH(_word, _addr) \
	_word = ntohs(*((u_int16_t *)_addr))
#define S16HTON(_word, _addr) \
	*((u_int16_t *)_addr) = htons(_word)
#endif

#define dh_getint(_word)						\
do {									\
	if (ntohs(dh->len) == 4)					\
		_word = (u_int32_t)ntohl(*((u_int32_t *)dh->data));	\
	else /* if (ntohs(dh->len) == 2) */				\
		_word = (u_int32_t)ntohs(*((u_int16_t *)dh->data));	\
} while (0)

#define dh_end()	\
	}		\
}

static inline struct htlc_conn *
isclient (int uid)
{
	struct htlc_conn *htlcp;

	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next)
		if (htlcp->uid == uid)
			return htlcp;

	return 0;
}

#define X2X(_ptr, _len, _x1, _x2) \
do {						\
	char *_p = _ptr, *_end = _ptr + _len;	\
	for ( ; _p < _end; _p++)		\
		if (*_p == _x1)			\
			*_p = _x2;		\
} while (0)

#define CR2LF(_ptr, _len)	X2X(_ptr, _len, '\r', '\n')
#define LF2CR(_ptr, _len)	X2X(_ptr, _len, '\n', '\r')

#endif /* ndef _HXD_H */
