#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include "hxd.h"
#include "rcv.h"
#include "xmalloc.h"
#ifdef CONFIG_HOPE
#include "sha.h"
#include "md5.h"
#endif

#define MAX_HOTLINE_PACKET_LEN 0x40000

void
away_timer (struct htlc_conn *htlc)
{
	toggle_away(htlc);
	htlc->away = AWAY_INTERRUPTABLE;
}

void
test_away (struct htlc_conn *htlc)
{
	if (htlc->access_extra.is_server)
		return;
	if (htlc->away == AWAY_INTERRUPTED)
		htlc->away = 0;
	if (htlc->away == AWAY_INTERRUPTABLE) {
		toggle_away(htlc);
		htlc->away = AWAY_INTERRUPTED;
	}
	if (hxd_cfg.options.away_time && htlc->away != AWAY_PERM) {
		timer_delete_ptr(htlc);
		timer_add_secs(hxd_cfg.options.away_time, away_timer, htlc);
	}
}

void
rcv_magic (struct htlc_conn *htlc)
{
	htlc->rcv = rcv_hdr;
	qbuf_set(&htlc->in, 0, SIZEOF_HL_HDR);
}

#ifdef CONFIG_NETWORK
static void
htln_rcv_echo (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	u_int32_t len;

	if (ntohl(h->len) < 2)
		len = 0;
	else
		len = (ntohl(h->len) > MAX_HOTLINE_PACKET_LEN ? MAX_HOTLINE_PACKET_LEN : ntohl(h->len)) - 2;
	qbuf_add(&htlc->out, h, SIZEOF_HL_HDR + len);
	FD_SET(htlc->fd, &hxd_wfds);
}

static void
htln_rcv_user_change (struct htlc_conn *htlc)
{
	struct htlc_conn *client, *htlcp;
	u_int32_t uid = 0, icon = 0, color = 0;
	u_int16_t nlen = 0, got_color = 0, uid16, icon16, color16;
	char name[32];

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
			case HTLS_DATA_ICON:
				dh_getint(icon);
				break;
			case HTLS_DATA_COLOUR:
				dh_getint(color);
				got_color = 1;
				break;
			case HTLS_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(name, dh->data, nlen);
				name[nlen] = 0;
				break;
		}
	dh_end()
	if (!(client = isclient(uid))) {
		client = xmalloc(sizeof(struct htlc_conn));
		memset(client, 0, sizeof(struct htlc_conn));
		client->sockaddr = htlc->sockaddr;
		client->fd = htlc->fd;
		client->access.send_chat = 1;
		client->server_htlc = htlc;
		client->uid = uid;
	}
	if (icon)
		client->icon = icon;
	if (got_color)
		client->color = color;
	if (nlen) {
		memcpy(client->name, name, nlen);
		client->name[nlen] = 0;
	}
	hxd_log("join: %s %u %u %u", client->name, client->uid, client->icon, client->color);
	uid16 = htons(uid);
	icon16 = htons(icon);
	color16 = htons(color);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
		if (!htlcp->access_extra.user_getlist)
			continue;
		hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
			HTLS_DATA_UID, sizeof(uid16), &uid16,
			HTLS_DATA_ICON, sizeof(icon16), &icon16,
			HTLS_DATA_COLOUR, sizeof(color16), &color16,
			HTLS_DATA_NAME, nlen, client->name);
	}
	client->next = 0;
	client->prev = htlc_tail;
	htlc_tail->next = client;
	htlc_tail = client;
	nhtlc_conns++;
}

static void
htln_rcv_user_part (struct htlc_conn *htlc)
{
	u_int32_t uid = 0;
	struct htlc_conn *client;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLS_DATA_UID:
				dh_getint(uid);
				break;
		}
	dh_end()
	client = isclient(uid);
	if (client)
		htlc_close(client);
}

static void
htln_rcv (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	u_int32_t len;
	struct htlc_conn *htlcp;

	if (ntohl(h->len) < 2)
		len = 0;
	else
		len = (ntohl(h->len) > MAX_HOTLINE_PACKET_LEN ? MAX_HOTLINE_PACKET_LEN : ntohl(h->len)) - 2;
	h->uid = htons(htlc->uid);
	for (htlcp = server_htlc_list->next; htlcp; htlcp = htlcp->next) {
		if (htlcp == htlc->server_htlc)
			continue;
		qbuf_add(&htlcp->out, h, SIZEOF_HL_HDR + len);
		FD_SET(htlcp->fd, &hxd_wfds);
	}
	if (htlc->real_rcv)
		htlc->real_rcv(htlc);
}
#endif /* CONFIG_NETWORK */

void
rcv_hdr (struct htlc_conn *htlc)
{
	struct hl_hdr *h = (struct hl_hdr *)htlc->in.buf;
	u_int32_t len;
#ifdef CONFIG_NETWORK
	int is_server;
#endif

	if (ntohl(h->len) < 2)
		len = 0;
	else
		len = (ntohl(h->len) > MAX_HOTLINE_PACKET_LEN ? MAX_HOTLINE_PACKET_LEN : ntohl(h->len)) - 2;

#ifdef CONFIG_NETWORK
	is_server = htlc->access_extra.is_server;
	if (is_server) {
		if (!ntohs(h->uid) || ntohl(h->type) != HTLC_HDR_CHAT) {
#define HTLN_HDR_ECHO		((u_int32_t) 0x6e01)
			htlc->rcv = 0;
			switch (ntohl(h->type)) {
				case HTLN_HDR_ECHO:
					htlc->rcv = htln_rcv_echo;
					break;
				case HTLS_HDR_USER_CHANGE:
					htlc->rcv = htln_rcv_user_change;
					break;
				case HTLS_HDR_USER_PART:
					htlc->rcv = htln_rcv_user_part;
					break;
					
			}
			if (len)
				qbuf_set(&htlc->in, htlc->in.pos, len);
			return;
		} else {
			struct htlc_conn *client = isclient(ntohs(h->uid));
			if (!client) {
				hxd_log("!client: %u", ntohs(h->uid));
				client = xmalloc(sizeof(struct htlc_conn));
				memset(client, 0, sizeof(struct htlc_conn));
				client->sockaddr = htlc->sockaddr;
				client->fd = htlc->fd;
				client->uid = ntohs(h->uid);
				client->access.send_chat = 1;
				client->server_htlc = htlc;
				nhtlc_conns++;
			}
			qbuf_add(&client->in, h, SIZEOF_HL_HDR);
			client->in.len -= 22;
			client->in.pos += 22;
			htlc->current_uid = ntohs(h->uid);
			htlc = hxd_files[htlc->fd].conn.htlc = client;
			is_server = 0;
		}
	}
#endif /* CONFIG_NETWORK */

	htlc->trans = ntohl(h->trans);
	htlc->rcv = 0;
	switch (ntohl(h->type)) {
		case HTLC_HDR_CHAT:
			if (htlc->access.send_chat)
				htlc->rcv = rcv_chat;
			break;
		case HTLC_HDR_MSG:
			if (htlc->access_extra.msg)
				htlc->rcv = rcv_msg;
			break;
		case HTLC_HDR_USER_CHANGE:
			if (htlc->access.use_any_name)
				htlc->rcv = rcv_user_change;
			break;
		case HTLC_HDR_USER_GETINFO:
			if (htlc->access.get_user_info)
				htlc->rcv = rcv_user_getinfo;
			break;
		case HTLC_HDR_USER_KICK:
			if (htlc->access.disconnect_users)
				htlc->rcv = rcv_user_kick;
			break;
		case HTLC_HDR_FILE_LIST:
			if (htlc->access_extra.file_list)
				htlc->rcv = rcv_file_list;
			break;
		case HTLC_HDR_FILE_GET:
			if (htlc->access.download_files)
				htlc->rcv = rcv_file_get;
			break;
		case HTLC_HDR_FILE_PUT:
			if (htlc->access.upload_files)
				htlc->rcv = rcv_file_put;
			break;
		case HTLC_HDR_FILE_GETINFO:
			if (htlc->access_extra.file_getinfo)
				htlc->rcv = rcv_file_getinfo;
			break;
		case HTLC_HDR_FILE_SETINFO:
			if (htlc->access.rename_files   ||
			    htlc->access.rename_folders ||
			    htlc->access.comment_files  ||
			    htlc->access.comment_folders)
				htlc->rcv = rcv_file_setinfo;
			break;
		case HTLC_HDR_FILE_DELETE:
			if (htlc->access.delete_files ||
			    htlc->access.delete_folders)
				htlc->rcv = rcv_file_delete;
			break;
		case HTLC_HDR_FILE_MOVE:
			if (htlc->access.move_files ||
			    htlc->access.move_folders)
				htlc->rcv = rcv_file_move;
			break;
		case HTLC_HDR_FILE_MKDIR:
			if (htlc->access.create_folders)
				htlc->rcv = rcv_file_mkdir;
			break;
		case HTLC_HDR_FILE_SYMLINK:
			if (htlc->access.make_aliases)
				htlc->rcv = rcv_file_symlink;
			break;
#ifdef CONFIG_HOPE
		case HTLC_HDR_FILE_HASH:
			if (htlc->access_extra.file_hash)
				htlc->rcv = rcv_file_hash;
			break;
#endif
		case HTLC_HDR_CHAT_CREATE:
			if (htlc->access_extra.chat_private)
				htlc->rcv = rcv_chat_create;
			break;
		case HTLC_HDR_CHAT_INVITE:
			if (htlc->access_extra.chat_private)
				htlc->rcv = rcv_chat_invite;
			break;
		case HTLC_HDR_CHAT_JOIN:
			if (htlc->access_extra.chat_private)
				htlc->rcv = rcv_chat_join;
			break;
		case HTLC_HDR_CHAT_PART:
			if (htlc->access_extra.chat_private)
				htlc->rcv = rcv_chat_part;
			break;
		case HTLC_HDR_CHAT_DECLINE:
			if (htlc->access_extra.chat_private)
				htlc->rcv = rcv_chat_decline;
			break;
		case HTLC_HDR_CHAT_SUBJECT:
			if (htlc->access_extra.chat_private)
				htlc->rcv = rcv_chat_subject;
			break;
		case HTLC_HDR_NEWS_POST:
			if (htlc->access.post_news)
				htlc->rcv = rcv_news_post;
			break;
		case HTLC_HDR_ACCOUNT_READ:
			if (htlc->access.read_users)
				htlc->rcv = rcv_account_read;
			break;
		case HTLC_HDR_ACCOUNT_MODIFY:
			if (htlc->access.modify_users)
				htlc->rcv = rcv_account_modify;
			break;
		case HTLC_HDR_ACCOUNT_CREATE:
			if (htlc->access.create_users)
				htlc->rcv = rcv_account_create;
			break;
		case HTLC_HDR_ACCOUNT_DELETE:
			if (htlc->access.delete_users)
				htlc->rcv = rcv_account_delete;
			break;
		case HTLC_HDR_MSG_BROADCAST:
			if (htlc->access.can_broadcast)
				htlc->rcv = rcv_msg_broadcast;
			break;
		case HTLC_HDR_USER_GETLIST:
			if (htlc->access_extra.user_getlist)
				htlc->rcv = rcv_user_getlist;
			break;
		case HTLC_HDR_NEWS_GETFILE:
			if (htlc->access.read_news)
				htlc->rcv = rcv_news_getfile;
			break;
		case HTLC_HDR_LOGIN:
			if (htlc->access_extra.can_login)
				htlc->rcv = rcv_login;
			break;
	}
#ifdef CONFIG_NETWORK
	if (!is_server && htlc->rcv == rcv_chat) {
		htlc->real_rcv = htlc->rcv;
		htlc->rcv = htln_rcv;
	}
#endif
	if (!htlc->rcv)
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1,
			HTLS_DATA_TASKERROR, 17, "Permission denied");

	if (len) {
		qbuf_set(&htlc->in, htlc->in.pos, len);
	} else {
		int fd = htlc->fd;
		if (htlc->rcv)
			htlc->rcv(htlc);
		if (!hxd_files[fd].conn.htlc)
			return;
		test_away(htlc);
		htlc->rcv = rcv_hdr;
		qbuf_set(&htlc->in, 0, SIZEOF_HL_HDR);
	}
}

void
rcv_user_kick (struct htlc_conn *htlc)
{
	u_int32_t uid = 0, ban = 0;
	struct htlc_conn *htlcp;

	dh_start(htlc)
		if (ntohs(dh->type) == HTLC_DATA_BAN) {
			dh_getint(ban);
			continue;
		}
		if (ntohs(dh->type) != HTLC_DATA_UID)
			continue;
		dh_getint(uid);
		if ((htlcp = isclient(uid))) {
			if (!htlcp->access.cant_be_disconnected) {
				if (ban) {
					char abuf[16], *name, *login, *userid;
					inet_ntoa_r(htlcp->sockaddr.sin_addr, abuf, 16);
					if (htlcp->name[0] && 0) name = htlcp->name; else name = "*";
					if (htlcp->login[0] && 0) login = htlcp->login; else login = "*";
					if (htlcp->userid[0] && 0) userid = htlcp->userid; else userid = "*";
					addto_banlist(name, login, userid, abuf, htlc->name);
				}
				hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
				htlc_close(htlcp);
				return;
			} else {
				char buf[64];
				int len = snprintf(buf, sizeof(buf), "%s can not be disconnected", htlcp->name);
				hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, len, buf);
			}
		} else {
			hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "who?!?");
		}
	dh_end()
}

static inline void
dirmask (char *dst, char *src, char *mask)
{
#if 0
	while (*mask && *src && *mask++ == *src++) ;
	strcpy(dst, src);
#else
	char *p = strrchr(src, '/');
	if (!p)
		p = src;
	else
		p++;
	strcpy(dst, p);
#endif
}

#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
#include <signal.h>

static void
mask_signal (int how, int sig)
{
	sigset_t set;

	sigemptyset(&set);
	sigaddset(&set, sig);
	sigprocmask(how, &set, 0);
}
#endif

void
rcv_user_getinfo (struct htlc_conn *htlc)
{
	u_int32_t uid = 0;
	struct htlc_conn *htlcp;
	u_int8_t infobuf[8192], address[24];
	ssize_t len;
	struct htxf_conn *htxf;
	u_int8_t in_name[MAXPATHLEN], out_name[MAXPATHLEN];
	u_int16_t i;

	dh_start(htlc)
		if (ntohs(dh->type) != HTLC_DATA_UID)
			continue;
		dh_getint(uid);
		if (!(htlcp = isclient(uid))) {
			hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "who?!?");
			return;
		}
		inet_ntoa_r(htlcp->sockaddr.sin_addr, address, sizeof(address));
		len = snprintf(infobuf, sizeof(infobuf), "\
   name: %s\r\
  login: %s\r\
address: %s:%u\r\
 userid: %s\r", htlcp->name, htlcp->login, address, ntohs(htlcp->sockaddr.sin_port), htlcp->userid);
		len += snprintf(infobuf+len, sizeof(infobuf)-len, "\
    uid: %u\r\
  color: %u\r\
   icon: %u\r", htlcp->uid, htlcp->color, htlcp->icon);
		len += snprintf(infobuf+len, sizeof(infobuf)-len, " - Downloads - \r");
#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
		mask_signal(SIG_BLOCK, SIGCHLD);
#endif
		LOCK_HTXF(htlc);
		for (i = 0; i < HTXF_GET_MAX; i++) {
			htxf = htlcp->htxf_out[i];
			if (!htxf)
				continue;
			dirmask(out_name, htxf->path, htlcp->rootdir);
			len += snprintf(infobuf+len, sizeof(infobuf)-len, "%s  %u/%u %u/%u\r",
				out_name, htxf->data_pos, htxf->data_size, htxf->rsrc_pos, htxf->rsrc_size);
		}
		len += snprintf(infobuf+len, sizeof(infobuf)-len, " - Uploads - \r");
		for (i = 0; i < HTXF_PUT_MAX; i++) {
			htxf = htlcp->htxf_in[i];
			if (!htxf)
				continue;
			dirmask(in_name, htxf->path, htlcp->rootdir);
			len += snprintf(infobuf+len, sizeof(infobuf)-len, "%s  %u/%u %u/%u\r",
				in_name, htxf->data_pos, htxf->data_size, htxf->rsrc_pos, htxf->rsrc_size);
		}
		UNLOCK_HTXF(htlc);
#if defined(CONFIG_HTXF_CLONE) || defined(CONFIG_HTXF_FORK)
		mask_signal(SIG_UNBLOCK, SIGCHLD);
#endif

		if (len < 0)
			len = sizeof(infobuf);
		infobuf[len-1] = 0;
		hlwrite(htlc, HTLS_HDR_TASK, 0, 2,
			HTLS_DATA_USER_INFO, (u_int16_t)len, infobuf,
			HTLS_DATA_NAME, strlen(htlcp->name), htlcp->name);
	dh_end()
}

void
rcv_news_getfile (struct htlc_conn *htlc)
{
	news_send_file(htlc);
}

void
rcv_news_post (struct htlc_conn *htlc)
{
	struct htlc_conn *htlcp;
	u_int8_t buf[0xffff];
	u_int16_t len;
	struct tm tm;
	time_t t;

	hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
	t = time(0);
	localtime_r(&t, &tm);
	dh_start(htlc)
		if (ntohs(dh->type) != HTLC_DATA_NEWS_POST)
			continue;
		len = snprintf(buf, sizeof(buf), "From %s ", htlc->name);
		len += strftime(buf+len, sizeof(buf)-len, hxd_cfg.options.timeformat, &tm);
		len += snprintf(buf+len, sizeof(buf)-len, "\r\r%.*s\r\
_________________________________________________________\r",
			ntohs(dh->len), dh->data);
		for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
			if (!htlcp->access.read_news)
				continue;
			if (strcmp(htlc->newsfile, htlcp->newsfile))
				continue;
			hlwrite(htlcp, HTLS_HDR_NEWS_POST, 0, 1,
				HTLS_DATA_NEWS, len, buf);
		}
		CR2LF(buf, len);
		news_save_post(htlc->newsfile, buf, len);
	dh_end()
}

#ifdef CONFIG_HOPE
u_int16_t
hmac_xxx (u_int8_t *md, u_int8_t *key, u_int32_t keylen, u_int8_t *text, u_int32_t textlen, u_int8_t *macalg)
{
	u_int8_t mdkey[20];
	u_int8_t k_ipad[64], k_opad[64];
	unsigned int i;

	if (!strcmp(macalg, "SHA1")) {
		struct sha_ctx ctx;

		sha_init(&ctx);
		sha_update(&ctx, key, keylen);
		sha_update(&ctx, text, textlen);
		sha_final(md, &ctx);

		return 20;
	} else if (!strcmp(macalg, "MD5")) {
		struct md5_ctx ctx;

		md5_init_ctx(&ctx);
		md5_process_bytes(key, keylen, &ctx);
		md5_process_bytes(text, textlen, &ctx);
		md5_finish_ctx(&ctx, md);

		return 16;
	}

	if (keylen > 64) {
		if (!strcmp(macalg, "HMAC-SHA1")) {
			struct sha_ctx ctx;

			sha_init(&ctx);
			sha_update(&ctx, key, keylen);
			sha_final(mdkey, &ctx);
			keylen = 20;
		} else if (!strcmp(macalg, "HMAC-MD5")) {
			struct md5_ctx ctx;

			md5_init_ctx(&ctx);
			md5_process_bytes(key, keylen, &ctx);
			md5_finish_ctx(&ctx, mdkey);
			keylen = 16;
		} else if (!strncmp(macalg, "HMAC-HAVAL", 10)) {
			return 0;
		} else {
			return 0;
		}
		key = mdkey;
	}

	memcpy(k_ipad, key, keylen);
	memcpy(k_opad, key, keylen);
	memset(k_ipad+keylen, 0, 64 - keylen);
	memset(k_opad+keylen, 0, 64 - keylen);

	for (i = 0; i < 64; i++) {
		k_ipad[i] ^= 0x36;
		k_opad[i] ^= 0x5c;
	}

	if (!strcmp(macalg, "HMAC-SHA1")) {
		struct sha_ctx ctx;

		sha_init(&ctx);
		sha_update(&ctx, k_ipad, 64);
		sha_update(&ctx, text, textlen);
		sha_final(md, &ctx);

		sha_init(&ctx);
		sha_update(&ctx, k_opad, 64);
		sha_update(&ctx, md, 20);
		sha_final(md, &ctx);

		return 20;
	} else if (!strcmp(macalg, "HMAC-MD5")) {
		struct md5_ctx ctx;

		md5_init_ctx(&ctx);
		md5_process_bytes(k_ipad, 64, &ctx);
		md5_process_bytes(text, textlen, &ctx);
		md5_finish_ctx(&ctx, md);

		md5_init_ctx(&ctx);
		md5_process_bytes(k_opad, 64, &ctx);
		md5_process_bytes(md, 16, &ctx);
		md5_finish_ctx(&ctx, md);

		return 16;
	} else if (!strncmp(macalg, "HMAC-HAVAL", 10)) {
		return 0;
	} else {
		return 0;
	}
}
#endif

void
rcv_login (struct htlc_conn *htlc)
{
	u_int16_t plen = 0, llen = 0, nlen = 0, uid, color, icon16, high_uid;
	u_int32_t icon = 0;
	char abuf[24], login[32], name[32], given_name[32], password[32], given_password[32];
	u_int8_t ubuf[SIZEOF_HL_USERLIST_HDR + 32];
	struct hl_userlist_hdr *uh;
	struct htlc_conn *htlcp;
	int err;
#ifdef CONFIG_HOPE
	u_int8_t given_password_mac[20], given_login_mac[20];
	u_int16_t password_mac_len = 0, login_mac_len = 0;
	u_int16_t macalglen;
#endif

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				memcpy(given_name, dh->data, nlen);
				given_name[nlen] = 0;
				break;
			case HTLC_DATA_ICON:
				dh_getint(icon);
				break;
			case HTLC_DATA_LOGIN:
				llen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
#ifdef CONFIG_HOPE
				if (llen == 1 && !dh->data[0]) {
					login[0] = 0;
					break;
				} else if (htlc->macalg[0]) {
					if (llen > 20)
						break;
					login_mac_len = llen;
					memcpy(given_login_mac, dh->data, login_mac_len);
				} else
#endif
					hl_decode(login, dh->data, llen);
				login[llen] = 0;
				break;
			case HTLC_DATA_PASSWORD:
				plen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
#ifdef CONFIG_HOPE
				if (htlc->macalg[0]) {
					if (plen > 20)
						break;
					password_mac_len = plen;
					memcpy(given_password_mac, dh->data, password_mac_len);
					break;
				}
#endif
				hl_decode(given_password, dh->data, plen);
				given_password[plen] = 0;
				break;
#ifdef CONFIG_HOPE
			case HTLC_DATA_MAC_ALG:
				macalglen = ntohs(dh->len);
				if (macalglen >= 4) {
					macalglen = dh->data[2] > 31 ? 31 : dh->data[2];
					if (macalglen <= ntohs(dh->len) - 3) {
						memcpy(htlc->macalg, dh->data+3, macalglen);
						htlc->macalg[macalglen] = 0;
					}
				}
				break;
#endif
		}
	dh_end()

	if (llen == 1 && !login[0]) {
#ifdef CONFIG_HOPE
		u_int8_t macalglist[64];
		u_int16_t macalglistlen;
		u_int8_t sessionkey[64];
		int rfd;
		struct sockaddr_in saddr;
		int x;

		rfd = open("/dev/urandom", O_RDONLY);
		if (rfd < 0)
			goto xxx;
		if (read(rfd, sessionkey+6, 58) != 58) {
			close(rfd);
xxx:			hlwrite(htlc, HTLS_HDR_TASK, 1, 0);
			return;
		}
		close(rfd);

		x = sizeof(saddr);
		if (getsockname(htlc->fd, (struct sockaddr *)&saddr, &x)) {
			hxd_log("login: getsockname: %s", strerror(errno));
			saddr.sin_addr.s_addr = 0;
		}
		*((u_int32_t *)sessionkey) = saddr.sin_addr.s_addr;
		*((u_int16_t *)(sessionkey + 4)) = saddr.sin_port;

		macalglist[0] = 0;
		macalglist[1] = 1;
		macalglist[2] = strlen(htlc->macalg);
		memcpy(macalglist+3, htlc->macalg, macalglist[2]);
		macalglistlen = 3 + macalglist[2];
		if (htlc->macalg[0] && (strcmp(htlc->macalg, "HMAC-SHA1") && strcmp(htlc->macalg, "HMAC-MD5")
					&& strcmp(htlc->macalg, "SHA1") && strcmp(htlc->macalg, "MD5")))
			hlwrite(htlc, HTLS_HDR_TASK, 0, 2,
				HTLS_DATA_MAC_ALG, 0, 0,
				HTLS_DATA_SESSIONKEY, 64, sessionkey);
		else
			hlwrite(htlc, HTLS_HDR_TASK, 0, 3,
				HTLS_DATA_LOGIN, strlen(htlc->macalg), htlc->macalg,
				HTLS_DATA_MAC_ALG, macalglistlen, macalglist,
				HTLS_DATA_SESSIONKEY, 64, sessionkey);

		memcpy(htlc->sessionkey, sessionkey, 64);
		htlc->sklen = 64;
		return;
#else
		hlwrite(htlc, HTLS_HDR_TASK, 1, 0);
		return;
#endif
	}

	inet_ntoa_r(htlc->sockaddr.sin_addr, abuf, sizeof(abuf));

#ifdef CONFIG_HOPE
	if (login_mac_len) {
		DIR *dir;
		struct dirent *de;
		u_int8_t real_login_mac[20];
		int len;

		len = hmac_xxx(real_login_mac, "", 0, htlc->sessionkey, htlc->sklen, htlc->macalg);
		if (len && !memcmp(real_login_mac, given_login_mac, len)) {
			llen = 0;
			goto guest_login;
		}
		dir = opendir(hxd_cfg.paths.accounts);
		if (!dir) {
			hlwrite(htlc, HTLS_HDR_TASK, 1, 0);
			return;
		}
		while ((de = readdir(dir))) {
			len = hmac_xxx(real_login_mac, de->d_name, strlen(de->d_name), htlc->sessionkey, htlc->sklen, htlc->macalg);
			if (len && !memcmp(real_login_mac, given_login_mac, len)) {
				llen = strlen(de->d_name) > 31 ? 31 : strlen(de->d_name);
				memcpy(login, de->d_name, llen);
				login[llen] = 0;
				break;
			}
		}
		closedir(dir);
		if (!de) {
			hxd_log("%s@%s:%u -- could not find login", htlc->userid, abuf, ntohs(htlc->sockaddr.sin_port));
			htlc_close(htlc);
			return;
		}
	}
guest_login:
#endif

	if (!llen) {
		strcpy(login, "guest");
	}

	if (!account_trusted(login, htlc->userid, abuf)) {
		hxd_log("%s@%s:%u -- not trusted with login %s", htlc->userid, abuf, ntohs(htlc->sockaddr.sin_port), login);
		htlc_close(htlc);
		return;
	}
	if ((err = account_read(login, password, name, &htlc->access))) {
		hxd_log("%s@%s:%u -- error reading account %s: %s", htlc->userid, abuf, ntohs(htlc->sockaddr.sin_port), login, strerror(err));
		htlc_close(htlc);
		return;
	}
#ifdef CONFIG_HOPE
	if (password_mac_len) {
		u_int8_t real_password_mac[20];
		int len;

		len = hmac_xxx(real_password_mac, password, strlen(password), htlc->sessionkey, htlc->sklen, htlc->macalg);
		if (!len || memcmp(real_password_mac, given_password_mac, len)) {
			hxd_log("%s@%s:%u -- wrong password for %s", htlc->userid, abuf, ntohs(htlc->sockaddr.sin_port), login);
			htlc_close(htlc);
			return;
		}
	} else {
#endif
		if (strlen(password) != plen || (plen && strcmp(password, given_password))) {
			hxd_log("%s@%s:%u -- wrong password for %s", htlc->userid, abuf, ntohs(htlc->sockaddr.sin_port), login);
			htlc_close(htlc);
			return;
		}
#ifdef CONFIG_HOPE
	}
#endif

	if (htlc->access.use_any_name && nlen) {
		strcpy(htlc->name, given_name);
	} else {
		nlen = strlen(name);
		strcpy(htlc->name, name);
	}
	strcpy(htlc->login, login);
	if (check_banlist(htlc))
		return;

	htlc->icon = (u_int16_t)icon;
	account_get_access_extra(htlc);
	htlc->access_extra.can_login = 0;
	htlc->access_extra.is_server = 0;
#ifdef CONFIG_NETWORK
	if (htlc->icon == 0xffff) {
		char *l;
		int i;
		for (i = 0; (l = hxd_cfg.network.logins[i]); i++)
			if (!strcmp(l, htlc->login)) {
				htlc->access_extra.is_server = 1;
				break;
			}
	}
	if (htlc->access_extra.is_server) {
		htlc->next = 0;
		htlc->prev = server_htlc_tail;
		server_htlc_tail->next = htlc;
		server_htlc_tail = htlc;
	} else {
#endif
		htlc->next = 0;
		htlc->prev = htlc_tail;
		htlc_tail->next = htlc;
		htlc_tail = htlc;
#ifdef CONFIG_NETWORK
	}
#endif

	high_uid = htlc->fd;
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next)
		if (htlcp->uid > high_uid)
			high_uid = htlcp->uid;
	htlc->uid = high_uid + 1;

	nhtlc_conns++;
	if (htlc->access.disconnect_users)
		htlc->color = 2;
	else
		htlc->color = 0;

	hxd_log("%s@%s:%u - %s:%u:%u:%s - name:icon:uid:login",
		htlc->userid, abuf, ntohs(htlc->sockaddr.sin_port),
		htlc->name, htlc->icon, htlc->uid, htlc->login);

	uid = htons(htlc->uid);
	icon16 = htons(htlc->icon);
	color = htons(htlc->color);
#ifdef CONFIG_NETWORK
	if (!htlc->access_extra.is_server) {
		for (htlcp = server_htlc_list->next; htlcp; htlcp = htlcp->next) {
			hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
				HTLS_DATA_UID, sizeof(uid), &uid,
				HTLS_DATA_ICON, sizeof(icon16), &icon16,
				HTLS_DATA_COLOUR, sizeof(color), &color,
				HTLS_DATA_NAME, nlen, htlc->name);
		}
	}
#endif
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
		if (htlcp == htlc || !htlcp->access_extra.user_getlist)
			continue;
		hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
			HTLS_DATA_UID, sizeof(uid), &uid,
			HTLS_DATA_ICON, sizeof(icon16), &icon16,
			HTLS_DATA_COLOUR, sizeof(color), &color,
			HTLS_DATA_NAME, nlen, htlc->name);
	}
	account_getconf(htlc);
#ifdef CONFIG_TRACKER_REGISTER
	tracker_register_timer(0);
#endif
	uh = (struct hl_userlist_hdr *)(ubuf - SIZEOF_HL_DATA_HDR);
	uh->uid = htons(htlc->uid);
	uh->icon = htons(htlc->icon);
	uh->color = htons(htlc->color);
	uh->nlen = htons(nlen);
	memcpy(uh->name, htlc->name, nlen);
	hlwrite(htlc, HTLS_HDR_TASK, 0, 1,
		HTLS_DATA_UID, 2, &uh->uid);
	hlwrite(htlc, HTLS_HDR_USER_SELFINFO, 0, 2,
		HTLS_DATA_ACCESS, 8, &htlc->access,
		HTLS_DATA_USER_LIST, (SIZEOF_HL_USERLIST_HDR - SIZEOF_HL_DATA_HDR) + nlen, ubuf);
	if (!htlc->access.dont_show_agreement)
		agreement_send_file(htlc);
}

static char *
cr_strtok_r (char *ptr, char **state_ptr)
{
	char *p, *ret;

	if (ptr) {
		if (!(p = strchr(ptr, '\r')))
			return 0;
	} else if (!*state_ptr) {
		return 0;
	} else if (!(p = strchr(*state_ptr, '\r'))) {
		p = *state_ptr;
		*state_ptr = 0;
		return p;
	}
	*p = 0;
	ret = *state_ptr ? *state_ptr : ptr;
	*state_ptr = &(p[1]);

	return ret;
}

char *chat_fmt = "\r%13.13s:  %s";
char *chat_opt_fmt = "\r *** %s %s";

void
rcv_chat (struct htlc_conn *htlc)
{
	u_int32_t chatref = 0, away = 0, style = 0;
	struct htlc_chat *chat = 0;
	char *p, *last = 0;
	u_int16_t len = 0;
	char buf[4096], chatbuf[4096], *fmt = chat_fmt;
	struct htlc_conn *htlcp;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_STYLE:
				dh_getint(style);
				if (style == 1)
					fmt = chat_opt_fmt;
				break;
			case HTLC_DATA_CHAT_ID:
				dh_getint(chatref);
				break;
			case HTLC_DATA_CHAT:
				len = ntohs(dh->len) > 2048 ? 2048 : ntohs(dh->len);
				memcpy(chatbuf, dh->data, len);
				break;
			case HTLC_DATA_CHAT_AWAY:	
				dh_getint(away);
				break;
		}
	dh_end()
	if (away && htlc->away != AWAY_INTERRUPTED) {
		toggle_away(htlc);
		if (!htlc->away)
			htlc->away = AWAY_PERM;
		else
			htlc->away = 0;
		if (hxd_cfg.options.away_time) {
			timer_delete_ptr(htlc);
			if (!htlc->away)
				timer_add_secs(hxd_cfg.options.away_time, away_timer, htlc);
		}
	}
	if (chatref) {
		chat = chat_lookup_ref(chatref);
		chatref = htonl(chatref);
	}
	chatbuf[len] = 0;
	if (chatbuf[0] == '/') {
		command_chat(htlc, &chatbuf[1]);
		return;
	}
	if (!(p = cr_strtok_r(chatbuf, &last)))
		p = chatbuf;
	do {
		len = snprintf(buf, sizeof(buf), fmt, htlc->name, p);
		if (len == (u_int16_t)-1)
			len = sizeof(buf);
		if (chat) {
			for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
				if (!chat_isset(htlcp, chat))
					continue;
				hlwrite(htlcp, HTLS_HDR_CHAT, 0, 2,
					HTLS_DATA_CHAT, len, buf,
					HTLS_DATA_CHAT_ID, sizeof(chatref), &chatref);
			}
		} else {
			for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
				if (!htlcp->access.read_chat)
					continue;
#if 0
	if (htlcp == htlc) {
		u_int16_t uid = htons(htlc->uid), icon16 = htons(htlc->icon);
		u_int16_t color = htons((htlc->color&1)?(htlc->color-1):(htlc->color+1));
		htlc->color=ntohs(color);
		hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
			HTLS_DATA_UID, sizeof(uid), &uid,
			HTLS_DATA_ICON, sizeof(icon16), &icon16,
			HTLS_DATA_COLOUR, sizeof(color), &color,
			HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
		hlwrite(htlcp, HTLS_HDR_NEWS_POST, 0, 1,
			HTLS_DATA_NEWS, len, buf);
	}
#endif
				hlwrite(htlcp, HTLS_HDR_CHAT, 0, 1, HTLS_DATA_CHAT, len, buf);
			}
		}
	} while ((p = cr_strtok_r(0, &last)));
}

void
rcv_msg (struct htlc_conn *htlc)
{
	u_int32_t uid = 0;
	u_int16_t msglen = 0, uid16;
	u_int8_t msgbuf[4096];
	struct htlc_conn *htlcp;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_UID:
				dh_getint(uid);
				break;
			case HTLC_DATA_MSG:
				msglen = ntohs(dh->len) > sizeof(msgbuf) ? sizeof(msgbuf) : ntohs(dh->len);
				memcpy(msgbuf, dh->data, msglen);
				break;
		}
	dh_end()
	if (uid && msglen) {
		if ((htlcp = isclient(uid))) {
			hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
			uid16 = htons(htlc->uid);
			hlwrite(htlcp, HTLS_HDR_MSG, 0, 3,
				HTLS_DATA_UID, sizeof(uid16), &uid16,
				HTLS_DATA_MSG, msglen, msgbuf,
				HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
		} else
			hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "who?!?");
	} else
		hlwrite(htlc, HTLS_HDR_TASK, 1, 1, HTLS_DATA_TASKERROR, 6, "huh?!?");
}

void
rcv_msg_broadcast (struct htlc_conn *htlc)
{
	u_int16_t msglen = 0, uid;
	u_int8_t msgbuf[4096];
	struct htlc_conn *htlcp;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_MSG:
				msglen = ntohs(dh->len) > sizeof(msgbuf) ? sizeof(msgbuf) : ntohs(dh->len);
				memcpy(msgbuf, dh->data, msglen);
				break;
		}
	dh_end()
	if (!msglen)
		return;
	uid = htons(htlc->uid);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
		hlwrite(htlcp, HTLS_HDR_MSG_BROADCAST, 0, 3,
			HTLS_DATA_UID, sizeof(uid), &uid,
			HTLS_DATA_MSG, msglen, msgbuf,
			HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
	}
	hlwrite(htlc, HTLS_HDR_TASK, 0, 0);
}

void
rcv_user_change (struct htlc_conn *htlc)
{
	int diff = 0;
	u_int8_t nlen;
	u_int32_t icon = 0;
	struct htlc_conn *htlcp;
	u_int16_t uid, icon16, color;

	dh_start(htlc)
		switch (ntohs(dh->type)) {
			case HTLC_DATA_NAME:
				nlen = ntohs(dh->len) > 31 ? 31 : ntohs(dh->len);
				if (strlen(htlc->name) != nlen || memcmp(dh->data, htlc->name, nlen))
					diff = 1;
				memcpy(htlc->name, dh->data, nlen);
				htlc->name[nlen] = 0;
				break;
			case HTLC_DATA_ICON:
				dh_getint(icon);
				if (htlc->icon != icon) {
					diff = 1;
					htlc->icon = (u_int16_t)icon;
				}
				break;
		}
	dh_end()
	if (!diff)
		return;

	if (!htlc->flags.visible)
		return;

	uid = htons(htlc->uid);
	icon16 = htons(htlc->icon);
	color = htons(htlc->color);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
		if (!htlcp->access_extra.user_getlist)
			continue;
		hlwrite(htlcp, HTLS_HDR_USER_CHANGE, 0, 4,
			HTLS_DATA_UID, sizeof(uid), &uid,
			HTLS_DATA_ICON, sizeof(icon16), &icon16,
			HTLS_DATA_COLOUR, sizeof(color), &color,
			HTLS_DATA_NAME, strlen(htlc->name), htlc->name);
	}
}

void
rcv_user_getlist (struct htlc_conn *htlc)
{
	struct qbuf *q = &htlc->out;
	u_int32_t this_off = q->pos + q->len, pos = this_off + SIZEOF_HL_HDR;
	struct hl_hdr h;
	struct hl_userlist_hdr uh;
	u_int16_t nlen, nclients = 0;
	struct htlc_conn *htlcp;

	q->len += SIZEOF_HL_HDR;
	q->buf = xrealloc(q->buf, q->pos + q->len);
	for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
		if (!htlcp->flags.visible)
			continue;
		nlen = strlen(htlcp->name);
		uh.type = htons(HTLS_DATA_USER_LIST);
		uh.len = htons(8 + nlen);
		uh.uid = htons(htlcp->uid);
		uh.icon = htons(htlcp->icon);
		uh.color = htons(htlcp->color);
		uh.nlen = htons(nlen);
		q->len += SIZEOF_HL_USERLIST_HDR + nlen;
		q->buf = xrealloc(q->buf, q->pos + q->len);
		memcpy(&q->buf[pos], &uh, SIZEOF_HL_USERLIST_HDR);
		pos += SIZEOF_HL_USERLIST_HDR;
		memcpy(&q->buf[pos], htlcp->name, nlen);
		pos += nlen;
		nclients++;
	}

	h.type = htonl(HTLS_HDR_TASK);
	h.trans = htonl(htlc->trans);
	htlc->trans++;
	h.uid = 0;
	h.flag = 0;
	h.len = h.len2 = htonl((pos - this_off) - (SIZEOF_HL_HDR - sizeof(h.hc)));
	h.hc = htons(nclients);
	memory_copy(q->buf + this_off, &h, SIZEOF_HL_HDR);

	FD_SET(htlc->fd, &hxd_wfds);
}
